@@ -17,9 +17,9 @@
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
 
-    You should have received a copy of the GNU Lesser General Public
+    You should have received a copy of the GNU General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 
     As a special exception, if you link this library with other files,
     compiled with a Free Software compiler, to produce an executable, this
@@ -382,12 +382,10 @@ dwim_predicate_list(M:Head, DWIMs) :-
 	setof(DWIM, dwim_pred(M:Head, DWIM), DWIMs), !.
 dwim_predicate_list(Head, DWIMs) :-
 	setof(DWIM, '$similar_module'(Head, DWIM), DWIMs), !.
-dwim_predicate_list(M:Goal, DWIMs) :-
+dwim_predicate_list(_:Goal, DWIMs) :-
 	setof(Module:Goal,
-	      ( current_module(Module),
-		default_module(Module, M),
-		current_predicate(_, Module:Goal)
-	      ), DWIMs).
+	      current_predicate(_, Module:Goal),
+	      DWIMs).
 
 %%	dwim_pred(:Head, -DWIM) is nondet.
 %
