@@ -3,9 +3,9 @@
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        jan@swi.psy.uva.nl
+    E-mail:        J.Wielemaker@uva.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2002, University of Amsterdam
+    Copyright (C): 1985-2009, University of Amsterdam
 
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 
     As a special exception, if you link this library with other files,
     compiled with a Free Software compiler, to produce an executable, this
@@ -32,12 +32,15 @@
 :- module(check,
 	[ check/0,			% run all checks
 	  list_undefined/0,		% list undefined predicates
+	  list_undefined/1,		% +Options
 	  list_autoload/0,		% list predicates that need autoloading
 	  list_redefined/0		% list redefinitions
 	]).
 :- use_module(library(lists)).
+:- use_module(library(system)).
+:- use_module(library(option)).
 
-:- style_check(+dollar).		% lock these predicates
+:- set_prolog_flag(generate_debug_info, false).
 
 /** <module> Consistency checking
 
@@ -47,6 +50,8 @@ loaded Prolog program.
 @see	prolog_xref.pl
 */
 
+:- predicate_options(list_undefined/1, 1, [scan(oneof([local,global]))]).
+
 %%	check
 %
 %	Run all consistency checks defined in this library
@@ -54,7 +59,7 @@ loaded Prolog program.
 check :-
 	print_message(informational,
 		      check(pass(1, 'Undefined predicates'))),
-	list_undefined(silent),
+	list_undefined(silent, global, false),
 	print_message(informational,
 		      check(pass(2, 'Redefined system and global predicates'))),
 	list_redefined,
@@ -62,28 +67,50 @@ check :-
 		      check(pass(3, 'Predicates that need autoloading'))),
 	list_autoload.
 
-%%	list_undefined
+%%	list_undefined is det.
+%%	list_undefined(+Options) is det.
 %
 %	List predicates names refered to  in  a  clause  body,  but  not
 %	defined.  This forms a "Quick and Dirty" alternative for a cross
-%	referencing tool.
+%	referencing tool.  Options:
+%
+%	    * scan(+Scan)
+%	    If =local=, only scan the module holding the undefined
+%	    predicate for references; if =global= (default), scan the
+%	    whole program.  Global scanning finds references through
+%	    qualified goals (i.e., M:G), but can take long on big
+%	    programs.
+%
+%	    * unreferenced(+Boolean)
+%	    If =true= (default), report undefined predicates that are
+%	    not referenced by any clause body.
 %
 %	@see gxref/0 provides a graphical cross-referencer.
+%	@see make/0 calls list_undefined/1 using scan(local)
 
 list_undefined :-
-	list_undefined(informational).
+	list_undefined(informational, global, false).
 
-list_undefined(Level) :-
-	'$style_check'(Old, Old),
-	style_check(+dollar),
-	call_cleanup(list_undefined_(Level), '$style_check'(_, Old)).
+list_undefined(Options) :-
+	option(scan(Scan), Options, global),
+	option(unreferenced(Unreferenced), Options, false),
+	list_undefined(informational, Scan, Unreferenced).
 
-list_undefined_(Level) :-
+
+list_undefined(Level, How, Unreferenced) :-
+	setup_call_cleanup(
+	    ( current_prolog_flag(access_level, OldLevel),
+	      set_prolog_flag(access_level, system)
+	    ),
+	    list_undefined_(Level, How, Unreferenced),
+	    set_prolog_flag(access_level, OldLevel)).
+
+list_undefined_(Level, How, Unreferenced) :-
 	findall(Pred, undefined_predicate(Pred), Preds),
 	(   Preds == []
 	->  true
 	;   print_message(Level, check(find_references(Preds))),
-	    find_references(Preds, Pairs),
+	    find_references(Preds, How, Pairs, Unrefs),
 	    (	Pairs == []
 	    ->	true
 	    ;   print_message(warning, check(undefined_predicates)),
@@ -93,6 +120,17 @@ list_undefined_(Level) :-
 		    fail
 		;   true
 		)
+	    ),
+	    (	Unreferenced == true,
+		Unrefs \== []
+	    ->  print_message(warning, check(undefined_unreferenced_predicates)),
+		(   member(Module:Head, Unrefs),
+		    print_message(warning,
+				  check(undefined_unreferenced(Module:Head))),
+		    fail
+		;   true
+		)
+	    ;   true
 	    )
 	).
 
@@ -101,7 +139,7 @@ undefined_predicate(Module:Head) :-
 	\+ predicate_property(Module:Head, imported_from(_)).
 
 
-%%	find_references(+Heads, -HeadRefs:list) is det.
+%%	find_references(+Heads, +How, -HeadRefs:list, -UnrefHeads:list) is det.
 %
 %	Find references to the given  predicates.   For  speedup we only
 %	look for references from the  same   module.  This  isn't really
@@ -109,17 +147,20 @@ undefined_predicate(Module:Head) :-
 %	through meta-calls, it isn't too bad either.
 %
 %	@param HeadRefs List of Head-Refs
+%	@param UnrefHeads List of Head
+%	@param How is one of =local= or =global=
 
-find_references([], []).
-find_references([H|T0], [H-Refs|T]) :-
+find_references([], _, [], []).
+find_references([H|T0], How, [H-Refs|T1], T2) :-
 	(   ignore(H = M:_),
 	    findall(Ref, referenced(H, M, Ref), Refs)
-	;   findall(Ref, referenced(H, _, Ref), Refs)
+	;   How == global,
+	    findall(Ref, referenced(H, _, Ref), Refs)
 	),
 	Refs \== [], !,
-	find_references(T0, T).
-find_references([_|T0], T) :-
-	find_references(T0, T).
+	find_references(T0, How, T1, T2).
+find_references([H|T0], How, T1, [H|T2]) :-
+	find_references(T0, How, T1, T2).
 
 %%	referenced(+Predicate, ?Module, -ClauseRef) is nondet.
 %
@@ -137,28 +178,39 @@ referenced(Term, Module, Ref) :-
 
 %%	list_autoload
 %
-%	Show predicates that need be linked via the autoload mechanism
+%	Show predicates that are not defined, but will be loaded on
+%	demand through the autoloader.
+%
+%	The behaviour of this predicate depends  on the system-mode (see
+%	system_mode/1): in normal  operation  it   only  lists  autoload
+%	requirements from user-module. In system-mode it also lists such
+%	requirements for the system modules.
 
 list_autoload :-
-	'$style_check'(Old, Old),
-	style_check(+dollar),
-	current_prolog_flag(autoload, OldAutoLoad),
-	set_prolog_flag(autoload, false),
-	call_cleanup(list_autoload_,
-		     (	 set_prolog_flag(autoload, OldAutoLoad),
-			 '$style_check'(_, Old)
-		     )).
+	setup_call_cleanup(
+	    ( current_prolog_flag(access_level, OldLevel),
+	      current_prolog_flag(autoload, OldAutoLoad),
+	      set_prolog_flag(access_level, system),
+	      set_prolog_flag(autoload, false)
+	    ),
+	    list_autoload_(OldLevel),
+	    ( set_prolog_flag(access_level, OldLevel),
+	      set_prolog_flag(autoload, OldAutoLoad)
+	    )).
 
-list_autoload_ :-
-	(   setof(Lib-Pred, autoload_predicate(Module, Lib, Pred), Pairs),
+list_autoload_(SystemMode) :-
+	(   setof(Lib-Pred,
+		  autoload_predicate(Module, Lib, Pred, SystemMode),
+		  Pairs),
 	    print_message(informational,
 			  check(autoload(Module, Pairs))),
 	    fail
 	;   true
 	).
 
-autoload_predicate(Module, Library, Name/Arity) :-
+autoload_predicate(Module, Library, Name/Arity, SystemMode) :-
 	predicate_property(Module:Head, undefined),
+	check_module_enabled(Module, SystemMode),
 	(   \+ predicate_property(Module:Head, imported_from(_)),
 	    functor(Head, Name, Arity),
 	    '$find_library'(Module, Name, Arity, _LoadModule, Library),
@@ -166,15 +218,22 @@ autoload_predicate(Module, Library, Name/Arity) :-
 	->  true
 	).
 
+check_module_enabled(_, system) :- !.
+check_module_enabled(Module, _) :-
+	\+ import_module(Module, system).
+
 
 %%	list_redefined
 %
 %	Show redefined system predicates
 
 list_redefined :-
-	'$style_check'(Old, Old),
-	style_check(+dollar),
-	call_cleanup(list_redefined_, '$style_check'(_, Old)).
+	setup_call_cleanup(
+	    ( current_prolog_flag(access_level, OldLevel),
+	      set_prolog_flag(access_level, system)
+	    ),
+	    list_redefined_,
+	    set_prolog_flag(access_level, OldLevel)).
 
 list_redefined_ :-
 	current_module(Module),
@@ -184,6 +243,7 @@ list_redefined_ :-
 	(   global_module(Super),
 	    Super \== Module,
 	    '$c_current_predicate'(_, Super:Head),
+	    \+ redefined_ok(Head),
 	    '$syspreds':'$defined_predicate'(Super:Head),
 	    \+ predicate_property(Super:Head, (dynamic)),
 	    \+ predicate_property(Super:Head, imported_from(Module)),
@@ -194,6 +254,9 @@ list_redefined_ :-
 	fail.
 list_redefined_.
 
+redefined_ok('$mode'(_,_)).
+redefined_ok('$pldoc'(_,_,_,_)).
+redefined_ok('$pred_option'(_,_,_,_)).
 
 global_module(user).
 global_module(system).
@@ -211,7 +274,7 @@ prolog:message(check(pass(N, Comment))) -->
 prolog:message(check(find_references(Preds))) -->
 	{ length(Preds, N)
 	},
-	[ 'Scanning references for ~D possibly undefined predicates'-[N] ].
+	[ 'Scanning for references to ~D possibly undefined predicates'-[N] ].
 prolog:message(check(undefined_predicates)) -->
 	[ 'The predicates below are not defined. If these are defined', nl,
 	  'at runtime using assert/1, use :- dynamic Name/Arity.', nl, nl
@@ -220,6 +283,12 @@ prolog:message(check(undefined(Pred, Refs))) -->
 	predicate(Pred),
 	[ ', which is referenced by', nl ],
 	referenced_by(Refs).
+prolog:message(check(undefined_unreferenced_predicates)) -->
+	[ 'The predicates below are not defined, and are not', nl,
+	  'referenced.', nl, nl
+	].
+prolog:message(check(undefined_unreferenced(Pred))) -->
+	predicate(Pred).
 prolog:message(check(autoload(Module, Pairs))) -->
 	{ module_property(Module, file(Path))
 	}, !,
@@ -311,7 +380,7 @@ short_filename(Path, Spec) :-
 short_filename(Path, Path).
 
 aliased_path(Path, Len-Spec) :-
-	setof(Alias, Spec^file_search_path(Alias, Spec), Aliases),
+	setof(Alias, Spec^(user:file_search_path(Alias, Spec)), Aliases),
 	member(Alias, Aliases),
 	Term =.. [Alias, '.'],
 	absolute_file_name(Term,
