@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 
     As a special exception, if you link this library with other files,
     compiled with a Free Software compiler, to produce an executable, this
@@ -31,6 +31,8 @@
 
 :- module((record),
 	  [ (record)/1,			% +Record
+	    current_record/2,		% ?Name, ?Term
+	    current_record_predicate/2,	% ?Record, :PI
 	    op(1150, fx, record)
 	  ]).
 :- use_module(library(error)).
@@ -57,6 +59,14 @@ _directive_.  Here is a simple example declaration and some calls.
 @author Richard O'Keefe
 */
 
+:- multifile
+	error:has_type/2,
+	prolog:generated_predicate/1.
+
+error:has_type(record(M:Name), X) :-
+	current_record(Name, M, _, X, IsX), !,
+	call(M:IsX).
+
 %%	record(+RecordDef)
 %
 %	Define access predicates for a compound-term. RecordDef is of
@@ -69,6 +79,7 @@ _directive_.  Here is a simple example declaration and some calls.
 %	info the following predicates:
 %
 %	  * <constructor>_<name>(Record, Value)
+%	  * <constructor>_data(?Name, ?Record, ?Value)
 %	  * default_<constructor>(-Record)
 %	  * is_<constructor>(@Term)
 %	  * make_<constructor>(+Fields, -Record)
@@ -79,6 +90,7 @@ _directive_.  Here is a simple example declaration and some calls.
 %	  * set_<constructor>_fields(+Fields, +Record0, -Record).
 %	  * set_<constructor>_fields(+Fields, +Record0, -Record, -RestFields).
 %	  * set_<constructor>_field(+Field, +Record0, -Record).
+%	  * user:current_record(:<constructor>)
 
 record(Record) :-
 	throw(error(context_error(nodirective, record(Record)), _)).
@@ -94,7 +106,7 @@ compile_records(Spec, Clauses) :-
 
 compile_records(Var) -->
 	{ var(Var), !,
-	  type_error(record_declaration, Var)
+	  instantiation_error(Var)
 	}.
 compile_records((A,B)) -->
 	compile_record(A),
@@ -111,17 +123,94 @@ compile_record(RecordDef) -->
 	  defaults(Args, Defs, TypedArgs),
 	  types(TypedArgs, Names, Types),
 	  atom_concat(default_, Constructor, DefName),
+	  atom_concat(Constructor, '_data', DataName),
 	  DefRecord =.. [Constructor|Defs],
 	  DefClause =.. [DefName,DefRecord],
 	  length(Names, Arity)
 	},
 	[ DefClause ],
 	access_predicates(Names, 1, Arity, Constructor),
+	data_predicate(Names, 1, Arity, Constructor, DataName),
 	set_predicates(Names, 1, Arity, Types, Constructor),
 	set_field_predicates(Names, 1, Arity, Types, Constructor),
 	make_predicate(Constructor),
-	is_predicate(Constructor, Types).
+	is_predicate(Constructor, Types),
+	current_clause(RecordDef).
 
+:- meta_predicate
+	current_record(?, :),
+	current_record_predicate(?, :).
+:- multifile
+	current_record/5.		% Name, Module, Term, X, IsX
+
+%%	current_record(?Name, :Term)
+%
+%	True if Name is the  name  of   a  record  defined in the module
+%	associated with Term  and  Term   is  the  user-provided  record
+%	declaration.
+
+current_record(Name, M:Term) :-
+	current_record(Name, M, Term, _, _).
+
+current_clause(RecordDef) -->
+	{ prolog_load_context(module, M),
+	  functor(RecordDef, Name, _),
+	  atom_concat(is_, Name, IsName),
+	  IsX =.. [IsName, X]
+	},
+	[ (record):current_record(Name, M, RecordDef, X, IsX)
+	].
+
+
+%%	current_record_predicate(?Record, ?PI) is nondet.
+%
+%	True if PI is the predicate indicator for an access predicate to
+%	Record. This predicate is intended   to support cross-referencer
+%	tools.
+
+current_record_predicate(Record, M:PI) :-
+	(   ground(PI)
+	->  Det = true
+	;   true
+	),
+	current_record(Record, M:RecordDef),
+	(   general_record_pred(Record, M:PI)
+	;   RecordDef =.. [_|Args],
+	    defaults(Args, _Defs, TypedArgs),
+	    types(TypedArgs, Names, _Types),
+	    member(Field, Names),
+	    field_record_pred(Record, Field, M:PI)
+	),
+	(   Det == true
+	->  !
+	;   true
+	).
+
+general_record_pred(Record, _:Name/1) :-
+	atom_concat(is_, Record, Name).
+general_record_pred(Record, _:Name/1) :-
+	atom_concat(default_, Record, Name).
+general_record_pred(Record, _:Name/A) :-
+	member(A, [2,3]),
+	atom_concat(make_, Record, Name).
+general_record_pred(Record, _:Name/3) :-
+	atom_concat(Record, '_data', Name).
+general_record_pred(Record, _:Name/A) :-
+	member(A, [3,4]),
+	atomic_list_concat([set_, Record, '_fields'], Name).
+general_record_pred(Record, _:Name/3) :-
+	atomic_list_concat([set_, Record, '_field'], Name).
+
+field_record_pred(Record, Field, _:Name/2) :-
+	atomic_list_concat([Record, '_', Field], Name).
+field_record_pred(Record, Field, _:Name/A) :-
+	member(A, [2,3]),
+	atomic_list_concat([set_, Field, '_of_', Record], Name).
+field_record_pred(Record, Field, _:Name/2) :-
+	atomic_list_concat([nb_set_, Field, '_of_', Record], Name).
+
+prolog:generated_predicate(P) :-
+	current_record_predicate(_, P).
 
 %%	make_predicate(+Constructor)// is det.
 %
@@ -205,11 +294,25 @@ type_checks([Type|T], [V|Vars], (Goal, Body)) :-
 %	Inline type checking calls.
 
 type_goal(Type, Var, Body) :-
-	Type \= list(_),
-	clause(error:has_type(Type, Var), Body), !.
-type_goal(Type, Var, is_of_type(Type, Var)).
+	defined_type(Type, Var, Body), !.
+type_goal(record(Record), Var, Body) :- !,
+	atom_concat(is_, Record, Pred),
+	Body =.. [Pred,Var].
+type_goal(Record, Var, Body) :-
+	atom(Record), !,
+	atom_concat(is_, Record, Pred),
+	Body =.. [Pred,Var].
+type_goal(Type, _, _) :-
+	domain_error(type, Type).
+
+defined_type(Type, Var, error:Body) :-
+	clause(error:has_type(Type, Var), Body).
 
 
+clean_body(M:(A0,B0), G) :- !,
+	clean_body(M:A0, A),
+	clean_body(M:B0, B),
+	clean_body((A,B), G).
 clean_body((A0,true), A) :- !,
 	clean_body(A0, A).
 clean_body((true,A0), A) :- !,
@@ -239,6 +342,22 @@ access_predicates([Name|NT], I, Arity, Constructor) -->
 	access_predicates(NT, I2, Arity, Constructor).
 
 
+%%	data_predicate(+Names, +Idx0, +Arity, +Constructor, +DataName)// is det.
+%
+%	Create the <constructor>_data(Name, Record, Value) predicate.
+
+data_predicate([], _, _, _, _) -->
+	[].
+data_predicate([Name|NT], I, Arity, Constructor, DataName) -->
+	{ functor(Record, Constructor, Arity),
+	  arg(I, Record, Value),
+	  Clause =.. [DataName, Name, Record, Value],
+	  I2 is I + 1
+	},
+	[Clause],
+	data_predicate(NT, I2, Arity, Constructor, DataName).
+
+
 %%	set_predicates(+Names, +Idx0, +Arity, +Types, +Constructor)// is det.
 %
 %	Create the clauses
@@ -262,10 +381,11 @@ set_predicates([Name|NT], I, Arity, [Type|TT], Constructor) -->
 	  ->  Clause = Head,
 	      SetClause = (SetHead :- setarg(I, Term, Value)),
 	      NBSetClause = (NBSetHead :- nb_setarg(I, Term, Value))
-	  ;   Clause = (Head :- must_be(Type, Value)),
-	      SetClause = (SetHead :- must_be(Type, Value),
+	  ;   type_check(Type, Value, MustBe),
+	      Clause = (Head :- MustBe),
+	      SetClause = (SetHead :- MustBe,
 				      setarg(I, Term, Value)),
-	      NBSetClause = (NBSetHead :- must_be(Type, Value),
+	      NBSetClause = (NBSetHead :- MustBe,
 				          nb_setarg(I, Term, Value))
 	  ),
 	  I2 is I + 1
@@ -273,6 +393,15 @@ set_predicates([Name|NT], I, Arity, [Type|TT], Constructor) -->
 	[ Clause, SetClause, NBSetClause ],
 	set_predicates(NT, I2, Arity, TT, Constructor).
 
+type_check(Type, Value, must_be(Type, Value)) :-
+	defined_type(Type, Value, _), !.
+type_check(record(Spec), Value, must_be(record(M:Name), Value)) :- !,
+	prolog_load_context(module, C),
+	strip_module(C:Spec, M, Name).
+type_check(Atom, Value, Check) :-
+	atom(Atom), !,
+	type_check(record(Atom), Value, Check).
+
 
 %%	set_field_predicates(+Names, +Idx0, +Arity, +Types, +Constructor)// is det.
 %
@@ -292,7 +421,8 @@ set_field_predicates([Name|NT], I, Arity, [Type|TT], Constructor) -->
 	  SetFieldHead =.. [FieldPredName, NameTerm, Old, New],
 	  (   Type == any
 	  ->  SetField = SetFieldHead
-	  ;   SetField = (SetFieldHead :- must_be(Type, Value))
+	  ;   type_check(Type, Value, MustBe),
+	      SetField = (SetFieldHead :- MustBe)
 	  ),
 	  I2 is I + 1
 	},
@@ -339,9 +469,9 @@ types([Name|T0], [Name|TN], [any|TT]) :-
 		 *******************************/
 
 :- multifile
-	user:term_expansion/2.
+	system:term_expansion/2.
 :- dynamic
-	user:term_expansion/2.
+	system:term_expansion/2.
 
-user:term_expansion((:- record(Record)), Clauses) :-
+system:term_expansion((:- record(Record)), Clauses) :-
 	compile_records(Record, Clauses).
