@@ -3,9 +3,10 @@
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        wielemak@science.uva.nl
+    E-mail:        J.Wielemaker@cs.vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2007, University of Amsterdam
+    Copyright (C): 1985-2011, University of Amsterdam
+			      VU University Amsterdam
 
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
@@ -19,7 +20,7 @@
 
     You should have received a copy of the GNU General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 
     As a special exception, if you link this library with other files,
     compiled with a Free Software compiler, to produce an executable, this
@@ -42,6 +43,7 @@
 	    current_setting/1,		% Module:Name
 	    setting_property/2,		% ?Setting, ?Property
 	    list_settings/0,
+	    list_settings/1,		% +Module
 
 	    convert_setting_text/3	% +Type, +Text, -Value
 	  ]).
@@ -49,6 +51,7 @@
 :- use_module(library(broadcast)).
 :- use_module(library(debug)).
 :- use_module(library(option)).
+:- use_module(library(arithmetic)).
 
 /** <module> Setting management
 
@@ -57,7 +60,7 @@ applications. Applications define settings  in   one  or  multiple files
 using the directive setting/4 as illustrated below:
 
 ==
-:- use_module(library(setting)).
+:- use_module(library(settings)).
 
 :- setting(version, atom,   '1.0', 'Current version').
 :- setting(timeout, number,    20, 'Timeout in seconds').
@@ -94,14 +97,16 @@ access, loading and saving of settings.
 	current_setting(:),
 	restore_setting(:).
 
-curr_setting(Name, Module, Type, Default, Comment) :-
-	current_setting(Name, Module, Type, Default0, Comment, _Src),
+:- predicate_options(load_settings/2, 2, [undefined(oneof([load,error]))]).
+
+curr_setting(Name, Module, Type, Default, Comment, Src) :-
+	current_setting(Name, Module, Type, Default0, Comment, Src),
 	(   st_default(Name, Module, Default1)
 	->  Default = Default1
 	;   Default = Default0
 	).
 
-%%	setting(Name, Type, Default, Comment) is det.
+%%	setting(:Name, +Type, +Default, +Comment) is det.
 %
 %	Define a setting. Name denotes the name of the setting, Type its
 %	type. Default is the value before  it is modified. Default refer
@@ -121,10 +126,11 @@ setting(Name, Type, Default, Comment) :-
 		    _)).
 
 :- multifile
-	user:term_expansion/2.
+	system:term_expansion/2.
 
-user:term_expansion((:- setting(QName, Type, Default, Comment)),
+system:term_expansion((:- setting(QName, Type, Default, Comment)),
 		    Expanded) :-
+	\+ current_prolog_flag(xref, true),
 	prolog_load_context(module, M0),
 	strip_module(M0:QName, Module, Name),
 	must_be(atom, Name),
@@ -157,7 +163,7 @@ setting(QName, Value) :-
 	(   ground(Name)
 	->  (   st_value(Name, Module, Value0)
 	    ->  Value = Value0
-	    ;   curr_setting(Name, Module, Type, Default, _)
+	    ;   curr_setting(Name, Module, Type, Default, _, _)
 	    ->	eval_default(Default, Module, Type, Value)
 	    ;	existence_error(setting, Module:Name)
 	    )
@@ -234,7 +240,7 @@ eval_default(setting(Name), Module, Type, Value) :- !,
 	must_be(Type, Value).
 eval_default(Expr, _, Type, Value) :-
 	numeric_type(Type, Basic), !,
-	Val0 is Expr,
+	arithmetic_expression_value(Expr, Val0),
 	(   Basic == float
 	->  Val is float(Val0)
 	;   Basic = integer
@@ -338,18 +344,21 @@ numeric_type(between(L,_), Type) :-
 set_setting(QName, Value) :-
 	strip_module(QName, Module, Name),
 	must_be(atom, Name),
-	(   curr_setting(Name, Module, Type, Default0, _Comment),
+	(   curr_setting(Name, Module, Type, Default0, _Comment, _Src),
 	    eval_default(Default0, Module, Type, Default)
-	->  (   Value == Default
+	->  setting(Module:Name, Old),
+	    (   Value == Default
 	    ->	retract_setting(Module:Name)
 	    ;	st_value(Name, Module, Value)
 	    ->	true
 	    ;	check_type(Type, Value)
-	    ->	setting(Module:Name, Old),
-	        retract_setting(Module:Name),
-	        assert_setting(Module:Name, Value),
-		broadcast(settings(changed(Module:Name, Old, Value))),
-		clear_setting_cache	% might influence dependent settings.
+	    ->	retract_setting(Module:Name),
+	        assert_setting(Module:Name, Value)
+	    ),
+	    (	Old == Value
+	    ->	true
+	    ;	broadcast(settings(changed(Module:Name, Old, Value))),
+		clear_setting_cache	% might influence dependent settings
 	    )
 	;   existence_error(setting, Name)
 	).
@@ -461,7 +470,7 @@ load_settings(Setting, In, Options) :-
 
 read_setting(In, Term) :-
 	read_term(In, Term,
-		  [ errors(dec10)
+		  [ syntax_errors(dec10)
 		  ]).
 
 %%	store_setting(Term, +Options)
@@ -469,7 +478,7 @@ read_setting(In, Term) :-
 %	Store setting loaded from file in the Prolog database.
 
 store_setting(setting(Module:Name, Value), _) :-
-	curr_setting(Name, Module, Type, Default0, _Commentm), !,
+	curr_setting(Name, Module, Type, Default0, _Commentm, _Src), !,
 	eval_default(Default0, Module, Type, Default),
 	(   Value == Default
 	->  true
@@ -518,13 +527,13 @@ write_setting_header(Out) :-
 	format(Out, '*/~n~n', []).
 
 save_setting(Out, Module:Name) :-
-	curr_setting(Name, Module, Type, Default, Comment),
+	curr_setting(Name, Module, Type, Default, Comment, _Src),
 	(   st_value(Name, Module, Value),
 	    \+ ( eval_default(Default, Module, Type, DefValue),
 		 debug(setting, '~w <-> ~w~n', [DefValue, Value]),
 	         DefValue =@= Value
 	       )
-	->  format(Out, '~n%	~w~n', [Comment]),
+	->  format(Out, '~n%\t~w~n', [Comment]),
 	    format(Out, 'setting(~q:~q, ~q).~n', [Module, Name, Value])
 	;   true
 	).
@@ -551,42 +560,60 @@ current_setting(Module:Name) :-
 %		* default(-Default)
 %		Default value.  If this is an expression, it is
 %		evaluated.
+%		* source(-File:-Line)
+%		Location where the setting is defined.
 
 setting_property(Setting, Property) :-
 	ground(Setting), !,
 	Setting = Module:Name,
-	curr_setting(Name, Module, Type, Default, Comment), !,
-	setting_property(Property, Module, Type, Default, Comment).
+	curr_setting(Name, Module, Type, Default, Comment, Src), !,
+	setting_property(Property, Module, Type, Default, Comment, Src).
 setting_property(Setting, Property) :-
 	Setting = Module:Name,
-	curr_setting(Name, Module, Type, Default, Comment),
-	setting_property(Property, Module, Type, Default, Comment).
+	curr_setting(Name, Module, Type, Default, Comment, Src),
+	setting_property(Property, Module, Type, Default, Comment, Src).
 
-setting_property(type(Type),       _, Type, _,        _).
-setting_property(default(Default), M, Type, Default0, _) :-
+setting_property(type(Type), _, Type, _, _, _).
+setting_property(default(Default), M, Type, Default0, _, _) :-
 	eval_default(Default0, M, Type, Default).
-setting_property(comment(Comment), _, _,    _,        Comment).
+setting_property(comment(Comment), _, _, _, Comment, _).
+setting_property(source(Src), _, _, _, _, Src).
 
-%%	list_settings
+%%	list_settings is det.
+%%	list_settings(+Module) is det.
 %
-%	List settings to =current_output=.
+%	List settings to =current_output=. The   second  form only lists
+%	settings on the matching module.
+%
+%	@tbd	Compute the required column widths
 
 list_settings :-
-	format('~`=t~72|~n'),
-	format('~w~t~20| ~w~w~t~40| ~w~n', ['Name', 'Value (*=modified)', '', 'Comment']),
-	format('~`=t~72|~n'),
-	forall(current_setting(Setting),
-	       list_setting(Setting)).
+	list_settings(_).
 
-list_setting(Module:Name) :-
-	curr_setting(Name, Module, Type, Default0, Comment),
+list_settings(Spec) :-
+	spec_term(Spec, Term),
+	TS1 = 25,
+	TS2 = 40,
+	format('~`=t~72|~n'),
+	format('~w~t~*| ~w~w~t~*| ~w~n',
+	       ['Name', TS1, 'Value (*=modified)', '', TS2, 'Comment']),
+	format('~`=t~72|~n'),
+	forall(current_setting(Term),
+	       list_setting(Term, TS1, TS2)).
+
+spec_term(M:S, M:S) :- !.
+spec_term(M, M:_).
+
+
+list_setting(Module:Name, TS1, TS2) :-
+	curr_setting(Name, Module, Type, Default0, Comment, _Src),
 	eval_default(Default0, Module, Type, Default),
 	setting(Module:Name, Value),
 	(   Value \== Default
 	->  Modified = (*)
 	;   Modified = ''
 	),
-        format('~w~t~20| ~q~w~t~40| ~w~n', [Module:Name, Value, Modified, Comment]).
+        format('~w~t~*| ~q~w~t~*| ~w~n', [Module:Name, TS1, Value, Modified, TS2, Comment]).
 
 
 		 /*******************************
