@@ -3,9 +3,9 @@
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        jan@swi.psy.uva.nl
+    E-mail:        J.Wielemaker@cs.vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2002, University of Amsterdam
+    Copyright (C): 1985-2009, University of Amsterdam
 
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 
     As a special exception, if you link this library with other files,
     compiled with a Free Software compiler, to produce an executable, this
@@ -35,14 +35,74 @@
 	    unload_foreign_library/1,	% +LibFile
 	    unload_foreign_library/2,	% +LibFile, +UninstallFunc
 	    current_foreign_library/2,	% ?LibFile, ?Public
-	    reload_foreign_libraries/0
+	    reload_foreign_libraries/0,
+					% Directives
+	    use_foreign_library/1,	% :LibFile
+	    use_foreign_library/2	% :LibFile, +InstallFunc
 	  ]).
 :- use_module(library(lists), [reverse/2]).
 :- set_prolog_flag(generate_debug_info, false).
 
-:- module_transparent
-	load_foreign_library/1,
-	load_foreign_library/2.
+/** <module> Utility library for loading foreign objects (DLLs, shared objects)
+
+This   section   discusses   the   functionality   of   the   (autoload)
+library(shlib), providing an interface to   manage  shared libraries. We
+describe the procedure for using a foreign  resource (DLL in Windows and
+shared object in Unix) called =mylib=.
+
+First, one must  assemble  the  resource   and  make  it  compatible  to
+SWI-Prolog. The details for this  vary   between  platforms. The plld(1)
+utility can be used to deal with this in a portable manner.  The typical
+commandline is:
+
+	==
+	plld -o mylib file.{c,o,cc,C} ...
+	==
+
+Make  sure  that  one  of   the    files   provides  a  global  function
+=|install_mylib()|=  that  initialises  the  module    using   calls  to
+PL_register_foreign(). Here is a  simple   example  file  mylib.c, which
+creates a Windows MessageBox:
+
+    ==
+    #include <windows.h>
+    #include <SWI-Prolog.h>
+
+    static foreign_t
+    pl_say_hello(term_t to)
+    { char *a;
+
+      if ( PL_get_atom_chars(to, &a) )
+      { MessageBox(NULL, a, "DLL test", MB_OK|MB_TASKMODAL);
+
+	PL_succeed;
+      }
+
+      PL_fail;
+    }
+
+    install_t
+    install_mylib()
+    { PL_register_foreign("say_hello", 1, pl_say_hello, 0);
+    }
+    ==
+
+Now write a file mylib.pl:
+
+    ==
+    :- module(mylib, [ say_hello/1 ]).
+    :- use_foreign_library(foreign(mylib)).
+    ==
+
+The file mylib.pl can be loaded as a normal Prolog file and provides the
+predicate defined in C.
+*/
+
+:- meta_predicate
+	load_foreign_library(:),
+	load_foreign_library(:, +),
+	use_foreign_library(:),
+	use_foreign_library(:, +).
 
 :- dynamic
 	loading/1,			% Lib
@@ -66,30 +126,44 @@
 		 *	     DISPATCHING	*
 		 *******************************/
 
-/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-Windows: If libpl.dll is compiled for debugging, prefer loading <lib>D.dll
-to allow for debugging.
-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+%%	find_library(+LibSpec, -Lib, -Delete) is det.
+%
+%	Find a foreign library from LibSpec.  If LibSpec is available as
+%	a resource, the content of the resource is copied to a temporary
+%	file and Delete is unified with =true=.
+%
+%	On Windows, if =|swipl.dll|= is   compiled for debugging, prefer
+%	loading <lib>D.dll to allow for debugging.
 
-find_library(Spec, Lib) :-
+find_library(Spec, Lib, Delete) :-
 	current_prolog_flag(windows, true),
 	current_prolog_flag(kernel_compile_mode, debug),
 	libd_spec(Spec, SpecD),
-	catch(find_library2(SpecD, Lib), _, fail).
-find_library(Spec, Lib) :-
-	find_library2(Spec, Lib).
+	catch(find_library2(SpecD, Lib, Delete), _, fail).
+find_library(Spec, Lib, Delete) :-
+	find_library2(Spec, Lib, Delete).
 
-find_library2(Spec, Lib) :-
-	absolute_file_name(Spec,
+find_library2(Spec, TmpFile, true) :-
+	'$rc_handle'(RC),
+	term_to_atom(Spec, Name),
+	setup_call_cleanup(
+	    '$rc_open'(RC, Name, shared, read, In),
+	    setup_call_cleanup(
+		tmp_file_stream(binary, TmpFile, Out),
+		copy_stream_data(In, Out),
+		close(Out)),
+	    close(In)), !.
+find_library2(Spec, Lib, false) :-
+	absolute_file_name(Spec, Lib,
 			   [ file_type(executable),
 			     access(read),
 			     file_errors(fail)
-			   ], Lib), !.
-find_library2(Spec, Spec) :-
+			   ]), !.
+find_library2(Spec, Spec, false) :-
 	atom(Spec), !.			% use machines finding schema
-find_library2(foreign(Spec), Spec) :-
+find_library2(foreign(Spec), Spec, false) :-
 	atom(Spec), !.			% use machines finding schema
-find_library2(Spec, _) :-
+find_library2(Spec, _, _) :-
 	throw(error(existence_error(source_sink, Spec), _)).
 
 libd_spec(Name, NameD) :-
@@ -123,32 +197,49 @@ entry(_, default(Function), Function).
 		 *	    (UN)LOADING		*
 		 *******************************/
 
-%%	load_foreign_library(+FileSpec) is det.
-%%	load_foreign_library(+FileSpec, +Entry:atom) is det.
+%%	load_foreign_library(:FileSpec) is det.
+%%	load_foreign_library(:FileSpec, +Entry:atom) is det.
 %
 %	Load a _|shared object|_  or  _DLL_.   After  loading  the Entry
 %	function is called without arguments. The default entry function
-%	is composed from =install_=, followed by the file base-name.
+%	is composed from =install_=,  followed   by  the file base-name.
+%	E.g.,    the    load-call    below      calls    the    function
+%	=|install_mylib()|=. If the platform   prefixes extern functions
+%	with =_=, this prefix is added before calling.
+%
+%	  ==
+%		...
+%		load_foreign_library(foreign(mylib)),
+%		...
+%	  ==
+%
+%	@param	FileSpec is a specification for absolute_file_name/3.  If searching
+%		the file fails, the plain name is passed to the OS to try the default
+%		method of the OS for locating foreign objects.  The default definition
+%		of file_search_path/2 searches <prolog home>/lib/<arch> on Unix and
+%		<prolog home>/bin on Windows.
+%
+%	@see	use_foreign_library/1,2 are intended for use in directives.
 
 load_foreign_library(Library) :-
 	load_foreign_library(Library, default(install)).
 
-load_foreign_library(LibFileSpec, Entry) :-
-	strip_module(LibFileSpec, Module, LibFile),
+load_foreign_library(Module:LibFile, Entry) :-
 	with_mutex('$foreign',
-		   shlib:load_foreign_library(LibFile, Module, Entry)).
+		   load_foreign_library(LibFile, Module, Entry)).
 
 load_foreign_library(LibFile, _Module, _) :-
 	current_library(LibFile, _, _, _, _), !.
 load_foreign_library(LibFile, Module, DefEntry) :-
 	retractall(error(_, _)),
-	find_library(LibFile, Path),
+	find_library(LibFile, Path, Delete),
 	asserta(loading(LibFile)),
 	catch(Module:open_shared_object(Path, Handle), E, true),
 	(   nonvar(E)
-	->  assert(error(Path, E)),
+	->  delete_foreign_lib(Delete, Path),
+	    assert(error(Path, E)),
 	    fail
-	;   true
+	;   delete_foreign_lib(Delete, Path)
 	), !,
 	(   (	entry(LibFile, DefEntry, Entry),
 		Module:call_shared_object_function(Handle, Entry)
@@ -170,6 +261,34 @@ load_foreign_library(LibFile, _, _) :-
 	;   throw(error(existence_error(foreign_library, LibFile), _))
 	).
 
+delete_foreign_lib(true, Path) :-
+	catch(delete_file(Path), _, true).
+delete_foreign_lib(_, _).
+
+
+%%	use_foreign_library(+FileSpec) is det.
+%%	use_foreign_library(+FileSpec, +Entry:atom) is det.
+%
+%	Load and install a foreign   library as load_foreign_library/1,2
+%	and register the installation using   initialization/2  with the
+%	option =now=. This is similar to using:
+%
+%	  ==
+%	  :- initialization(load_foreign_library(foreign(mylib))).
+%	  ==
+%
+%	but using the initialization/1 wrapper causes  the library to be
+%	loaded _after_ loading of  the  file   in  which  it  appears is
+%	completed,  while  use_foreign_library/1  loads    the   library
+%	_immediately_. I.e. the  difference  is   only  relevant  if the
+%	remainder of the file uses functionality of the C-library.
+
+use_foreign_library(FileSpec) :-
+	initialization(load_foreign_library(FileSpec), now).
+
+use_foreign_library(FileSpec, Entry) :-
+	initialization(load_foreign_library(FileSpec, Entry), now).
+
 %%	unload_foreign_library(+FileSpec) is det.
 %%	unload_foreign_library(+FileSpec, +Exit:atom) is det.
 %
