@@ -17,7 +17,7 @@
 
     You should have received a copy of the GNU General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 
     As a special exception, if you link this library with other files,
     compiled with a Free Software compiler, to produce an executable, this
@@ -78,6 +78,10 @@ otherwise during startup of the application.
 
 :- meta_predicate
 	thread_create_in_pool(+, 0, -, +).
+:- predicate_options(thread_create_in_pool/4, 4,
+                     [ wait(boolean),
+                       pass_to(system:thread_create/3, 3)
+                     ]).
 
 
 %%	thread_pool_create(+Pool, +Size, +Options) is det.
@@ -105,6 +109,7 @@ otherwise during startup of the application.
 %		internal use by this library.
 
 thread_pool_create(Name, Size, Options) :-
+	must_be(list, Options),
 	pool_manager(Manager),
 	thread_self(Me),
 	thread_send_message(Manager, create_pool(Name, Size, Options, Me)),
@@ -180,13 +185,15 @@ thread_pool_property(Name, Property) :-
 %
 %	@error	resource_error(threads_in_pool(Pool)) is raised if wait
 %		is =false= or the backlog limit has been reached.
+%	@error	existence_error(thread_pool, Pool) if Pool does not
+%		exist.
 
 thread_create_in_pool(Pool, Goal, Id, Options) :-
-	strip_module(Goal, M, G),
 	select_option(wait(Wait), Options, ThreadOptions, true),
 	pool_manager(Manager),
 	thread_self(Me),
-	thread_send_message(Manager, create(Pool, M:G, Me, Wait, ThreadOptions)),
+	thread_send_message(Manager,
+			    create(Pool, Goal, Me, Wait, ThreadOptions)),
 	wait_reply(Id).
 
 
@@ -199,25 +206,34 @@ thread_create_in_pool(Pool, Goal, Id, Options) :-
 %	ThreadID is the thread (alias) identifier of the manager. Starts
 %	the manager if it is not running.
 
-pool_manager('__thread_pool_manager') :-
-	catch(thread_property('__thread_pool_manager', status(Status)),
-	      _, fail),
-	(   Status == running
-	->  !
-	;   thread_join('__thread_pool_manager', Status),
-	    print_message(warning, thread_pool(manager_died(Status))),
-	    fail
+pool_manager(TID) :-
+	TID = '__thread_pool_manager',
+	(   thread_running(TID)
+	->  true
+	;   with_mutex('__thread_pool', create_pool_manager(TID))
 	).
-pool_manager('__thread_pool_manager') :-
-	with_mutex('__thread_pool', create_pool_manager).
 
-create_pool_manager :-
+thread_running(Thread) :-
+	catch(thread_property(Thread, status(Status)),
+	      E, true),
+	(   var(E)
+	->  (   Status == running
+	    ->  true
+	    ;	thread_join(Thread, _),
+		print_message(warning, thread_pool(manager_died(Status))),
+		fail
+	    )
+	;   E = error(existence_error(thread, Thread), _)
+	->  fail
+	;   throw(E)
+	).
+
+create_pool_manager(Thread) :-
+	thread_running(Thread), !.
+create_pool_manager(Thread) :-
 	rb_new(State0),
 	thread_create(manage_thread_pool(State0), _,
-		      [ alias('__thread_pool_manager'),
-			local(1000),
-			global(2000),
-			trail(2000)
+		      [ alias(Thread)
 		      ]).
 
 
