@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
 #define O_DEBUG 1
@@ -112,6 +112,15 @@ leave the details to this function.
 #undef HAVE_H_ERRNO
 #endif
 
+#if defined(__MINGW32__)
+#define __SEH_NOOP 1
+#endif
+
+#if defined(__MINGW32__)
+#define WINVER 0x0501
+#include <ws2tcpip.h>
+#endif
+
 #include "nonblockio.h"
 
 #include <SWI-Stream.h>
@@ -266,6 +275,7 @@ static int socketIsPendingClose(plsocket *s);
 static const char *WinSockError(unsigned long eno);
 #endif
 
+#ifndef __WINDOWS__
 static int
 need_retry(int error)
 { if ( error == EINTR || error == EAGAIN || error == EWOULDBLOCK )
@@ -273,6 +283,7 @@ need_retry(int error)
 
   return FALSE;
 }
+#endif
 
 #ifdef O_DEBUG
 static int debugging;
@@ -364,8 +375,12 @@ nbio_fcntl(nbio_sock_t socket, int op, int arg)
       switch(arg)
       { case O_NONBLOCK:
 	{ int rval;
+#if defined(__MINGW32__)
+	  u_long non_block;
+#else
+          /* FIXME: is this really `int' for MSC? */
 	  int non_block;
-
+#endif
 	  non_block = 1;
 	  rval = ioctlsocket(s->socket, FIONBIO, &non_block);
 	  if ( rval )
@@ -444,6 +459,8 @@ waitRequest(plsocket *s)
     } else if ( GetMessage(&msg, NULL, 0, 0) )
     { TranslateMessage(&msg);
       DispatchMessage(&msg);
+      if ( PL_exception(0) )
+	return FALSE;			/* DispatchMessage handled a signal */
     } else
     { ExitThread(0);			/* WM_QUIT received */
       return FALSE;			/* NOTREACHED */
@@ -934,6 +951,9 @@ socket_wnd_proc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
 	switch(s->request)
 	{ case REQ_CONNECT:
 	    break;
+          case REQ_NONE:
+            /* FIXME: is this OK? */
+            break;
 	  case REQ_ACCEPT:
 	    s->rdata.accept.slave = SOCKET_ERROR;
 	    break;
@@ -996,17 +1016,6 @@ HiddenFrameClass()
 }
 
 
-static void
-destroyHiddenWindow(int rval, void *closure)
-{ local_state *s = State();
-
-  if ( s->hwnd )
-  { DestroyWindow(s->hwnd);
-    s->hwnd = 0;
-  }
-}
-
-
 static HWND
 SocketHiddenWindow()
 { local_state *s = State();
@@ -1341,6 +1350,7 @@ allocSocket(SOCKET socket)
 }
 
 
+#ifndef __WINDOWS__
 static int
 freeSocket(plsocket *s)
 { int rval;
@@ -1380,6 +1390,7 @@ freeSocket(plsocket *s)
 
   return rval;
 }
+#endif
 
 
 		 /*******************************
@@ -2006,7 +2017,8 @@ nbio_connect(nbio_sock_t socket,
     { s->rdata.connect.addrlen = addrlen;
       memcpy(&s->rdata.connect.addr, serv_addr, addrlen);
       placeRequest(s, REQ_CONNECT);
-      waitRequest(s);
+      if ( !waitRequest(s) )
+	return -1;
     }
 
     if ( s->error )
