@@ -3,9 +3,10 @@
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        wielemak@science.uva.nl
+    E-mail:        J.Wielemaker@cs.vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2007, University of Amsterdam
+    Copyright (C): 1985-2011, University of Amsterdam,
+			      VU University Amsterdam
 
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
@@ -19,7 +20,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 
     As a special exception, if you link this library with other files,
     compiled with a Free Software compiler, to produce an executable, this
@@ -32,11 +33,14 @@
 
 :- module(http_client,
 	  [ http_get/3,			% +URL, -Reply, +Options
+	    http_delete/3,		% +URL, -Reply, +Options
 	    http_post/4,		% +URL, +In, -Reply, +Options
+	    http_put/4,			% +URL, +In, -Reply, +Options
 	    http_read_data/3,		% +Header, -Data, +Options
 	    http_disconnect/1		% +What
 	  ]).
 :- use_module(library(socket)).
+:- use_module(library(uri)).
 :- use_module(library(url)).
 :- use_module(http_header).
 :- use_module(http_stream).
@@ -44,14 +48,65 @@
 :- use_module(library(memfile)).
 :- use_module(library(lists)).
 :- use_module(library(error)).
+:- use_module(library(option)).
 :- use_module(dcg_basics).
 
 :- multifile
 	http_convert_data/4,		% http_read_data plugin-hook
-	post_data_hook/3.		% http_post_data/3 hook
+	post_data_hook/3,		% http_post_data/3 hook
+	open_connection/4,		% do_connect/5 hook
+	close_connection/4.
+
+:- predicate_options(http_get/3, 3,
+		     [ connection(oneof([close,'Keep-alive'])),
+		       http_version(any),
+%		       proxy(atom,integer),
+		       proxy_authorization(any),
+		       timeout(number),
+		       user_agent(atom),
+		       range(any),
+		       request_header(any),
+		       reply_header(-list),
+		       method(oneof(['GET','DELETE'])),
+		       pass_to(http_read_data/3, 3)
+		     ]).
+:- predicate_options(http_delete/3, 3, [pass_to(http_get/3, 3)]).
+:- predicate_options(http_read_data/3, 3,
+		     [ to(any),
+		       content_type(any)
+		     ]).
+
+
+/** <module> HTTP client library
+
+This library provides the four basic   HTTP client actions: GET, DELETE,
+POST and PUT. In addition, it   provides http_read_data/3, which is used
+by  library(http/http_parameters)  to  decode  POST    data   in  server
+applications.
+
+@see	library(http/http_open) provides a stream-based alternative to
+	http_get/3.
+@tbd	Replace functions from library(url) with those from library(uri).
+*/
+
+
+%%	open_connection(+Scheme, +Address, -In, -Out) is semidet.
+%
+%	Hook to open a connection for the  given URL-scheme to the given
+%	address. If successful, In and  Out   must  be  two valid Prolog
+%	streams that connect to the server.
+%
+%	@param Scheme is the URL schema (=http= or =https=)
+%	@param Address is a term Host:Port as used by tcp_connect/4.
+
+%%	close_connection(+Scheme, +Address, +In, +Out) is semidet.
+%
+%	Hook to close a specific connection.   If the hook succeeds, the
+%	HTTP client assumes that In and Out are no longer to be used. If
+%	the hook fails, the client calls close/2 on both streams.
 
 :- dynamic
-	connection/4.			% Host:Port, ThreadId, In, Out
+	connection/5.			% Host:Port, Protocol, Thread, In, Out
 
 user_agent('SWI-Prolog (http://www.swi-prolog.org)').
 
@@ -63,51 +118,57 @@ user_agent('SWI-Prolog (http://www.swi-prolog.org)').
 connect(Parts, Read, Write, _) :-
 	memberchk(socket(Read, Write), Parts), !.
 connect(Parts, Read, Write, Options) :-
-	address(Parts, Address, Options),
+	address(Parts, Address, Protocol, Options),
 	with_mutex(http_client_connect,
-		   connect2(Address, Read, Write, Options)).
+		   connect2(Address, Protocol, Read, Write, Options)).
 
-connect2(Address, In, Out, _) :-
+connect2(Address, Protocol, In, Out, _) :-
 	thread_self(Self),
-	connection(Address, Self, In, Out), !.
-connect2(Address, In, Out, Options) :-
+	connection(Address, Protocol, Self, In, Out), !.
+connect2(Address, Protocol, In, Out, Options) :-
 	thread_self(Self),
-	do_connect(Address, In, Out, Options),
-	assert(connection(Address, Self, In, Out)).
+	do_connect(Address, Protocol, In, Out, Options),
+	assert(connection(Address, Protocol, Self, In, Out)).
 
-do_connect(Address, In, Out, Options) :-
+do_connect(Address, Protocol, In, Out, Options) :-
 	debug(http(client), 'http_client: Connecting to ~p ...', [Address]),
-	tcp_socket(Socket),
-	catch(tcp_connect(Socket, Address),
-	      E,
-	      (	  tcp_close_socket(Socket),
-		  throw(E)
-	      )),
-	tcp_open_socket(Socket, In, Out),
+	(   open_connection(Protocol, Address, In, Out)
+	->  true
+	;   tcp_socket(Socket),
+	    catch(tcp_connect(Socket, Address, In, Out),
+		  E,
+		  (   tcp_close_socket(Socket),
+		      throw(E)
+		  ))
+	),
 	debug(http(client), '\tok ~p --> ~p', [In, Out]),
 	(   memberchk(timeout(Timeout), Options)
         ->  set_stream(In, timeout(Timeout))
         ;   true
 	), !.
-do_connect(Address, _, _, _) :-		% can this happen!?
+do_connect(Address, _, _, _, _) :-		% can this happen!?
 	throw(error(failed(connect, Address), _)).
 
 
 disconnect(Parts) :-
-	address(Parts, Address, []), !,
-	with_mutex(http_client_connect, disconnect2(Address)).
-disconnect(Address) :-
-	with_mutex(http_client_connect, disconnect2(Address)).
+	protocol(Parts, Protocol),
+	address(Parts, Protocol, Address, []), !,
+	disconnect(Address, Protocol).
 
+disconnect(Address, Protocol) :-
+	with_mutex(http_client_connect,
+		   disconnect_locked(Address, Protocol)).
 
-disconnect2(Address) :-
+disconnect_locked(Address, Protocol) :-
 	thread_self(Me),
 	debug(connection, '~w: Closing connection to ~w~n', [Me, Address]),
 	thread_self(Self),
-	retract(connection(Address, Self, In, Out)), !,
-	close_socket(In, Out).
+	retract(connection(Address, Protocol, Self, In, Out)), !,
+	disconnect(Protocol, Address, In, Out).
 
-close_socket(In, Out) :-
+disconnect(Protocol, Address, In, Out) :-
+	close_connection(Protocol, Address, In, Out), !.
+disconnect(_, _, In, Out) :-
 	close(Out, [force(true)]),
 	close(In,  [force(true)]).
 
@@ -118,27 +179,46 @@ close_socket(In, Out) :-
 
 http_disconnect(all) :-
 	(   thread_self(Self),
-	    connection(Address, Self, _, _),
-	    disconnect(Address),
+	    connection(Address, Protocol, Self, _, _),
+	    disconnect(Address, Protocol),
 	    fail
 	;   true
 	).
 
-address(_Parts, Host:Port, Options) :-
-	memberchk(proxy(Host, Port), Options), !.
-address(Parts, Host:Port, _Options) :-
+address(_Parts, Host:Port, Protocol, Options) :-
+	(   memberchk(proxy(Host, Port, Protocol), Options)
+	->  true
+	;   memberchk(proxy(Host, Port), Options),
+	    Protocol = http
+	).
+address(Parts, Host:Port, Protocol, _Options) :-
 	memberchk(host(Host), Parts),
-	port(Parts, Port).
+	port(Parts, Port),
+	protocol(Parts, Protocol).
 
 port(Parts, Port) :-
 	memberchk(port(Port), Parts), !.
 port(Parts, 80) :-
 	memberchk(protocol(http), Parts).
 
+protocol(Parts, Protocol) :-
+	memberchk(protocol(Protocol), Parts), !.
+protocol(_, http).
+
 		 /*******************************
 		 *	        GET		*
 		 *******************************/
 
+%%	http_delete(+URL, -Data, +Options) is det.
+%
+%	Execute a DELETE method on the server.
+%
+%	@tbd Properly map the 201, 202 and 204 replies.
+
+http_delete(URL, Data, Options) :-
+	http_get(URL, Data, [method('DELETE')|Options]).
+
+
 %%	http_get(+URL, -Data, +Options) is det.
 %
 %	Get data from an HTTP server.
@@ -148,7 +228,8 @@ http_get(URL, Data, Options) :-
 	parse_url(URL, Parts),
         http_get(Parts, Data, Options).
 http_get(Parts, Data, Options) :-
-	memberchk(connection(Connection), Options),
+	must_be(list, Options),
+	option(connection(Connection), Options),
 	downcase_atom(Connection, 'keep-alive'), !,
 	between(0, 1, _),
 	catch(http_do_get(Parts, Data, Options), E,
@@ -158,10 +239,10 @@ http_get(Parts, Data, Options) :-
 		  fail
 	      )), !.
 http_get(Parts, Data, Options) :-
-	address(Parts, Address, Options),
-	do_connect(Address, Read, Write, Options),
+	address(Parts, Address, Protocol, Options),
+	do_connect(Address, Protocol, Read, Write, Options),
 	call_cleanup(http_do_get([socket(Read, Write)|Parts], Data, Options),
-		     close_socket(Read, Write)).
+		     disconnect(Protocol, Address, Read, Write)).
 
 http_do_get(Parts, Data, Options) :-
 	connect(Parts, Read, Write, Options),
@@ -171,7 +252,8 @@ http_do_get(Parts, Data, Options) :-
 	    Options1 = Options
 	),
 	memberchk(host(Host), Parts),
-	http_write_header(Write, 'GET', Location, Host,
+	option(method(Method), Options, 'GET'),
+	http_write_header(Write, Method, Location, Host,
 			  Options1, ReplyOptions),
 	write(Write, '\r\n'),
 	flush_output(Write),
@@ -206,8 +288,8 @@ http_read_reply(In, Data, Options) :-
 	    downcase_atom(Connection, 'keep-alive')
 	->  true
 	;   thread_self(Self),
-	    connection(Address, Self, In, _Out)
-	->  disconnect(Address)
+	    connection(Address, Protocol, Self, In, _Out)
+	->  disconnect(Address, Protocol)
 	;   true
 	).
 http_read_reply(In, _Data, _Options) :-
@@ -249,7 +331,6 @@ http_write_header(Out, Method, Location, Host, Options, RestOptions) :-
 		     Connection: ~w\r\n', [Agent, Connection]),
 	x_headers(Options3, Out, RestOptions).
 
-
 %%	x_headers(+Options, +Out, -RestOptions) is det.
 %
 %	Pass additional request options.  For example:
@@ -259,13 +340,34 @@ http_write_header(Out, Method, Location, Host, Options, RestOptions) :-
 %	No checking is performed on the fieldname or value. Both are
 %	copied literally and in the order of appearance to the request.
 
-x_headers(Options0, Out, Options) :-
-	select(request_header(Name=Value), Options0, Options1), !,
-	format(Out, '~w: ~w\r\n', [Name, Value]),
-	x_headers(Options1, Out, Options).
-x_headers(Options, _, Options).
+x_headers([], _, []).
+x_headers([H|T0], Out, Options) :-
+	x_header(H, Out), !,
+	x_headers(T0, Out, Options).
+x_headers([H|T0], Out, [H|T]) :-
+	x_headers(T0, Out, T).
 
-%%	http_read_data(+Fields, -Data, +Options) is det.
+x_header(request_header(Name=Value), Out) :-
+	format(Out, '~w: ~w\r\n', [Name, Value]).
+x_header(proxy_authorization(ProxyAuthorization), Out) :-
+	proxy_auth_header(ProxyAuthorization, Out).
+x_header(range(Spec), Out) :-
+	Spec =.. [Unit, From, To],
+	(   To == end
+	->  ToT = ''
+	;   must_be(integer, To),
+	    ToT = To
+	),
+	format(Out, 'Range: ~w=~d-~w\r\n', [Unit, From, ToT]).
+
+proxy_auth_header(basic(User, Password), Out) :- !,
+	format(codes(Codes), '~w:~w', [User, Password]),
+	phrase(base64(Codes), Base64Codes),
+	format(Out, 'Proxy-Authorization: basic ~s\r\n', [Base64Codes]).
+proxy_auth_header(Auth, _) :-
+	domain_error(authorization, Auth).
+
+%%	http_read_data(+Request, -Data, +Options) is det.
 %
 %	Read data from an HTTP connection.   Options must contain a term
 %	input(In) that provides the input stream   from the HTTP server.
@@ -292,7 +394,7 @@ http_read_data(In, Fields, Data, Options) :-	% Transfer-encoding: chunked
 	call_cleanup(http_read_data(DataStream, RestFields, Data, Options),
 		     close(DataStream)).
 http_read_data(In, Fields, Data, Options) :-
-	memberchk(to(X), Options), !,
+	option(to(X), Options), !,
 	(   X = stream(Stream)
 	->  (   memberchk(content_length(Bytes), Fields)
 	    ->  copy_stream_data(In, Stream, Bytes)
@@ -316,11 +418,12 @@ http_read_data(In, Fields, Data, Options) :-
 	    Data = Data0
 	).
 http_read_data(In, Fields, Data, _) :-
-	memberchk(content_type('application/x-www-form-urlencoded'), Fields), !,
+	option(content_type(ContentType), Fields),
+	form_data_content_type(ContentType), !,
 	http_read_data(In, Fields, Codes, [to(codes)]),
-	parse_url_search(Codes, Data).
-http_read_data(In, Fields, Data, Options) :- 			% call hook
-	(   select(content_type(Type), Options, Options1)
+	uri_query_components(Codes, Data).
+http_read_data(In, Fields, Data, Options) :-			% call hook
+	(   select_option(content_type(Type), Options, Options1)
 	->  delete(Fields, content_type(_), Fields1),
 	    http_convert_data(In, [content_type(Type)|Fields1], Data, Options1)
 	;   http_convert_data(In, Fields, Data, Options)
@@ -337,20 +440,38 @@ encoding(Fields, utf8) :-
 	), !.
 encoding(_, octet).
 
+%%	form_data_content_type(+ContentType) is semidet.
+%
+%	True if ContentType   is  =|application/x-www-form-urlencoded|=.
+%	Note that there can be parameters,   but  the ; MUST immediately
+%	follow the type.
+%
+%	@tbd	Within ClioPatria, there is a library http_request_value.pl
+%		that generalises these things. That library must be
+%		completed and used for this match.
+
+form_data_content_type('application/x-www-form-urlencoded') :- !.
+form_data_content_type(ContentType) :-
+	sub_atom(ContentType, 0, _, _, 'application/x-www-form-urlencoded;').
+
 
 		 /*******************************
 		 *	       POST		*
 		 *******************************/
 
-%%	http_post(+URL, +In, -Out, +Options)
+%%	http_put(+URL, +In, -Out, +Options)
 %
-%	Issue an HTTP POST request, In is modelled after the reply
-%	from the HTTP server module.  In is one of:
+%	Issue an HTTP PUT request.
+
+http_put(URL, In, Out, Options) :-
+	http_post(URL, In, Out, [method('PUT')|Options]).
+
+
+%%	http_post(+URL, +Data, -Reply, +Options) is det.
 %
-%		* string(String)
-%		* string(MimeType, String)
-%		* html(Tokens)
-%		* file(MimeType, File)
+%	Issue   an   HTTP   POST   request.   Data   is   posted   using
+%	http_post_data/3. The HTTP server reply   is  returned in Reply,
+%	using the same rules as http_get/3.
 
 http_post(URL, In, Out, Options) :-
 	atomic(URL), !,
@@ -365,11 +486,11 @@ http_post(Parts, In, Out, Options) :-
 		  fail
 	      )), !.
 http_post(Parts, In, Out, Options) :-
-	address(Parts, Address, Options),
-	do_connect(Address, Read, Write, Options),
+	address(Parts, Address, Protocol, Options),
+	do_connect(Address, Protocol, Read, Write, Options),
 	call_cleanup(http_do_post([socket(Read, Write)|Parts],
 				  In, Out, Options),
-		     close_socket(Read, Write)).
+		     disconnect(Protocol, Address, Read, Write)).
 
 http_do_post(Parts, In, Out, Options) :-
 	connect(Parts, Read, Write, Options),
@@ -383,11 +504,13 @@ http_do_post(Parts, In, Out, Options) :-
 	http_read_reply(Read, Out, ReplyOptions).
 
 write_post_header(Out, Location, Host, In, Options) :-
-	http_write_header(Out, 'POST', Location, Host, Options, DataOptions),
+	option(method(Method), Options, 'POST'),
+	http_write_header(Out, Method, Location, Host, Options, DataOptions),
 	http_post_data(In, Out, DataOptions),
 	flush_output(Out).
 
 post_option(connection(_)).
+post_option(method(_)).
 post_option(http_version(_)).
 post_option(cache_control(_)).
 post_option(request_header(_)).
