@@ -3,9 +3,10 @@
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        J.Wielemaker@uva.nl
+    E-mail:        J.Wielemaker@cs.vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 2009, University of Amsterdam
+    Copyright (C): 2009-2010, University of Amsterdam
+			      VU University Amsterdam
 
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
@@ -19,7 +20,7 @@
 
     You should have received a copy of the GNU General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 
     As a special exception, if you link this library with other files,
     compiled with a Free Software compiler, to produce an executable, this
@@ -33,7 +34,7 @@
 	  [ doc_for_file/2,		% +FileSpec, +Options
 	    doc_write_html/3,		% +Stream, +Title, +Term
 	    doc_for_wiki_file/2,	% +FileSpec, +Options
-	    				% Support doc_index
+					% Support doc_index
 	    doc_page_dom/3,		% +Title, +Body, -DOM
 	    print_html_head/1,		% +Stream
 	    predref//1,			% +PI //
@@ -42,15 +43,17 @@
 	    doc_hide_private/3,		% +Doc0, -Doc, +Options
 	    edit_button//2,		% +File, +Options, //
 	    source_button//2,		% +File, +Options, //
-	    pred_edit_button//2,		% +PredInd, +Options, //
+	    pred_edit_button//2,	% +PredInd, +Options, //
 	    object_edit_button//2,	% +Obj, +Options, //
 	    object_source_button//2,	% +Obj, +Options, //
 					% Support other backends
 	    doc_file_objects/5,		% +FSpec, -File, -Objs, -FileOpts, +Opts
 	    existing_linked_file/2,	% +FileSpec, -Path
+	    unquote_filespec/2,		% +FileSpec, -Unquoted
 	    doc_tag_title/2,		% +Tag, -Title
 	    pred_anchor_name/3,		% +Head, -PI, -Anchor
 	    private/2,			% +Obj, +Options
+	    (multifile)/2,		% +Obj, +Options
 	    is_pi/1,			% @Term
 	    is_op_type/2,		% +Atom, ?Type
 					% Output routines
@@ -63,19 +66,22 @@
 	    objects//2,			% +Objects, +Options, //
 	    object_ref//2,		% +Object, +Options, //
 	    object_href/2,		% +Object, -URL
-	    object_page//2		% +Object, +Options, //
+	    object_page//2,		% +Object, +Options, //
+	    object_synopsis//2		% +Object, +Options, //
 	  ]).
 :- use_module(library(lists)).
 :- use_module(library(option)).
-:- use_module(library(url)).
+:- use_module(library(uri)).
 :- use_module(library(readutil)).
 :- use_module(library(http/html_write)).
 :- use_module(library(http/http_dispatch)).
+:- use_module(library(http/http_wrapper)).
 :- use_module(library(http/http_path)).
 :- use_module(library(http/html_head)).
 :- use_module(library(doc_http)).
 :- use_module(library(debug)).
 :- use_module(library(apply)).
+:- use_module(library(pairs)).
 :- use_module(doc_process).
 :- use_module(doc_man).
 :- use_module(doc_modes).
@@ -93,6 +99,83 @@ extracting module doc_wiki.pl into HTML+CSS.
 	in other backends.
 */
 
+:- public
+	params//1,			% Called from \Term output created
+	pred_dt//3,			% by the wiki renderer
+	section//2,
+	tag//2.
+
+
+:- predicate_options(doc_for_wiki_file/2, 2,
+		     [ edit(boolean)
+		     ]).
+:- predicate_options(doc_hide_private/3, 3,
+		     [module(atom), public(list), public_only(boolean)]).
+:- predicate_options(edit_button//2, 2,
+		     [ edit(boolean)
+		     ]).
+:- predicate_options(file//2, 2,
+		     [ label(any),
+		       absolute_path(atom),
+		       href(atom),
+		       map_extension(list),
+		       files(list),
+		       edit_handler(atom)
+		     ]).
+:- predicate_options(file_header//2, 2,
+		     [ edit(boolean),
+		       files(list),
+		       public_only(boolean)
+		     ]).
+:- predicate_options(include//3, 3,
+		     [ absolute_path(atom),
+		       class(atom),
+		       files(list),
+		       href(atom),
+		       label(any),
+		       map_extension(list)
+		     ]).
+:- predicate_options(object_edit_button//2, 2,
+		     [ edit(boolean),
+		       pass_to(pred_edit_button//2, 2)
+		     ]).
+:- predicate_options(object_page//2, 2,
+		     [ for(any),
+		       header(boolean),
+		       links(boolean),
+		       no_manual(boolean),
+		       search_in(oneof([all,app,man])),
+		       search_match(oneof([name,summary])),
+		       search_options(boolean)
+		     ]).
+:- predicate_options(object_ref//2, 2,
+		     [ files(list),
+		       qualify(boolean),
+		       secref_style(oneof([number,title,number_title]))
+		     ]).
+:- predicate_options(object_synopsis//2, 2,
+		     [ href(atom)
+		     ]).
+:- predicate_options(pred_dt//3, 3,
+		     [ edit(boolean)
+		     ]).
+:- predicate_options(pred_edit_button//2, 2,
+		     [ edit(boolean)
+		     ]).
+:- predicate_options(predref//2, 2,
+		     [ files(list),
+		       prefer(oneof([manual,app])),
+		       pass_to(object_ref/4, 2)
+		     ]).
+:- predicate_options(private/2, 2,
+		     [ module(atom),
+		       public(list)
+		     ]).
+:- predicate_options(source_button//2, 2,
+		     [ files(list)
+		     ]).
+
+
 		 /*******************************
 		 *	     RESOURCES		*
 		 *******************************/
@@ -121,7 +204,8 @@ extracting module doc_wiki.pl into HTML+CSS.
 
 %%	doc_for_file(+File, +Options) is det
 %
-%	Write documentation for File as HTML.  Options:
+%	HTTP  handler  that  writes  documentation  for  File  as  HTML.
+%	Options:
 %
 %		* public_only(+Bool)
 %		If =true= (default), only emit documentation for
@@ -132,7 +216,6 @@ extracting module doc_wiki.pl into HTML+CSS.
 %		are suppressed.
 %
 %	@param File	Prolog file specification.
-%	@param Out	Output stream
 
 doc_for_file(FileSpec, Options) :-
 	absolute_file_name(FileSpec,
@@ -142,7 +225,8 @@ doc_for_file(FileSpec, Options) :-
 			   File),
 	file_base_name(File, Base),
 	Title = Base,
-	reply_html_page(title(Title),
+	reply_html_page(pldoc(result),
+			title(Title),
 			\prolog_file(FileSpec, Options)).
 
 prolog_file(FileSpec, Options) -->
@@ -174,18 +258,34 @@ prolog_file(FileSpec, Options) -->
 %	@param File	Prolog canonical filename
 
 doc_file_objects(FileSpec, File, Objects, FileOptions, Options) :-
-	absolute_file_name(FileSpec,
+	absolute_file_name(FileSpec, File,
 			   [ file_type(prolog),
 			     access(read)
-			   ],
-			   File),
-	Pos = File:_Line,
-	findall(doc(Obj,Pos,Comment),
-		doc_comment(Obj, Pos, _, Comment), Objs0),
+			   ]),
+	Pos = File:Line,
+	ensure_doc_objects(File),
+	findall(Line-doc(Obj,Pos,Comment),
+		doc_comment(Obj, Pos, _, Comment), Pairs),
+	keysort(Pairs, ByLine),
+	pairs_values(ByLine, Objs0),
 	module_info(File, ModuleOptions, Options),
 	file_info(Objs0, Objs1, FileOptions, ModuleOptions),
 	doc_hide_private(Objs1, Objects, ModuleOptions).
 
+%%	ensure_doc_objects(+File)
+%
+%	Ensure we have documentation about  File.   If  the  file is not
+%	loaded, run the cross-referencer on it   to collect the comments
+%	and meta-information.
+%
+%	@tbd	We could also run the cross-referencer if the file
+%		was loaded, but before we are collecting comments.
+%		Is that useful?
+
+ensure_doc_objects(File) :-
+	source_file(File), !.
+ensure_doc_objects(File) :-
+	xref_source(File).
 
 %%	module_info(+File, -ModuleOptions, +OtherOptions) is det.
 %
@@ -195,8 +295,14 @@ doc_file_objects(FileSpec, File, Objects, FileOptions, Options) :-
 module_info(File, [module(Module), public(Exports)|Options], Options) :-
 	module_property(Module, file(File)), !,
 	module_property(Module, exports(Exports)).
+module_info(File, [module(Module), public(Exports)|Options], Options) :-
+	xref_module(File, Module), !,
+	findall(PI, xref_exported_pi(File, PI), Exports).
 module_info(_, Options, Options).
 
+xref_exported_pi(Src, Name/Arity) :-
+	xref_exported(Src, Head),
+	functor(Head, Name, Arity).
 
 %%	doc_hide_private(+Objs, +Public, +Options)
 %
@@ -236,17 +342,42 @@ obj(Term, Obj) :-
 %	defined a module and Obj is  not   member  of the exports of the
 %	module.
 
+private(Module:PI, Options) :-
+	multifile(Module:PI, Options), !, fail.
 private(Module:PI, Options) :-
 	option(module(Module), Options),
 	option(public(Public), Options), !,
-	\+ ( member(PI2, Public) ,
+	\+ ( member(PI2, Public),
 	     eq_pi(PI, PI2)
 	   ).
 private(Module:PI, _Options) :-
+	module_property(Module, file(_)), !,	% A loaded module
 	export_list(Module, Exports),
 	\+ ( member(PI2, Exports),
 	     eq_pi(PI, PI2)
 	   ).
+private(Module:PI, _Options) :-
+	\+ (pi_to_head(PI, Head),
+	    xref_exported(Source, Head),
+	    xref_module(Source, Module)).
+
+%%	multifile(+Obj, +Options) is semidet.
+%
+%	True if Obj is a multifile predicate.
+
+multifile(Obj, _Options) :-
+	strip_module(user:Obj, Module, PI),
+	pi_to_head(PI, Head),
+	predicate_property(Module:Head, multifile).
+
+
+pi_to_head(Var, _) :-
+	var(Var), !, fail.
+pi_to_head(Name/Arity, Term) :-
+	functor(Term, Name, Arity).
+pi_to_head(Name//DCGArity, Term) :-
+	Arity is DCGArity+2,
+	functor(Term, Name, Arity).
 
 %%	file_info(+Comments, -RestComment, -FileOptions, +OtherOptions) is det.
 %
@@ -288,7 +419,7 @@ file_title(Title, File, Options) -->
 	},
 	html(h1(class=file,
 		[ span(style('float:right'),
-		       [ \reload_button(Base, Options),
+		       [ \reload_button(File, Base, Options),
 			 \zoom_button(Base, Options),
 			 \source_button(Base, Options),
 			 \edit_button(File, Options)
@@ -297,24 +428,27 @@ file_title(Title, File, Options) -->
 		])).
 
 
-%%	reload_button(+File)// is det.
+%%	reload_button(+File, +Base, +Options)// is det.
 %
 %	Create a button for  reloading  the   sources  and  updating the
-%	documentation page.
+%	documentation page. Note that the  button   is  not shown if the
+%	file is not loaded because we do  not want to load files through
+%	the documentation system.
 
-reload_button(Base, Options) -->
+reload_button(File, _Base, _Options) -->
+	{ \+ source_file(File) }, !,
+	html(span(class(file_anot), '[not loaded]')).
+reload_button(_File, Base, Options) -->
 	{ option(edit(true), Options), !,
 	  option(public_only(Public), Options, true)
 	},
 	html(a(href(Base+[reload(true), public_only(Public)]),
-	       img([ %class(icon),
-		     height=24,
+	       img([ class(action),
 		     alt('Reload'),
-		     style('padding-top:4px; border:0;'),
+		     title('Make & Reload'),
 		     src(location_by_id(pldoc_resource)+'reload.gif')
 		   ]))).
-reload_button(_, _) -->
-	[].
+reload_button(_, _, _) --> [].
 
 %%	edit_button(+File, +Options)// is det.
 %
@@ -323,17 +457,15 @@ reload_button(_, _) -->
 %	current page.  JavaScript code is in the file pldoc.js.
 
 edit_button(File, Options) -->
-	{ option(edit(true), Options), !,
-	  option(button_height(H), Options, 24)
-	},
+	{ option(edit(true), Options)
+	}, !,
 	html(a([ onClick('HTTPrequest(\'' +
 			 location_by_id(pldoc_edit) + [file(File)] +
-			 '\')'),
-		 onMouseOver('window.status=\'Edit file\'; return true;')
+			 '\')')
 	       ],
-	       img([ height(H),
+	       img([ class(action),
 		     alt(edit),
-		     style('border:0'),
+		     title('Edit file'),
 		     src(location_by_id(pldoc_resource)+'edit.gif')
 		 ]))).
 edit_button(_, _) -->
@@ -348,18 +480,20 @@ zoom_button(_, Options) -->
 	{ option(files(_Map), Options) }, !.	% generating files
 zoom_button(Base, Options) -->
 	{   (   option(public_only(true), Options, true)
-	    ->  Zoom = 'zoomin.gif',
-		Alt = 'Show all',
+	    ->  Zoom = 'public.png',
+		Alt = 'Public',
+		Title = 'Click to include private',
 		PublicOnly = false
-	    ;   Zoom = 'zoomout.gif',
-		Alt = 'Show public',
+	    ;   Zoom = 'private.png',
+		Alt = 'All predicates',
+		Title = 'Click to show exports only',
 		PublicOnly = true
 	    )
 	},
 	html(a(href(Base+[public_only(PublicOnly)]),
-	       img([ height=24,
+	       img([ class(action),
 		     alt(Alt),
-		     style('padding-top:4px; border:0;'),
+		     title(Title),
 		     src(location_by_id(pldoc_resource)+Zoom)
 		   ]))).
 
@@ -370,18 +504,17 @@ zoom_button(Base, Options) -->
 
 source_button(_File, Options) -->
 	{ option(files(_Map), Options) }, !.	% generating files
-source_button(File, Options) -->
+source_button(File, _Options) -->
 	{ (   is_absolute_file_name(File)
 	  ->  doc_file_href(File, HREF0)
 	  ;   HREF0 = File
-	  ),
-	  option(button_height(H), Options, 24)
+	  )
 	},
-	html(a(href(HREF0+[source(true)]),
-	       img([ height(H),
+	html(a(href(HREF0+[show(src)]),
+	       img([ class(action),
 		     alt('Show source'),
-		     style('padding-top:4px; border:0;'),
-		     src(location_by_id(pldoc_resource)+'source.gif')
+		     title('Show source'),
+		     src(location_by_id(pldoc_resource)+'source.png')
 		   ]))).
 
 
@@ -398,40 +531,68 @@ objects([Obj|T], Mode, Options) -->
 	object(Obj, Mode, Mode1, Options),
 	objects(T, Mode1, Options).
 
-object(doc(Obj,Pos,Comment), Mode0, Mode, Options) --> !,
-	object(Obj, Pos, Comment, Mode0, Mode, Options).
-object(Obj, Mode0, Mode, Options) -->
-	{ doc_comment(Obj, Pos, _Summary, Comment)
-	}, !,
-	object(Obj, Pos, Comment, Mode0, Mode, Options).
+%%	object(+Spec, +ModeIn, -ModeOut, +Options) is det.
+%
+%	Emit the documentation of a single object.
+%
+%	@param	Spec is one of doc(Obj,Pos,Comment), which is used
+%		to list the objects documented in a file or a plain
+%		Obj, used for documenting the object regardless of
+%		its location.
 
-object(Obj, Pos, Comment, Mode0, Mode, Options) -->
+object(doc(Obj,Pos,Comment), Mode0, Mode, Options) --> !,
+	object(Obj, [Pos-Comment], Mode0, Mode, [scope(file)|Options]).
+object(Obj, Mode0, Mode, Options) -->
+	{ findall(Pos-Comment,
+		  doc_comment(Obj, Pos, _Summary, Comment),
+		  Pairs)
+	}, !,
+	object(Obj, Pairs, Mode0, Mode, Options).
+
+object(Obj, Pairs, Mode0, Mode, Options) -->
 	{ is_pi(Obj), !,
-	  is_structured_comment(Comment, Prefixes),
-	  string_to_list(Comment, Codes),
-	  indented_lines(Codes, Prefixes, Lines),
-	  strip_module(user:Obj, Module, _),
-	  process_modes(Lines, Module, Pos, Modes, Args, Lines1),
-	  (   private(Obj, Options)
-	  ->  Class = privdef		% private definition
-	  ;   Class = pubdef		% public definition
-	  ),
-	  (   Obj = Module:_
-	  ->  POptions = [module(Module)|Options]
-	  ;   POptions = Options
-	  ),
-	  DOM = [\pred_dt(Modes, Class, POptions), dd(class=defbody, DOM1)],
-	  wiki_lines_to_dom(Lines1, Args, DOM0),
-	  strip_leading_par(DOM0, DOM1)
+	  maplist(pred_dom(Obj, Options), Pairs, DOMS),
+	  append(DOMS, DOM)
 	},
 	need_mode(dl, Mode0, Mode),
 	html(DOM).
-object([Obj|_Same], Pos, Comment, Mode0, Mode, Options) --> !,
-	object(Obj, Pos, Comment, Mode0, Mode, Options).
-object(Obj, _Pos, _Comment, Mode, Mode, _Options) -->
+object([Obj|_Same], Pairs, Mode0, Mode, Options) --> !,
+	object(Obj, Pairs, Mode0, Mode, Options).
+object(Obj, _Pairs, Mode, Mode, _Options) -->
 	{ debug(pldoc, 'Skipped ~p', [Obj]) },
 	[].
 
+pred_dom(Obj, Options, Pos-Comment, DOM) :-
+	is_structured_comment(Comment, Prefixes),
+	string_to_list(Comment, Codes),
+	indented_lines(Codes, Prefixes, Lines),
+	strip_module(user:Obj, Module, _),
+	process_modes(Lines, Module, Pos, Modes, Args, Lines1),
+	(   private(Obj, Options)
+	->  Class = privdef		% private definition
+	;   multifile(Obj, Options)
+	->  (   option(scope(file), Options)
+	    ->	(   more_doc(Obj, Pos)
+		->  Class = multidef(object(Obj))
+		;   Class = multidef
+		)
+	    ;	Class = multidef(file((Pos)))
+	    )
+	;   Class = pubdef		% public definition
+	),
+	(   Obj = Module:_
+	->  POptions = [module(Module)|Options]
+	;   POptions = Options
+	),
+	Pos = File:Line,
+	DTOptions = [file(File),line(Line)|POptions],
+	DOM = [\pred_dt(Modes, Class, DTOptions), dd(class=defbody, DOM1)],
+	wiki_lines_to_dom(Lines1, Args, DOM0),
+	strip_leading_par(DOM0, DOM1).
+
+more_doc(Obj, File:_) :-
+	doc_comment(Obj, File2:_, _, _),
+	File2 \== File, !.
 
 %%	need_mode(+Mode:atom, +Stack:list, -NewStack:list)// is det.
 %
@@ -546,22 +707,112 @@ object_page(Obj, Options) -->
 	prolog:doc_object_page(Obj, Options).
 object_page(Obj, Options) -->
 	{ doc_comment(Obj, File:_Line, _Summary, _Comment)
-	},
-	html([ \html_requires(pldoc),
-	       \object_page_header(File, Options),
-	       \objects([Obj], Options)
-	     ]).
+	}, !,
+	(   { \+ ( doc_comment(Obj, File2:_, _, _),
+		   File2 \== File )
+	    }
+	->  html([ \html_requires(pldoc),
+		   \object_page_header(File, Options),
+		   \object_synopsis(Obj, []),
+		   \objects([Obj], Options)
+		 ])
+	;   html([ \html_requires(pldoc),
+		   \object_page_header(-, Options),
+		   \objects([Obj], [synopsis(true)|Options])
+		 ])
+	).
 
 object_page_header(File, Options) -->
 	{ option(header(true), Options, true) }, !,
 	html(div(class(navhdr),
-		 [ div(class(jump),
-		       a(href(location_by_id(pldoc_doc)+File), File)),
+		 [ \file_link(File),
 		   div(class(search), \search_form(Options)),
 		   br(clear(right))
 		 ])).
 object_page_header(_, _) --> [].
 
+file_link(-) --> !.
+file_link(File) -->
+	html(div(class(jump),
+		 a(href(location_by_id(pldoc_doc)+File), File))).
+
+
+%%	object_synopsis(Obj, Options) is det.
+%
+%	Provide additional information  about  Obj.   Note  that  due to
+%	reexport facilities, predicates may be   available from multiple
+%	modules.
+%
+%	@tbd Currently we provide a synopsis   for the one where the
+%	definition resides. This is not   always  correct. Notably there
+%	are cases where multiple implementation modules are bundled in a
+%	larger interface that is the `preferred' module.
+
+object_synopsis(Name/Arity, _) -->
+	{ functor(Head, Name, Arity),
+	  predicate_property(system:Head, built_in)
+	},
+	synopsis([span(class(builtin), 'built-in')]).
+object_synopsis(Name/Arity, Options) --> !,
+	object_synopsis(_:Name/Arity, Options).
+object_synopsis(M:Name/Arity, Options) -->
+	{ functor(Head, Name, Arity),
+	  predicate_property(M:Head, exported),
+	  \+ predicate_property(M:Head, imported_from(_)),
+	  module_property(M, file(File)),
+	  file_name_on_path(File, Spec), !,
+	  unquote_filespec(Spec, Unquoted),
+	  (   predicate_property(Head, autoload(FileBase)),
+	      file_name_extension(FileBase, _Ext, File)
+	  ->  Extra = [span(class(autoload), '(can be autoloaded)')]
+	  ;   Extra = []
+	  )
+	},
+	(   { option(href(HREF), Options) }
+	->  synopsis([code([':- use_module(',a(href(HREF), '~q'-[Unquoted]),').'])|Extra])
+	;   synopsis([code(':- use_module(~q).'-[Unquoted])|Extra])
+	).
+object_synopsis(f(_/_), _) -->
+	synopsis(span(class(function),
+		      [ 'Arithmetic function (see ',
+			\object_ref(is/2, []),
+			')'
+		      ])).
+object_synopsis(c(Func), _) -->
+	{ sub_atom(Func, 0, _, _, 'PL_')
+	}, !,
+	synopsis([span(class(cfunc), 'C-language interface function')]).
+object_synopsis(_, _) --> [].
+
+synopsis(Text) -->
+	html(div(class(synopsis),
+		 [ span(class('synopsis-hdr'), 'Availability:')
+		 | Text
+		 ])).
+
+%%	unquote_filespec(+Spec, -Unquoted) is det.
+%
+%	Translate       e.g.       library('semweb/rdf_db')         into
+%	library(semweb/rdf_db).
+
+unquote_filespec(Spec, Unquoted) :-
+	compound(Spec),
+	Spec =.. [Alias,Path],
+	atomic_list_concat(Parts, /, Path),
+	maplist(need_no_quotes, Parts), !,
+	parts_to_path(Parts, UnquotedPath),
+	Unquoted =.. [Alias, UnquotedPath].
+unquote_filespec(Spec, Spec).
+
+need_no_quotes(Atom) :-
+	format(atom(A), '~q', [Atom]),
+	\+ sub_atom(A, 0, _, _, '\'').
+
+parts_to_path([One], One) :- !.
+parts_to_path(List, More/T) :-
+	(   append(H, [T], List)
+	->  parts_to_path(H, More)
+	).
 
 
 		 /*******************************
@@ -703,10 +954,34 @@ pred_dt(Modes, Class, Options) -->
 pred_dt([], _, Done, Done, _) -->
 	[].
 pred_dt([H|T], Class, Done0, Done, Options) -->
-	html(dt(class=Class,
-		\pred_mode(H, Done0, Done1, Options))),
+	{ functor(Class, CSSClass, _) },
+	html(dt(class=CSSClass,
+		[ \pred_mode(H, Done0, Done1, Options),
+		  \mode_anot(Class)
+		])),
 	pred_dt(T, Class, Done1, Done, Options).
 
+mode_anot(privdef) --> !,
+	html(span([class(anot), style('float:right')],
+		  '[private]')).
+mode_anot(multidef(object(Obj))) --> !,
+	{ object_href(Obj, HREF) },
+	html(span([class(anot), style('float:right')],
+		  ['[', a(href(HREF), multifile), ']'
+		  ])).
+mode_anot(multidef(file(File:_))) --> !,
+	{ file_name_on_path(File, Spec),
+	  unquote_filespec(Spec, Unquoted),
+	  doc_file_href(File, HREF)
+	},
+	html(span([class(anot), style('float:right')],
+		  ['[multifile, ', a(href(HREF), '~q'-[Unquoted]), ']'
+		  ])).
+mode_anot(multidef) --> !,
+	html(span([class(anot), style('float:right')],
+		  '[multifile]')).
+mode_anot(_) -->
+	[].
 
 pred_mode(mode(Head,Vars), Done0, Done, Options) --> !,
 	{ bind_vars(Head, Vars) },
@@ -755,7 +1030,15 @@ anchored_pred_head(Head, Done0, Done, Options) -->
 
 %%	pred_edit_button(+PredIndicator, +Options)// is det.
 %
-%	Create a button for editing the given predicate.
+%	Create a button for editing the given predicate.  Options
+%	processed:
+%
+%	    * module(M)
+%	    Resolve to module M
+%	    * file(F)
+%	    For multi-file predicates: link to version in file.
+%	    * line(L)
+%	    Line to edit (in file)
 
 pred_edit_button(_, Options) -->
 	{ \+ option(edit(true), Options) }, !.
@@ -764,29 +1047,34 @@ pred_edit_button(PI0, Options0) -->
 	pred_edit_button2(PI, Options).
 
 pred_edit_button2(Name/Arity, Options) -->
-	{ functor(Head, Name, Arity),
+	{ \+ ( memberchk(file(_), Options), % always edit if file and line
+	       memberchk(line(_), Options)  % are given.
+	     ),
+	  functor(Head, Name, Arity),
 	  option(module(M), Options, _),
 	  \+ ( current_module(M),
 	       source_file(M:Head, _File)
 	     )
 	}, !.
 pred_edit_button2(Name/Arity, Options) -->
-	{ (   option(module(M), Options)
-	  ->  Extra = [module(M)]
-	  ;   Extra = []
-	  )
+	{ include(edit_param, Options, Extra),
+	  http_link_to_id(pldoc_edit,
+			  [name(Name),arity(Arity)|Extra],
+			  EditHREF)
 	},
-	html(a([ onClick('HTTPrequest(\'' +
-			 location_by_id(pldoc_edit)+[name(Name),arity(Arity)|Extra] +
-			 '\')')
-	       ],
-	       img([ height=12,
-		     style('border:0;'),
+	html(a(onClick('HTTPrequest(\'' + EditHREF + '\')'),
+	       img([ class(action),
+		     alt('Edit predicate'),
+		     title('Edit predicate'),
 		     src(location_by_id(pldoc_resource)+'edit.gif')
 		   ]))).
 pred_edit_button2(_, _) --> !,
 	[].
 
+edit_param(module(_)).
+edit_param(file(_)).
+edit_param(line(_)).
+
 
 %%	object_edit_button(+Object, +Options)// is det.
 %
@@ -812,9 +1100,10 @@ pred_source_button(PI0, Options0) -->
 	},
 	html(a([ href(HREF)
 	       ],
-	       img([ height=12,
-		     style('border:0;'),
-		     src(location_by_id(pldoc_resource)+'source.gif')
+	       img([ class(action),
+		     alt('Source'),
+		     title('Show source'),
+		     src(location_by_id(pldoc_resource)+'source.png')
 		   ]))).
 pred_source_button(_, _) -->
 	[].
@@ -836,7 +1125,7 @@ object_source_button(_, _) -->
 %	Canonise a predicate reference. A   possible module qualifier is
 %	added as module(M) to Options.
 
-canonise_predref(M:PI0, PI, Options0, [module(M),Options]) :- !,
+canonise_predref(M:PI0, PI, Options0, [module(M)|Options]) :- !,
 	canonise_predref(PI0, PI, Options0, Options).
 canonise_predref(//(Head), PI, Options0, Options) :-	!,
 	functor(Head, Name, Arity),
@@ -857,9 +1146,16 @@ canonise_predref(Head, PI, Options0, Options) :-
 %	Emit a predicate head. The functor is  typeset as a =span= using
 %	class =pred= and the arguments and =var= using class =arglist=.
 
+pred_head(Var) -->
+	{ var(Var), !,
+	  instantiation_error(Var)
+	}.
 pred_head(//(Head)) --> !,
 	pred_head(Head),
 	html(//).
+pred_head(M:Head) -->
+	html([span(class=module, M), :]),
+	pred_head(Head).
 pred_head(Head) -->
 	{ atom(Head) }, !,
 	html(b(class=pred, Head)).
@@ -996,6 +1292,11 @@ predref(Term) -->
 	{ catch(nb_getval(pldoc_options, Options), _, Options = []) },
 	predref(Term, Options).
 
+predref(Obj, Options) -->
+	{ Obj = _:_,
+	  doc_comment(Obj, _, _, _)
+	}, !,
+	object_ref(Obj, [qualify(true)|Options]).
 predref(M:Term, Options) --> !,
 	predref(Term, M, Options).
 predref(Term, Options) -->
@@ -1006,6 +1307,12 @@ predref(Name/Arity, _, Options) -->		% Builtin; cannot be overruled
 	  manref(Name/Arity, HREF, Options)
 	},
 	html(a([class=builtin, href=HREF], [Name, /, Arity])).
+predref(Name/Arity, _, Options) -->		% From packages
+	{ option(prefer(manual), Options),
+	  prolog:doc_object_summary(Name/Arity, Category, _, _), !,
+	  manref(Name/Arity, HREF, Options)
+	},
+	html(a([class=Category, href=HREF], [Name, /, Arity])).
 predref(Obj, Module, Options) -->		% Local
 	{ doc_comment(Module:Obj, _, _, _)
 	}, !,
@@ -1040,14 +1347,14 @@ predref(Callable, Module, Options) -->
 
 manref(Name/Arity, HREF, Options) :-
 	format(string(FragmentId), '~w/~d', [Name, Arity]),
-	www_form_encode(FragmentId, EncId),
 	(   option(files(_Map), Options)
-	->  option(man_server(ManHandler), Options,
-		   'http://gollem.science.uva.nl/SWI-Prolog/pldoc/man')
-	;   http_location_by_id(pldoc_man, ManHandler)
-	),
-	http_location_by_id(pldoc_man, ManHandler),
-	format(string(HREF), '~w?predicate=~w', [ManHandler, EncId]).
+	->  uri_query_components(Query, [predicate=FragmentId]),
+	    uri_data(authority, Components, 'www.swi-prolog.org'),
+	    uri_data(path, Components, '/pldoc/man'),
+	    uri_data(search, Components, Query),
+	    uri_components(HREF, Components)
+	;   http_link_to_id(pldoc_man, [predicate=FragmentId], HREF)
+	).
 
 
 %%	pred_href(+NameArity, +Module, -HREF) is semidet.
@@ -1063,20 +1370,20 @@ manref(Name/Arity, HREF, Options) :-
 
 pred_href(Name/Arity, Module, HREF) :-
 	format(string(FragmentId), '~w/~d', [Name, Arity]),
-	www_form_encode(FragmentId, EncId),
+	uri_data(fragment, Components, FragmentId),
 	functor(Head, Name, Arity),
 	(   catch(relative_file(Module:Head, File), _, fail)
-	->  format(string(HREF), '~w#~w', [File, EncId])
+	->  uri_data(path, Components, File),
+	    uri_components(HREF, Components)
 	;   in_file(Module:Head, File)
 	->  (	current_prolog_flag(home, SWI),
 		sub_atom(File, 0, _, _, SWI),
 		prolog:doc_object_summary(Name/Arity, packages, _, _)
-	    ->	format(string(FragmentId), '~w/~d', [Name, Arity]),
-		www_form_encode(FragmentId, EncId),
-		http_location_by_id(pldoc_man, ManHandler),
-		format(string(HREF), '~w?predicate=~w', [ManHandler, EncId])
+	    ->	http_link_to_id(pldoc_man, [predicate=FragmentId], HREF)
 	    ;	http_location_by_id(pldoc_doc, DocHandler),
-	        format(string(HREF), '~w~w#~w', [DocHandler, File, EncId])
+		atom_concat(DocHandler, File, Path),
+		uri_data(path, Components, Path),
+		uri_components(HREF, Components)
 	    )
 	).
 
@@ -1094,13 +1401,18 @@ relative_file(Head, RelFile) :-
 
 pred_source_href(Name/Arity, Module, HREF) :-
 	format(string(FragmentId), '~w/~d', [Name, Arity]),
-	www_form_encode(FragmentId, EncId),
+	uri_data(fragment, Components, FragmentId),
+	uri_query_components(Query, [source=true]),
+	uri_data(search, Components, Query),
 	functor(Head, Name, Arity),
 	(   catch(relative_file(Module:Head, File), _, fail)
-	->  format(string(HREF), '~w?source=true#~w', [File, EncId])
+	->  uri_data(path, Components, File),
+	    uri_components(HREF, Components)
 	;   in_file(Module:Head, File),
 	    http_location_by_id(pldoc_doc, DocHandler),
-	    format(string(HREF), '~w~w?source=true#~w', [DocHandler, File, EncId])
+	    atom_concat(DocHandler, File, Path),
+	    uri_data(path, Components, Path),
+	    uri_components(HREF, Components)
 	).
 
 
@@ -1136,17 +1448,16 @@ object_href(M:PI0, HREF, Options) :-
 	option(files(Map), Options),
 	module_property(M, file(File)),
 	memberchk(file(File, DocFile), Map), !,
+	file_base_name(DocFile, LocalFile),	% TBD: proper directory index
 	expand_pi(PI0, PI),
 	term_to_string(PI, PIS),
-	www_form_encode(PIS, PIEnc),
-	file_base_name(DocFile, LocalFile),	% TBD: proper directory index
-	format(string(HREF), '~w#~w', [LocalFile, PIEnc]).
+	uri_data(path, Components, LocalFile),
+	uri_data(fragment, Components, PIS),
+	uri_components(HREF, Components).
 object_href(Obj0, HREF, _Options) :-
 	localise_object(Obj0, Obj),
 	term_to_string(Obj, String),
-	www_form_encode(String, Enc),
-	http_location_by_id(pldoc_object, ObjHandler),
-	format(string(HREF), '~w?object=~w', [ObjHandler, Enc]).
+	http_link_to_id(pldoc_object, [object=String], HREF).
 
 expand_pi(Name//Arity0, Name/Arity) :- !,
 	Arity is Arity0+2.
@@ -1188,20 +1499,24 @@ term_to_string(Term, String) :-
 
 object_link(Obj, Options) -->
 	prolog:doc_object_link(Obj, Options), !.
-object_link(PI, _) -->
+object_link(PI, Options) -->
 	{ is_pi(PI) }, !,
-	pi(PI).
+	pi(PI, Options).
 object_link(Module:module(_Title), _) -->
 	{ module_property(Module, file(File)),
 	  file_base_name(File, Base)
 	}, !,
 	html(Base).
 
-pi(_M:PI) --> !,
-	pi(PI).
-pi(Name/Arity) --> !,
+pi(M:PI, Options) --> !,
+	(   { option(qualify(true), Options) }
+	->  html([span(class(module), M), :])
+	;   []
+	),
+	pi(PI, Options).
+pi(Name/Arity, _) --> !,
 	html([Name, /, Arity]).
-pi(Name//Arity) -->
+pi(Name//Arity, _) -->
 	html([Name, //, Arity]).
 
 %%	in_file(+Head, ?File) is nondet.
@@ -1234,7 +1549,7 @@ in_file(Module, Head, File) :-
 
 %%	relative_file_name(+Path:atom, +RelTo:atom, -RelPath:atom) is det.
 %
-%	Create a relative path from an absolute one.
+%	True when RelPath is a relative path to Path, relative to RelTo.
 %
 %	@tbd	move to library?
 
@@ -1278,6 +1593,9 @@ to_dot_dot([_|T0], ['..'|T], Tail) :-
 %          List of file(Name, Link) that specifies that we must
 %	   user Link for the given physical file Name.
 %
+%	   * edit_handler(+Id)
+%	   HTTP handler Id to call if the user clicks the edit button.
+%
 %	@tbd	Translation of files to HREFS is a mess.  How to relate
 %		these elegantly?
 
@@ -1289,6 +1607,17 @@ file(File, Options) -->
 	  merge_options(Options, GenOptions, FinalOptions)
 	},
 	link_file(File, FinalOptions), !.
+file(File, Options) -->
+	{ option(edit_handler(Handler), Options),
+	  http_current_request(Request),
+	  memberchk(path(Path), Request),
+	  absolute_file_name(File, Location,
+			     [ relative_to(Path)
+			     ]),
+	  http_link_to_id(Handler, [location(Location)], HREF),
+	  format(atom(Title), 'Click to create ~w', [File])
+	},
+	html(a([href(HREF), class(nofile), title(Title)], File)).
 file(File, _) -->
 	html(code(class(nofile), File)).
 
@@ -1334,7 +1663,7 @@ file_href_real(File, HREF, Options) :-
 	),
 	file_href(LinkFile, HREF).
 file_href_real(File, HREF, _) :-
-	http:location(Alias, _, _),
+	directory_alias(Alias),
 	Term =.. [Alias,File],
 	absolute_file_name(Term, _,
 			   [ access(read),
@@ -1342,6 +1671,10 @@ file_href_real(File, HREF, _) :-
 			   ]), !,
 	http_absolute_location(Term, HREF, []).
 
+directory_alias(icons).
+directory_alias(css).
+
+
 %%	file_href(+FilePath, -HREF) is det.
 %
 %	Create a relative URL from  the   current  location to the given
@@ -1349,6 +1682,9 @@ file_href_real(File, HREF, _) :-
 %	file being processed  that  is   available  through  the  global
 %	variable =pldoc_file=.
 
+file_href(Path, HREF) :-		% a loaded Prolog file
+	source_file(Path), !,
+	doc_file_href(Path, HREF).
 file_href(Path, HREF) :-
 	nb_current(pldoc_file, CFile),
 	CFile \== [], !,
@@ -1384,7 +1720,18 @@ include(PI, predicate, _) --> !,
 	;   html(['[[', \predref(PI), ']]'])
 	).
 include(File, image, Options) -->
-	{ file_href(File, HREF, Options),
+	{ file_name_extension(_, svg, File),
+	  file_href(File, HREF, Options), !,
+	  include(image_attribute, Options, Attrs0),
+	  merge_options(Attrs0,
+			[ alt(File),
+			  data(HREF),
+			  type('image/svg+xml')
+			], Attrs)
+	},
+	html(object(Attrs, [])).
+include(File, image, Options) -->
+	{ file_href(File, HREF, Options), !,
 	  include(image_attribute, Options, Attrs0),
 	  merge_options(Attrs0,
 			[ alt(File),
@@ -1428,7 +1775,7 @@ html_tokens_for_predicates(PI, Options) -->
 	  ;   Comment = ''
 	  )
 	},
-	object(PI, Pos, Comment, [dl], _, Options).
+	object(PI, [Pos-Comment], [dl], _, Options).
 html_tokens_for_predicates(Spec, Options) -->
 	{ findall(PI, documented_pi(Spec, PI), List),
 	  List \== [], !
@@ -1455,19 +1802,21 @@ generalise_spec(Name//Arity, _M:Name//Arity).
 %
 %	Write HTML for the File containing wiki data.
 
-doc_for_wiki_file(FileSpec, _Options) :-
+doc_for_wiki_file(FileSpec, Options) :-
 	absolute_file_name(FileSpec, File,
 			   [ access(read)
 			   ]),
 	read_file_to_codes(File, String, []),
 	b_setval(pldoc_file, File),
-	call_cleanup(reply_wiki_page(File, String),
+	call_cleanup(reply_wiki_page(File, String, Options),
 		     nb_delete(pldoc_file)).
 
-reply_wiki_page(File, String) :-
-	wiki_codes_to_dom(String, [], DOM),
-	title(DOM, File, Title),
-	reply_html_page(title(Title),
+reply_wiki_page(File, String, Options) :-
+	wiki_codes_to_dom(String, [], DOM0),
+	title(DOM0, File, Title),
+	insert_edit_button(DOM0, File, DOM, Options),
+	reply_html_page(pldoc(wiki),
+			title(Title),
 			[ \html_requires(pldoc)
 			| DOM
 			]).
@@ -1477,6 +1826,21 @@ title(DOM, _, Title) :-
 title(_, File, Title) :-
 	file_base_name(File, Title).
 
+insert_edit_button(DOM, _, DOM, Options) :-
+	option(edit(false), Options, false), !.
+insert_edit_button([h1(Attrs,Title)|DOM], File,
+		   [h1(Attrs,[ span(style('float:right'),
+				   \edit_button(File, [edit(true)]))
+			     | Title
+			     ])|DOM], _) :- !.
+insert_edit_button(DOM, File,
+		   [ h1(class(wiki),
+			[ span(style('float:right'),
+			       \edit_button(File, [edit(true)]))
+			])
+		   | DOM
+		   ], _).
+
 
 		 /*******************************
 		 *	      ANCHORS		*
@@ -1489,7 +1853,7 @@ title(_, File, Title) :-
 pred_anchor_name(//(Head), Name/Arity, Anchor) :- !,
 	functor(Head, Name, DCGArity),
 	Arity is DCGArity+2,
-	format(string(Anchor), '~w/~d', [Name, Arity]).
+	format(atom(Anchor), '~w/~d', [Name, Arity]).
 pred_anchor_name(Head, Name/Arity, Anchor) :-
 	functor(Head, Name, Arity),
-	format(string(Anchor), '~w/~d', [Name, Arity]).
+	format(atom(Anchor), '~w/~d', [Name, Arity]).
