@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 
     As a special exception, if you link this library with other files,
     compiled with a Free Software compiler, to produce an executable, this
@@ -37,6 +37,7 @@
 	    indented_lines/3,		% +Text, +PrefixChars, -Lines
 	    strip_leading_par/2,	% +DOM0, -DOM
 	    normalise_white_space/3,	% -Text, //
+	    autolink_extension/2,	% ?Extension, ?Type
 	    autolink_file/2		% +FileName, -Type
 	  ]).
 :- use_module(library(lists)).
@@ -418,7 +419,8 @@ tag_name(w(Name), Tag, Order) :-
 
 rest_tag([], _, [], []) :- !.
 rest_tag(Lines, Indent, [], Lines) :-
-	Lines = [Indent-[@,w(_Name)|_]|_], !.
+	Lines = [Indent-[@,Word|_]|_],
+	tag_name(Word, _, _), !.
 rest_tag([L|Lines0], Indent, [L|VT], Lines) :-
 	rest_tag(Lines0, Indent, VT, Lines).
 
@@ -436,19 +438,23 @@ renamed_tag(exception, throws).
 %	the tags are forced into  this   order  without  warning. Future
 %	versions may issue a warning if the order is inconsistent.
 
-tag_order(param,       1).
-tag_order(error,       2).		% same as throw
-tag_order(throws,      3).
-tag_order(author,      4).
-tag_order(version,     5).
-tag_order(see,	       6).
-tag_order(deprecated,  7).
-tag_order(compat,      8).		% PlDoc extension
-tag_order(copyright,   9).
-tag_order(license,    10).
-tag_order(bug,	      11).
-tag_order(tbd,	      12).
+:- multifile
+	pldoc:tag_order/2.
 
+tag_order(Tag, Order) :-
+	pldoc:tag_order(Tag, Order), !.
+tag_order(param,       100).
+tag_order(error,       200).		% same as throw
+tag_order(throws,      300).
+tag_order(author,      400).
+tag_order(version,     500).
+tag_order(see,	       600).
+tag_order(deprecated,  700).
+tag_order(compat,      800).		% PlDoc extension
+tag_order(copyright,   900).
+tag_order(license,    1000).
+tag_order(bug,	      1100).
+tag_order(tbd,	      1200).
 
 %%	combine_tags(+Tags:list(tag(Key, Value)), -Tags:list) is det.
 %
@@ -580,11 +586,21 @@ wiki_face(i(Italic), ArgNames) -->
 wiki_face(code(Code), _) -->
 	[=, w(Code), =], !.
 wiki_face(code(Code), _) -->
-	[=,'|'], wiki_faces(Code, []), ['|',=], !.
+	[=,'|'], wiki_words(Code), ['|',=], !.
+wiki_face(Face, _) -->
+	[ w(Name) ], arg_list(List),
+	{ atomic_list_concat([Name|List], Text),
+	  catch(atom_to_term(Text, Term, Vars), _, fail),
+	  term_face(Term, Vars, Face)
+	}, !.
 wiki_face(\predref(Name/Arity), _) -->
 	[ w(Name), '/' ], arity(Arity),
 	{ functor_name(Name)
 	}, !.
+wiki_face(\predref(Module:(Name/Arity)), _) -->
+	[ w(Module), ':', w(Name), '/' ], arity(Arity),
+	{ functor_name(Name)
+	}, !.
 wiki_face(\predref(Name/Arity), _) -->
 	prolog_symbol_char(S0),
 	symbol_string(SRest), [ '/' ], arity(Arity), !,
@@ -594,6 +610,10 @@ wiki_face(\predref(Name//Arity), _) -->
 	[ w(Name), '/', '/' ], arity(Arity),
 	{ functor_name(Name)
 	}, !.
+wiki_face(\predref(Module:(Name//Arity)), _) -->
+	[ w(Module), ':', w(Name), '/', '/' ], arity(Arity),
+	{ functor_name(Name)
+	}, !.
 wiki_face(\include(Name, Type, Options), _) -->
 	['[','['], file_name(Base, Ext), [']',']'],
 	{ autolink_extension(Ext, Type), !,
@@ -619,6 +639,57 @@ wiki_face(FT, ArgNames) -->
 	{ wiki_faces(Structure, ArgNames, FT)
 	}.
 
+wiki_words([]) --> [].
+wiki_words([Word|T]) --> [w(Word)], wiki_words(T).
+wiki_words([Punct|T]) --> [Punct], {atomic(Punct)}, wiki_words(T).
+
+
+%%	arg_list(-Atoms) is nondet.
+%
+%	Atoms  is  a  token-list  for  a    Prolog   argument  list.  An
+%	argument-list is a sequence of tokens '(' ... ')'.
+%
+%	@bug	the current implementation does not deal correctly with
+%		brackets that are embedded in quoted strings.
+
+arg_list(['('|T]) -->
+	['('], arg_list_close(T, 1).
+
+arg_list_close(Tokens, Depth) -->
+	[')'], !,
+	(   { Depth == 1 }
+	->  { Tokens = [')'] }
+	;   { Depth > 1 }
+	->  { Tokens = [')'|More],
+	      NewDepth is Depth - 1
+	    },
+	    arg_list_close(More, NewDepth)
+	).
+arg_list_close(['('|T], Depth) -->
+	['('], { NewDepth is Depth+1 },
+	arg_list_close(T, NewDepth).
+arg_list_close([H|T], Depth) -->
+	[w(H)], !,
+	arg_list_close(T, Depth).
+arg_list_close([H|T], Depth) -->
+	[H],
+	arg_list_close(T, Depth).
+
+
+%%	term_face(+Term, +Vars, -Face) is semidet.
+%
+%	Process embedded Prolog-terms. Currently   processes  Alias(Arg)
+%	terms that refer to files.  Future   versions  will also provide
+%	pretty-printing of Prolog terms.
+
+term_face(Term, _Vars, \file(Name, FileOptions)) :-
+	ground(Term),
+	compound(Term),
+	Term =.. [Alias,_],
+	user:file_search_path(Alias, _),
+	existing_file(Term, FileOptions, []), !,
+	format(atom(Name), '~q', [Term]).
+
 
 %%	make_label(+Parts, -Label) is det.
 %
@@ -735,7 +806,7 @@ wiki_link(a(href(Ref), Label), Options) -->
 	}.
 
 
-%%	filename(-Name:atom, -Ext:atom)// is semidet.
+%%	file_name(-Name:atom, -Ext:atom)// is semidet.
 %
 %	Matches a filename.  A filename is defined as a	sequence
 %	<segment>{/<segment}.<ext>.
@@ -750,6 +821,9 @@ segment(..) -->
 	['.','.'], !.
 segment(Word) -->
 	[w(Word)].
+segment(Dir) -->
+	[w(Word),'.',w(d)],
+	{ atom_concat(Word, '.d', Dir) }.
 
 segments([H|T]) -->
 	['/'], !,
@@ -772,16 +846,24 @@ file_extension(Ext) -->
 %	file.
 
 resolve_file(Name, Options, Rest) :-
+	existing_file(Name, Options, Rest), !.
+resolve_file(_, Options, Options).
+
+
+existing_file(Name, Options, Rest) :-
 	nb_current(pldoc_file, RelativeTo),
 	RelativeTo \== [],
+	(   compound(Name)
+	->  Extra = [file_type(prolog)]
+	;   Extra = []
+	),
 	absolute_file_name(Name, Path,
 			   [ relative_to(RelativeTo),
 			     access(read),
 			     file_errors(fail)
-			   ]), !,
+			   | Extra
+			   ]),
 	Options = [ absolute_path(Path) | Rest ].
-resolve_file(_, Options, Options).
-
 
 %%	arity(-Arity:int)// is semidet.
 %
@@ -847,8 +929,7 @@ url_protocol(mailto).
 peek_end_url(space) -->
 	peek(End),
 	{ space_atom(End) }, !.
-peek_end_url(space) -->
-	eos, !.
+peek_end_url(space, [], []) :- !.
 peek_end_url(Token) -->
 	peek(Token), !.
 
@@ -862,12 +943,13 @@ space_atom('\n').
 %	in the documentation.
 
 autolink_extension(Ext, prolog) :-
-	prolog_file_type(Ext,prolog).
+	user:prolog_file_type(Ext,prolog).
 autolink_extension(txt, wiki).
 autolink_extension(gif, image).
 autolink_extension(png, image).
 autolink_extension(jpg, image).
 autolink_extension(jpeg, image).
+autolink_extension(svg, image).
 
 %%	autolink_file(?File, -Type) is nondet.
 %
@@ -1092,10 +1174,10 @@ end_sentence([H|T0], [H|T]) :-
 
 indented_lines(Comment, Prefixes, Lines) :-
 	must_be(codes, Comment),
-	phrase(split_lines(Prefixes, Lines), Comment).
+	phrase(split_lines(Prefixes, Lines), Comment), !.
 
 split_lines(_, []) -->
-	end_of_comment, !.
+	end_of_comment.
 split_lines(Prefixes, [Indent-L1|Ls]) -->
 	take_prefix(Prefixes, 0, Indent0),
 	white_prefix(Indent0, Indent),
@@ -1103,16 +1185,16 @@ split_lines(Prefixes, [Indent-L1|Ls]) -->
 	split_lines(Prefixes, Ls).
 
 
-%%	end_of_comment// is det.
+%%	end_of_comment//
 %
 %	Succeeds if we hit the end of the comment.
 %
 %	@bug	%*/ will be seen as the end of the comment.
 
 end_of_comment -->
-	eos, !.
+	eos.
 end_of_comment -->
-	ws, stars, "*/", !.
+	ws, stars, "*/".
 
 stars --> [].
 stars --> "*", !, stars.
@@ -1322,3 +1404,18 @@ token(Token) -->
 
 token(w(_)) :- !.
 token(Token) :- atom(Token).
+
+
+		 /*******************************
+		 *	     MESSAGES		*
+		 *******************************/
+
+:- multifile
+	prolog:message//1.
+
+prolog:message(pldoc(deprecated_tag(Name, Tag))) -->
+	[ 'PlDoc: Deprecated tag @~w (use @~w)'-[Name, Tag]
+	].
+prolog:message(pldoc(unknown_tag(Name))) -->
+	[ 'PlDoc: unknown tag @~w'-[Name]
+	].
