@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
 #ifdef HAVE_CONFIG_H
@@ -45,7 +45,7 @@
 		 *	   TEXT HANDLING	*
 		 *******************************/
 
-inline int
+static inline int
 get_atom_text(atom_t atom, text *txt)
 { if ( (txt->a = (const charA*)PL_atom_nchars(atom, &txt->length)) )
   { txt->w = NULL;
@@ -93,12 +93,12 @@ cmpA(int c1, int c2, int *dl2)
   { int k1 = sort_pointA(c1);
     int k2 = sort_pointA(c2);
     int d;
-    
+
     if ( (d=((k1>>8)-(k2>>8))) == 0 )
     { if ( *dl2 == 0 )
 	*dl2 = (k1&0xff) - (k2&0xff);
     }
-    
+
     return d;
   }
 }
@@ -117,7 +117,7 @@ cmpW(int c1, int c2, int *dl2)
     { if ( *dl2 == 0 )
 	*dl2 = (k1&0xff) - (k2&0xff);
     }
-    
+
     return d;
   }
 }
@@ -129,7 +129,7 @@ cmp_atom_info(atom_info *info, atom_t a2)
   int i;
   int dl2 = 0;
   size_t n;
-  
+
   if ( info->handle == a2 )
     return 0;
 
@@ -161,7 +161,7 @@ cmp_atom_info(atom_info *info, atom_t a2)
     { if ( n-- == 0 )
       { if ( info->text.length == t2.length )
 	  goto eq;
-	
+
 	return info->text.length < t2.length ? -1 : 1;
       } else
       { int d;
@@ -171,12 +171,12 @@ cmp_atom_info(atom_info *info, atom_t a2)
       }
     }
   }
-  
+
   for(i=0; ; i++)
   { if ( n-- == 0 )
     { if ( info->text.length == t2.length )
 	  goto eq;
-	
+
       return info->text.length < t2.length ? -1 : 1;
     } else
     { wint_t c1 = fetch(&info->text, i);
@@ -228,7 +228,7 @@ string_hashA(const char *s, size_t len)
     t--;
     while(++t<e)
       *++o = sort_pointA(*t)>>8;
-    hash ^= MurmurHashAligned2(buf, cp, MURMUR_SEED);
+    hash ^= rdf_murmer_hash(buf, cp, MURMUR_SEED);
 
     len -= cp;
   }
@@ -249,7 +249,7 @@ string_hashW(const wchar_t *t, size_t len)
 
     while(t<e)
       *o++ = (short)(sort_point(*t++)>>8);
-    hash ^= MurmurHashAligned2(buf, cp*sizeof(short), MURMUR_SEED);
+    hash ^= rdf_murmer_hash(buf, cp*sizeof(short), MURMUR_SEED);
 
     len -= cp;
   }
@@ -297,25 +297,30 @@ first_atom(atom_t a, int match)
     wchar_t buf[256];
     wchar_t *out, *s;
     int i;
-    wint_t c;
     atom_t rc;
-    
+
     if ( len <= 256 )
       out = buf;
     else
       out = PL_malloc(len*sizeof(wchar_t));
-    
-    for(s=out,i=0; (c=fetch(&t,i)); s++,i++)
-    { if ( c == '*' && match == STR_MATCH_LIKE )
+
+    for(s=out,i=0; i<len; s++,i++)
+    { wint_t c = fetch(&t,i);
+
+      if ( c == '*' && match == STR_MATCH_LIKE )
       { if ( i == 0 )			/* like '*...' */
-	  return (atom_t)0;
+	{ rc = 0;
+	  goto out;
+	}
 	len = i;			/* only up to the first * */
+	break;
       }
       *s = sort_point(c)>>8;
     }
 
     rc = PL_new_atom_wchars(len, out);
 
+  out:
     if ( out != buf )
       PL_free(out);
 
@@ -362,7 +367,7 @@ matchA(int how, const charA *f, const charA *l)
       }
       if ( *l == '\0' && *f == '\0' )
 	return TRUE;
-  
+
       return FALSE;
     }
     case STR_MATCH_PREFIX:
@@ -372,13 +377,13 @@ matchA(int how, const charA *f, const charA *l)
       }
       if ( *f == '\0' )
 	return TRUE;
-  
+
       return FALSE;
     }
     case STR_MATCH_SUBSTRING:		/* use Boyle-More! */
     { const charA *h;
       const charA *f0 = f;
-  
+
       for(h=l; *h; h++)
       { for( l=h,f=f0; *l && *f; l++, f++ )
 	{ if ( cmp_pointA(*l) != cmp_pointA(*f) )
@@ -389,13 +394,13 @@ matchA(int how, const charA *f, const charA *l)
 	if ( *h == '\0' )
 	  return FALSE;
       }
-  
+
       return FALSE;
     }
     case STR_MATCH_WORD:
     { const charA *h;
       const charA *f0 = f;
-  
+
       for(h=l; *h; h = nextwordA(h))
       { for( l=h,f=f0; *l && *f; l++, f++ )
 	{ if ( cmp_pointA(*l) != cmp_pointA(*f) )
@@ -408,7 +413,7 @@ matchA(int how, const charA *f, const charA *l)
 	if ( *l == '\0' )
 	  return FALSE;
       }
-  
+
       return FALSE;
     }
     case STR_MATCH_LIKE:		/* SeRQL like: * --> wildcart */
@@ -449,7 +454,7 @@ matchA(int how, const charA *f, const charA *l)
       if ( *l == '\0' && (*f == '\0' ||
 			 (*f == '*' && f[1] == '\0')) )
 	return TRUE;
-  
+
 retry_like:
       if ( chn > 0 )
       { chn--;
@@ -488,7 +493,7 @@ match_atoms(int how, atom_t search, atom_t label)
   if ( !get_atom_text(label, &l) ||
        !get_atom_text(search, &f) )
     return FALSE;			/* error? */
-  
+
   if ( f.length == 0 )
     return TRUE;
 
@@ -507,7 +512,7 @@ match_atoms(int how, atom_t search, atom_t label)
 
         return TRUE;
       }
-  
+
       return FALSE;
     }
     case STR_MATCH_PREFIX:
@@ -520,7 +525,7 @@ match_atoms(int how, atom_t search, atom_t label)
 	}
 
 	return TRUE;
-      }  
+      }
 
       return FALSE;
     }
@@ -538,7 +543,7 @@ match_atoms(int how, atom_t search, atom_t label)
 	snext:;
 	}
       }
-  
+
       return FALSE;
     }
     case STR_MATCH_WORD:
@@ -556,7 +561,7 @@ match_atoms(int how, atom_t search, atom_t label)
 	wnext:;
 	}
       }
-  
+
       return FALSE;
     }
     case STR_MATCH_LIKE:		/* SeRQL like: * --> wildcart */
@@ -600,7 +605,7 @@ match_atoms(int how, atom_t search, atom_t label)
       if ( il == l.length && (ip == f.length ||
 			      (fetch(&f,ip) == '*' && ip+1 == f.length)) )
 	return TRUE;
-  
+
 retry_like:
       if ( chn > 0 )
       { chn--;
@@ -616,3 +621,116 @@ retry_like:
       return FALSE;
   }
 }
+
+
+		 /*******************************
+		 *	  LANGUAGE MATCH	*
+		 *******************************/
+
+typedef struct lang_choice
+{ int langp;				/* points after - */
+  int patp;				/* points after *- */
+} lang_choice;
+
+#define MAX_CHOICES 10			/* Max number of stars */
+
+typedef struct
+{ int il, ip;
+  text l, p;
+  lang_choice choicepoints[MAX_CHOICES];
+  int choice_count;
+} lang_state;
+
+
+static int
+create_chp(lang_state *s)
+{ if ( s->choice_count < MAX_CHOICES )
+  { lang_choice *cp = &s->choicepoints[s->choice_count];
+
+    cp->langp = s->il;
+    cp->patp = s->ip+2;
+    s->choice_count++;
+
+    return TRUE;
+  }
+
+  return FALSE;
+}
+
+
+static int
+next_choice(lang_state *s)
+{ for ( ; s->choice_count > 0; s->choice_count-- )
+  { lang_choice *cp = &s->choicepoints[s->choice_count-1];
+    int il = cp->langp;
+
+    for(; il<s->l.length; il++)
+    { if ( fetch(&s->l, il) == '-' )
+      { cp->langp = s->il = il+1;
+	s->ip = cp->patp;
+	return TRUE;
+      }
+    }
+  }
+
+  return FALSE;
+}
+
+
+static atom_t ATOM_;
+static atom_t ATOM_star;
+
+int
+atom_lang_matches(atom_t lang, atom_t pattern)
+{ lang_state s = {0};
+  int cl, cp;
+
+  if ( lang == pattern )		/* exact match */
+    return TRUE;
+
+  if ( !ATOM_ )
+  { ATOM_ = PL_new_atom("");
+    ATOM_star = PL_new_atom("*");
+  }
+
+  if ( lang == ATOM_ )			/* no language */
+    return FALSE;
+  if ( pattern == ATOM_star )		/* Everything matches "*" */
+    return TRUE;
+
+  if ( !get_atom_text(lang, &s.l) ||
+       !get_atom_text(pattern, &s.p) )
+    return FALSE;			/* exception? */
+
+  s.il=0; s.ip=0;
+  for(;; s.ip++, s.il++)
+  { if ( s.ip == s.p.length )
+      return TRUE;
+    if ( s.il == s.l.length )
+    { if ( fetch(&s.p, s.ip) == '*' )
+	return TRUE;
+      if ( !next_choice(&s) )
+	return FALSE;
+    }
+
+    cl = fetch(&s.l, s.il);
+    cp = fetch(&s.p, s.ip);
+    if ( cl == cp )
+      continue;
+    if ( sort_point(cl)>>8 == sort_point(cp)>>8 )
+      continue;
+
+    if ( cp == '*' )
+    { if ( s.ip+1 == s.p.length )
+	return TRUE;
+      if ( (s.ip == 0 || fetch(&s.p, s.ip-1) == '-') &&
+	   fetch(&s.p, s.ip+1) == '-' )
+      { if ( !create_chp(&s) )
+	  return FALSE;
+      }
+    }
+
+    if ( !next_choice(&s) )
+      return FALSE;
+  }
+}
