@@ -1,11 +1,12 @@
-/*  $Id: rdf_edit.pl,v 1.48 2007/07/09 09:23:51 jan Exp $
+/*  $Id$
 
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        wielemak@science.uva.nl
+    E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2007, University of Amsterdam
+    Copyright (C): 1985-2011, University of Amsterdam
+			      VU University Amsterdam
 
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
@@ -19,7 +20,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 
     As a special exception, if you link this library with other files,
     compiled with a Free Software compiler, to produce an executable, this
@@ -54,7 +55,7 @@
 
 	    rdfe_set_watermark/1,	% +Name
 
-	    rdfe_undo/0,		% 
+	    rdfe_undo/0,		%
 	    rdfe_redo/0,
 	    rdfe_can_undo/1,		% -TID
 	    rdfe_can_redo/1,		% -TID
@@ -76,11 +77,14 @@
 :- use_module(library(broadcast)).
 :- use_module(library(lists)).
 :- use_module(library(debug)).
-:- use_module(library(url)).
+:- use_module(library(uri)).
 
 :- meta_predicate
-	rdfe_transaction(:),
-	rdfe_transaction(:, +).
+	rdfe_transaction(0),
+	rdfe_transaction(0, +).
+
+:- predicate_options(rdfe_load/2, 2,
+		     [pass_to(rdf_db:rdf_load/2, 2)]).
 
 :- dynamic
 	undo_log/5,			% TID, Action, Subj, Pred, Obj
@@ -113,7 +117,7 @@ This library provides a number of functions on top of the rdf_db module:
 	rdfe_delete(r),
 	rdfe_transaction(:),
 	rdfe_transaction(:, +).
-	
+
 
 		 /*******************************
 		 *     BASIC EDIT OPERATIONS	*
@@ -140,11 +144,11 @@ rdfe_retractall(Subject, Predicate, Object, PayLoad) :-
 	;   true
 	),
 	rdf_retractall(Subject, Predicate, Object, PayLoad).
-	
+
 %%	rdfe_update(+Subject, +Predicate, +Object, +Action)
-%	
+%
 %	Update an existing triple.  Possible actions are:
-%	
+%
 %%		subject(+Subject)
 %%		predicate(+Predicate)
 %%		object(+Object)
@@ -177,7 +181,7 @@ rdfe_update(Subject, Predicate, Object, PayLoad, Action) :-
 	journal(update(TID, Subject, Predicate, Object, PayLoad, Action)).
 
 %%	rdfe_delete(+Subject)
-%	
+%
 %	Delete a subject and all we know about it. This is a bit tricky.
 %	If we are involved in transitive relations, should we re-joint
 %	these in this module?
@@ -197,7 +201,7 @@ delete(Subject) :-
 
 %%	rdfe_load(+File) is det.
 %%	rdfe_load(+File, +Options) is det.
-%	
+%
 %	Load an RDF file and record this action including version information
 %	to facilitate reliable reload.
 
@@ -211,18 +215,22 @@ rdfe_load(File, Options) :-
 			   [ access(read),
 			     extensions([rdf,rdfs,owl,''])
 			   ], Path),
-	atom_concat('file://', Path, BaseURI),
 	rdf_load(Path,
-		 [ base_uri(BaseURI),
-		   result(Action, Triples, MD5)
+		 [ graph(Graph),
+		   modified(Modified)
 		 | Options
 		 ]),
-	(   Action == none		% load, reload, none
+	(   Modified == not_modified
 	->  true
 	;   absolute_file_name('.', PWD),
 	    size_file(Path, Size),
-	    time_file(Path, Modified),
-	    SecTime is round(Modified),
+	    (	Modified = last_modified(Stamp)
+	    ->	true
+	    ;	time_file(Path, Stamp)
+	    ),
+	    SecTime is round(Stamp),
+	    rdf_statistics(triples_by_file(Graph, Triples)),
+	    rdf_md5(Graph, MD5),
 	    assert_action(TID, load_file(Path), -, -, -),
 	    journal(rdf_load(TID,
 			     Path,
@@ -245,7 +253,7 @@ rdfe_unload(Path) :-
 
 
 %%	ensure_snapshot(+Path)
-%	
+%
 %	Ensure we have a snapshot of Path if we are making a journal, so
 %	we can always reload the snapshot to ensure exactly the same
 %	state.
@@ -271,7 +279,7 @@ ensure_snapshot(_).
 
 
 %%	load_snapshot(+Source, +Path)
-%	
+%
 %	Load triples from the given snapshot   file. One of the troubles
 %	is the time-stamp to avoid rdf_make/0   from reloading the file.
 %	for the time being we use 1e12, which   is  a lot further in the
@@ -292,12 +300,12 @@ load_snapshot(Source, Path) :-
 
 
 %%	snapshot_file(+Path, +MD5, +Access, -File)
-%	
+%
 %	Find existing snapsnot file or location to save a new one.
 
 snapshot_file(Path, MD5, Options, SnapShot) :-
 	file_base_name(Path, Base),
-	concat_atom([Base, @, MD5], File),
+	atomic_list_concat([Base, @, MD5], File),
 	absolute_file_name(snapshot(File),
 			   [ extensions([trp])
 			   | Options
@@ -306,7 +314,7 @@ snapshot_file(Path, MD5, Options, SnapShot) :-
 
 
 %%	rdfe_snapshot_file(-File)
-%	
+%
 %	Enumerate the MD5 snapshot files required to restore the current
 %	journal file. Using this  call  we   can  write  a  routine that
 %	packages the journal file with all required snapshots to restore
@@ -326,7 +334,7 @@ rdfe_snapshot_file(File) :-
 	system_ns/2.
 
 %%	rdfe_register_ns(Id, URI)
-%	
+%
 %	Encapsulation of rdf_register_ns(Id, URI)
 
 rdfe_register_ns(Id, URI) :-
@@ -348,7 +356,7 @@ rdfe_unregister_ns(Id, URI) :-
 	journal(ns(TID, unregister(Id, URI))).
 
 %	rdfe_register_ns/0
-%	
+%
 %	Reset namespaces to the state they where before usage of the
 %	rdf_edit layer.
 
@@ -370,7 +378,7 @@ save_system_ns :-
 		 *******************************/
 
 %%	rdfe_transaction(:Goal)
-%	
+%
 %	Run Goal, recording all modifications   as a single transaction.
 %	If  Goal  raises  an  exception  or    fails,  all  changes  are
 %	rolled-back.
@@ -393,9 +401,9 @@ rdfe_transaction(Goal, Name) :-
 	;   rdfe_rollback,
 	    fail
 	).
-	    
+
 %%	rdfe_begin_transaction
-%	
+%
 %	Start a transaction.  This is followed by either rdfe_end_transaction
 %	or rdfe_rollback.  Transactions may be nested.
 
@@ -433,7 +441,7 @@ rdfe_rollback :-
 	rollback(TID).
 
 %%	rollback(+TID)
-%	
+%
 %	This is the same as undo/1, but it must not record the undone
 %	actions as rollbacks cannot be `redone'.  Somehow there should
 %	be a cleaner way to distinguish between transactional operations
@@ -451,7 +459,7 @@ rollback(TID) :-
 	    )
 	;   true
 	).
-	
+
 rollback(assert(PayLoad), Subject, Predicate, Object) :- !,
 	rdf_retractall(Subject, Predicate, Object, PayLoad).
 rollback(retract(PayLoad), Subject, Predicate, Object) :- !,
@@ -465,7 +473,7 @@ assert_action(TID, Action, Subject, Predicate, Object) :-
 	asserta(undo_log(TID, Action, Subject, Predicate, Object)).
 
 %%	undo(+TID)
-%	
+%
 %	Undo a transaction as well as possible transactions nested into
 %	it.
 
@@ -481,7 +489,7 @@ undo(TID) :-
 	    )
 	;   true
 	).
-	
+
 undo(assert(PayLoad), Subject, Predicate, Object) :- !,
 	rdfe_retractall(Subject, Predicate, Object, PayLoad).
 undo(retract(PayLoad), Subject, Predicate, Object) :- !,
@@ -507,7 +515,7 @@ action(predicate(_)).
 action(object(_)).
 
 %%	rdfe_undo
-%	
+%
 %	Undo a (toplevel) transaction. More calls do further undo. The
 %	`Undone' actions are re-added to the undo log, so the user can
 %	redo them.  Fails if there are no more undo/redo transactions.
@@ -548,7 +556,7 @@ last_transaction(TID) :-
 	undo_log([TID|_], _, _, _, _), !.
 
 %%	rdfe_redo
-%	
+%
 %	Start a redo-session
 
 rdfe_redo :-
@@ -566,7 +574,7 @@ rdfe_redo :-
 
 %%	rdfe_can_redo(-TID) is semidet.
 %%	rdfe_can_undo(-TID) is semidet.
-%	
+%
 %	Check if we can undo and if so return the id of the transaction
 %	that will be un/re-done.  A subsequent call to rdfe_transaction_name
 %	can be used to give a hint in the UI.
@@ -587,7 +595,7 @@ rdfe_can_undo(Undo) :-			% start undo
 	find_previous_undo(TID, Undo).
 
 %%	rdfe_transaction_name(+TID, -Name)
-%	
+%
 %	Return name if the transaction is named.
 
 rdfe_transaction_name(TID, Name) :-
@@ -595,7 +603,7 @@ rdfe_transaction_name(TID, Name) :-
 	Name \== [].
 
 %%	rdfe_set_transaction_name(+Name)
-%	
+%
 %	Set name of the current transaction
 
 rdfe_set_transaction_name(Name) :-
@@ -603,7 +611,7 @@ rdfe_set_transaction_name(Name) :-
 	assert(transaction_name(TID, Name)).
 
 %%	rdfe_transaction_member(+TID, -Action)
-%	
+%
 %	Query actions inside a transaction to allow for quick update
 %	of visualisers.
 
@@ -636,19 +644,19 @@ user_transaction_member(Update, Subject, Predicate, Object,
 	rdf_current_default_file/2.	% file, all/fallback
 
 %%	rdfe_set_file_property(+File, +Options)
-%	
+%
 %	Set properties on the file.  Options is one of
-%	
+%
 %		* access(ro/rw)
 %		* default(all/fallback)
 
 rdfe_set_file_property(File, access(Access)) :- !,
-	to_url(File, URL),
+	to_uri(File, URL),
 	retractall(rdf_source_permission(URL, _)),
 	assert(rdf_source_permission(URL, Access)),
 	broadcast(rdf_file_property(URL, access(Access))).
 rdfe_set_file_property(File, default(Type)) :-
-	to_url(File, URL),
+	to_uri(File, URL),
 	rdfe_set_file_property(URL, access(rw)), % must be writeable
 	retractall(rdf_current_default_file(_,_)),
 	assert(rdf_current_default_file(URL, Type)),
@@ -657,12 +665,12 @@ rdfe_set_file_property(File, default(Type)) :-
 
 %%	rdfe_get_file_property(+FileOrURL, ?Option).
 %%	rdfe_get_file_property(-URL, ?Option).
-%	
+%
 %	Fetch file properties set with rdfe_set_file_property/2.
 
 rdfe_get_file_property(FileOrURL, access(Access)) :-
 	(   ground(FileOrURL)
-	->  to_url(FileOrURL, URL)
+	->  to_uri(FileOrURL, URL)
 	;   rdf_source(_DB, URL),
 	    FileOrURL = URL
 	),
@@ -676,7 +684,7 @@ rdfe_get_file_property(FileOrURL, access(Access)) :-
 	).
 rdfe_get_file_property(FileOrURL, default(Default)) :-
 	ground(FileOrURL),
-	to_url(FileOrURL, URL),
+	to_uri(FileOrURL, URL),
 	(   rdf_current_default_file(URL, Default)
 	->  true
 	;   FileOrURL = user,
@@ -691,7 +699,7 @@ rdfe_get_file_property(URL, default(Default)) :-
 
 
 %%	check_file_protection(-Error)
-%	
+%
 %	Check modification of all protected files
 
 check_file_protection(Error) :-
@@ -702,24 +710,24 @@ check_file_protection(Error) :-
 	).
 
 
-%%	to_url(+Spec, -URL) is det.
+%%	to_uri(+Spec, -URL) is det.
 %
 %	Convert a specification into a URL.
 
-to_url(URL, URL) :-
-	atom(URL),
-	sub_atom(URL, B, _, _, '://'),
-	sub_atom(URL, 0, B, _, Protocol),
-	url_protocol(Protocol), !.
-to_url(File, URL) :-
-	file_name_to_url(File, URL).
+to_uri(URL, URL) :-
+	uri_components(URL, Components),
+	uri_data(scheme, Components, Scheme),
+	nonvar(Scheme),
+	uri_scheme(Scheme), !.
+to_uri(File, URL) :-
+	uri_file_name(URL, File).
 
 
-url_protocol(file).
-url_protocol(http).
-url_protocol(https).
-url_protocol(ftp).
-url_protocol(ftps).
+uri_scheme(file).
+uri_scheme(http).
+uri_scheme(https).
+uri_scheme(ftp).
+uri_scheme(ftps).
 
 
 		 /*******************************
@@ -727,7 +735,7 @@ url_protocol(ftps).
 		 *******************************/
 
 %%	rdfe_is_modified(?Source)
-%	
+%
 %	True if facts have been added, deleted or updated that have
 %	Source as `payload'.
 
@@ -765,7 +773,7 @@ rdfe_clear_modified(DB) :-
 		 *******************************/
 
 %%	rdfe_set_watermark(Name)
-%	
+%
 %	Create a watermark for undo and replay journal upto this point.
 %	The rest of the logic needs to be written later.
 
@@ -780,7 +788,7 @@ rdfe_set_watermark(Name) :-
 		 *******************************/
 
 %%	rdfe_reset
-%	
+%
 %	Clear database, undo, namespaces and journalling info.
 
 rdfe_reset :-
@@ -822,14 +830,14 @@ journal_version(1).
 
 %%	rdfe_open_journal(+File, +Mode) is det.
 %
-%	Open a journal writing to File in Mode.  Mode is one of 
-%	
+%	Open a journal writing to File in Mode.  Mode is one of
+%
 %		* read
 %		Open and replay the journal
-%		
+%
 %		* write
 %		Delete current journal and create a fresh one
-%		
+%
 %		* append
 %		Read and replay the existing journal and append new
 %		modifications to the File.
@@ -884,7 +892,7 @@ journal_open(Type, Time) :-
 
 journal_comment(start, Time) :-
 	journal(_, _, Stream),
-	convert_time(Time, String),
+	format_time(string(String), '%+', Time),
 	format(Stream,
 	       '/* Triple20 Journal File\n\n   \
 	       Created: ~w\n   \
@@ -893,7 +901,7 @@ journal_comment(start, Time) :-
 	       */~n~n', [String]).
 journal_comment(resume, Time) :-
 	journal(_, _, Stream),
-	convert_time(Time, String),
+	format_time(string(String), '%+', Time),
 	format(Stream,
 	       '\n\
 	       /* Resumed: ~w\n\
@@ -916,7 +924,7 @@ rdfe_close_journal :-
 	).
 
 %%	rdfe_current_journal(-Path)
-%	
+%
 %	Query the currently open journal
 
 rdfe_current_journal(Path) :-
@@ -946,7 +954,7 @@ write_journal(Term, Stream) :-
 
 
 %%	rdfe_replay_journal(+File)
-%	
+%
 %	Replay a journal file. For now  this   is  our cheap way to deal
 %	with save/load. Future versions may be  more clever when dealing
 %	with the version information stored in the journal.
@@ -1018,7 +1026,7 @@ replay_actions([H|T]) :-
 
 
 %%	replay_action(+Action)
-%	
+%
 %	Replay actions from the journal. Tricky is rdf_load/3. It should
 %	reload the file in the state it  was   in  at  the moment it was
 %	created. For now this has been hacked  for files that were empry
@@ -1073,7 +1081,7 @@ find_file(File, Options, Path) :-
 	exists_file(Path), !.
 
 %%	make_path(+File, +PWD, -Path)
-%	
+%
 %	Return location of File relative to PWD, Parent of PWD, etc. (TBD)
 
 make_path(File, PWD, Path) :-
