@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 
     As a special exception, if you link this library with other files,
     compiled with a Free Software compiler, to produce an executable, this
@@ -180,7 +180,7 @@ join_expansions_by_tag([H|T0], Tag, T, [V0|VT]) :-
 	join_expansions_by_tag(T0, Tag, T, VT).
 join_expansions_by_tag(L, _, L, []).
 
-lookup(false, _, [], []) :- !.
+lookup(@(false), _, [], []) :- !.
 lookup(or(H0,T0), Map, [CH|CT], [H|T]) :- !,
 	lookup(H0, Map, CH, H),
 	lookup(T0, Map, CT, T).
@@ -188,16 +188,19 @@ lookup(H0, Map, [C], [H]) :-
 	lookup1(H0, Map, C, H).
 
 lookup1(Conj, Map, Cond, Literals) :-
-	phrase(conj_to_list(Conj), List),
+	phrase(conj_to_list(Conj), List), !,
 	rdf_find_literal_map(Map, List, Literals),
 	(   Literals \== []
 	->  phrase(conj_to_cond(Conj), Cond)
 	;   Cond = []
 	).
+lookup1(_, _, _, []).
 
 conj_to_list(and(A,B)) --> !,
 	conj_to_list(A),
 	conj_to_list(B).
+conj_to_list(@(false)) --> !,
+	{fail}.
 conj_to_list(Tagged) -->
 	{ untag(Tagged, L) }, !,
 	[L].
@@ -246,12 +249,14 @@ expand_fuzzy(case(String), Or) :- !,
 	token_index(Map),
 	rdf_keys_in_literal_map(Map, case(String), Tokens),
 	list_to_or(Tokens, case(String), Or).
-expand_fuzzy(or(A0, B0), or(A,B)) :- !,
+expand_fuzzy(or(A0, B0), E) :- !,
 	expand_fuzzy(A0, A),
-	expand_fuzzy(B0, B).
-expand_fuzzy(and(A0, B0), and(A,B)) :- !,
+	expand_fuzzy(B0, B),
+	simplify(or(A,B), E).
+expand_fuzzy(and(A0, B0), E) :- !,
 	expand_fuzzy(A0, A),
-	expand_fuzzy(B0, B).
+	expand_fuzzy(B0, B),
+	simplify(and(A,B), E).
 expand_fuzzy(not(A0), not(A)) :- !,
 	expand_fuzzy(A0, A).
 expand_fuzzy(between(Low, High), Or) :- !,
@@ -271,8 +276,17 @@ expand_fuzzy(Token, Token) :-
 expand_fuzzy(Token, _) :-
 	throw(error(type_error(Token, boolean_expression), _)).
 
+simplify(Expr0, Expr) :-
+	simple(Expr0, Expr), !.
+simplify(Expr, Expr).
 
-list_to_or([], _, false) :- !.
+simple(and(@(false), _), @(false)).
+simple(and(_, @(false)), @(false)).
+simple(or(@(false), X), X).
+simple(or(X, @(false)), X).
+
+
+list_to_or([], _, @(false)) :- !.
 list_to_or([X], How, One) :- !,
 	tag(How, X, One).
 list_to_or([H0|T0], How, or(H, T)) :-
@@ -402,11 +416,7 @@ threaded_literal_index(_) :-
 
 create_index_threads(N, Q, [Id|T]) :-
 	N > 0, !,
-	thread_create(index_worker(Q), Id,
-		      [ local(1000),
-			global(1000),
-			trail(1000)
-		      ]),
+	thread_create(index_worker(Q), Id, []),
 	N2 is N - 1,
 	create_index_threads(N2, Q, T).
 create_index_threads(_, _, []) :- !.
@@ -452,10 +462,7 @@ create_update_literal_thread(Threads) :-
 	forall(between(1, Threads, N),
 	       (   atom_concat(rdf_literal_monitor_, N, Alias),
 		   thread_create(monitor_literals, _,
-				 [ alias(Alias),
-				   local(1000),
-				   global(1000),
-				   trail(1000)
+				 [ alias(Alias)
 				 ])
 	       )).
 
@@ -539,7 +546,7 @@ del_tokens([H|T], Literal, Map) :-
 %	generally domain dependent.
 
 rdf_tokenize_literal(Literal, Tokens) :-
-	tokenization(Literal, Tokens), !. 		% Hook
+	tokenization(Literal, Tokens), !.		% Hook
 rdf_tokenize_literal(Literal, Tokens) :-
 	text_of(Literal, Text),
 	atom(Text),
