@@ -17,7 +17,7 @@
 
     You should have received a copy of the GNU General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 
     As a special exception, if you link this library with other files,
     compiled with a Free Software compiler, to produce an executable, this
@@ -36,26 +36,27 @@
 	    tipc_accept/3,		% +Master, -Slave, -PeerName
 	    tipc_open_socket/3,		% +Socket, -Read, -Write
 	    tipc_get_name/2,		% +Socket, -Address
+	    tipc_get_peer_name/2,	% +Socket, -Address
 	    tipc_connect/2,		% +Socket, +Address
 	    tipc_receive/4,		% +Socket, -Data, -Sender, +Options
 	    tipc_send/4,  		% +Socket, +Data, +Receiver, +Options
 %	    tipc_subscribe/5, % +Socket, +Address, +Timeout, +Filter,
 %	    +Usr_handle
 
-%	    tipc_event/3,		% +Data, -Event, -Residue
+%	    tipc_receive_subscr_event/2, % +Socket, -Event
 	    tipc_canonical_address/2,	% -Address, +port_id/2
             tipc_service_probe/1,	% ?Address
 	    tipc_service_probe/2,	% ?Address, ?PortId
 	    tipc_service_port_monitor/2, % +Address, :Goal
 	    tipc_service_port_monitor/3, % +Address, :Goal, +Timeout
             tipc_service_exists/1,	% +Address
-	    tipc_service_exists/2       % +Address +Timeout
+	    tipc_service_exists/2,       % +Address +Timeout
+	    tipc_initialize/0            %
 	  ]).
 
 :- use_module(library(shlib)).
 
-:- initialization
-   load_foreign_library(foreign(tipc), install_tipc).
+:- use_foreign_library(foreign(tipc)).
 
 /** <module> TIPC Sockets
 
@@ -139,7 +140,7 @@ pattern. For an overview, please see: tipc_overview.txt.
 %	 transparently to the subscriber  with   every  notification. An
 %	 application that uses this service must   be  prepared to parse
 %	 event notifications received from  the   TIPC  topology server.
-%	 tipc_event/3, is provided for this purpose.
+%	 tipc_receive_subscr_event/4, is provided for this purpose.
 %
 %	 It is possible to have  multiple   subscriptions  active on the
 %	 same socket any one time, making periodic subscription renewal
@@ -166,11 +167,14 @@ pattern. For an overview, please see: tipc_overview.txt.
 %	 @param Timeout an integer that specifies   the  duration of the
 %	 subscription in milliseconds. A  duration   of  -1, specifies a
 %	 subscription of infinite duration.  @param   Filter  the  event
-%	 filter bit map. @param UserHandle an   eight-byte  code that is
-%	 passed transparently to the user  with each event notification.
-%	 @error socket_error('Invalid argument') is thrown under several
-%	 circumstances, the most obscure  being   that  only  name_seq/3
-%	 addresses are permissible. To  interrogate   a  specific server
+%	 filter bit map.
+%
+%	 @param UserHandle an eight-byte code that is passed
+%	 transparently to the user with each event notification.
+%
+%        @error	 socket_error('Invalid argument') is thrown under several
+%	 circumstances, the most obscure being that only name_seq/3
+%	 addresses are permissible. To interrogate a specific server
 %	 instance X, of type Y, use: name_seq(Y, X, X).
 
 %%	 tipc_open_socket(+SocketId, -InStream, -OutStream) is det.
@@ -221,16 +225,17 @@ pattern. For an overview, please see: tipc_overview.txt.
 %	 port-id is sufficient for this role.   And server sockets (e.g.
 %	 those that are bound to name/3   or  name_seq/3, addresses) may
 %	 not act as clients. That is, they may not originate connections
-%	 from the socket. Please see the   TIPC programmers's guide for
-%	 other restrictions.
+%	 from the socket using tipc_connect/2. Servers however, may
+%	 originate datagrams from bound sockets using tipc_send/4.
+%	 Please see the TIPC programmers's guide for other restrictions.
 
 %%	 tipc_listen(+Socket,+Backlog) is det.
 %
-%	 Listens  for  incoming  requests    for   connections.  Backlog
-%	 indicates how many pending  connection   requests  are allowed.
-%	 Pending requests are requests  that   are  not yet acknowledged
-%	 using tipc_accept/3. If the indicated   number is exceeded, the
-%	 requesting  client  will  be  signalled  that  the  service  is
+%	 Listens for incoming requests for connections. Backlog
+%	 indicates how many pending connection requests are allowed.
+%	 Pending requests are requests that are not yet acknowledged
+%	 using tipc_accept/3. If the indicated number is exceeded, the
+%	 requesting client will be signalled that the service is
 %	 currently not available. A suggested default value is 5.
 
 %%	 tipc_accept(+Socket, -Slave, -Peer) is det.
@@ -246,11 +251,28 @@ pattern. For an overview, please see: tipc_overview.txt.
 %	 socket to a given TIPC_address.   After  successful completion,
 %	 tipc_open_socket/3 may be used  to   create  I/O-Streams to the
 %	 remote socket.
+%
+%	 @throws socket_error('Connection refused'), if there are
+%	 no servers bound to the specified address.
+%
+%	 @throws socket_error('Connection timed out'), if no server that
+%	 is bound to the specified address accepts the connect
+%	 request within the specified time limit. See also
+%	 tipc_setopt/2.
 
 %%	 tipc_get_name(+Socket, -TIPC_address) is det.
 %
 %	 Unifies TIPC_address with the port-id assigned to the socket.
 
+%%	 tipc_get_peer_name(+Socket, -TIPC_address) is det.
+%
+%	 Unifies TIPC_address with the port-id assigned to the socket
+%	 that this socket is connected to.
+%
+%	 @throws socket_error('Transport endpoint is not connected'), if
+%	 an attempt is made to obtain a peer's name of an unconnected
+%	 socket.
+
 %%	 tipc_setopt(+Socket,+Option) is det.
 %
 %	 Sets options on the socket. Defined options are:
@@ -309,7 +331,7 @@ pattern. For an overview, please see: tipc_overview.txt.
 
 %%	 tipc_send(+Socket, +Data, +To, +Options) is det.
 %
-%	 Sends a TIPC datagram to one or more destinations. Like its UDP
+%	 sends a TIPC datagram to one or more destinations. Like its UDP
 %	 counterpart, Data is a string, atom  or code-list providing the
 %	 data to be sent.  To  is   a  name/3,  name_seq/3, or port_id/2
 %	 address structure. See tipc_overview.txt, for more information
@@ -339,102 +361,46 @@ pattern. For an overview, please see: tipc_overview.txt.
 % %	 tipc_event(+Data, -Event, -Residue) is det.
 %
 %	 Parses event notifications received from   the  topology server
-%	 into Prolog structures. While using the "publish and subscribe"
-%	 regime, the topology  server  will   notify  the  subscriber of
-%	 certain events related to the publisher.  These are received on
-%	 the socket as  user  data.   tipc_event/3,  parses  a  received
-%	 message into a more comfortable   Prolog structure, Event. Data
-%	 is a list of codes representing a  message received on a socket
-%	 where a tipc_subscribe/5, has been  issued previously. Event is
-%	 the  parsed  event  structure,  if  an  event  was  recognized,
-%	 otherwise Residue is unified with Data and the atom, =none=, is
-%	 returned in the Event argument. Residue  is the remaining data,
-%	 after the notification has been  stripped out. tipc_event/3, is
-%	 implemented using Prolog DCG.
-%
-%	 Event is a structure of the form:
-%
-%        $ tipc_event(-EventType, -NameSeq, -PortId, -Subscription) :
-%	   * EventType, is one of: =published=, =withdrawn=, or
-%	   =subscr_timeout=.
-%
-%          * NameSeq is the newly published/withdrawn
-%	   name-sequence address of the publisher.
-%
-%          * PortId is the port-id
-%	   address of an instance, though not necessarily the only
-%	   instance, of the service. A =|port_id(0,0)|=, appears in
-%	   "withdraw" events. It means that the address is no longer
-%	   valid. A client may use either the NameSeq address, or the
-%	   PortId address to connect to the service. Using the NameSeq
-%	   address causes the client to be connected to an unspecified
-%	   instance of the service. Using the PortId causes the client
-%	   to be connected to a specific instance of the service.
-%
-%        Subscription, is a structure of the form:
-%
-%	 $ subscr(-OriginalNameSeq, -Timeout, -Filter,-UserHandle) :
-%	 This information is identical to that specified in the
-%	 original subscription.
-%
-%	 _|Please note that this predicate should be considered
-%	 private. Its use in user programs is strongly discouraged.|_
+%	into Prolog structures. This  predicate   has  been  permanently
+%	removed.
 %
 
+% %	 tipc_receive_subscr_event(+Socket, -Event) is semidet.
+%
+%	Receives and parses event notifications received from the
+%	TIPC Topology Server.
+%
+%	_|Please note that this predicate is considered private. Its use
+%	in user programs is strongly discouraged. See the
+%	tipc-service predicates for alternatives.|_
+%
+%	@param Socket is a TIPC socket that has been previously
+%	connected to the topology server using tipc_connect/2, and for
+%	which a subscription has been sent using tipc_subscribe/5,
+%	above.
+%
+%	@param Event is the structure: =|tipc_event(-Action, -Subscr,
+%	-Found, -Port_id)|=. On subscription timeout, the
+%	atom, =subscr_timeout= is returned. Subscr is the name-sequence
+%	address of the original subscription.
+%
+%       Action is one of:
+%
+%	  $ published:
+%         The socket specified by Port_id has been bound to
+%	the name_seq/3 address specified in Found.
+%
+%	  $ withdrawn:
+%         The socket specified by Port_id has been
+%	  unbound from the name_seq/3 address specified in Found. See
+%	the no_scope/1 option of tipc_bind/3.
+%
 
-tipc_event(Data, Event, Residue) :-
-	phrase(struct_tipc_event(Event), Data, Residue), !.
-
-tipc_event(Data, none, Data).
-
-/*
-*   DCG parser for TIPC topology server events
-*/
-
-integerAsU32(In, Out) :-
-        nonvar(In), !,
-	(   ( In < 0) -> Out is In + 0x100000000; Out is In).
-
-integerAsU32(In, Out) :-
-        nonvar(Out), !,
-	(   ( Out > 0x7fffffff) -> In is Out - 0x100000000; In is Out).
-
-u32(X) -->  [X,0,0,0], { between(0,255, X) }, !.
-u32(-1) --> [255,255,255,255], !.
-u32(Int) -->
-	[X0, X1, X2, X3],
-	{ X is ((((X3 * 256) + X2) * 256 + X1) * 256) + X0, integerAsU32(Int, X) }.
-
-chars(0, []) --> [], !.
-chars(N, [A | B]) --> [A], {succ(N1, N)}, chars(N1, B).
-
-struct_tipc_name_seq(name_seq(Type, Lower, Upper)) -->
-	u32(Type),
-	u32(Lower),
-	u32(Upper).
-
-struct_tipc_portid(port_id(Ref, Node)) -->
-	u32(Ref),
-	u32(Node).
-
-struct_tipc_subscr(subscr(Seq, Timeout, Filter, UserHandle)) -->
-	struct_tipc_name_seq(Seq),
-	u32(Timeout),
-	u32(Filter),
-	chars(8, UserHandle).
-
-tipc_event_type(published) --> u32(1), !.
-tipc_event_type(withdrawn) --> u32(2), !.
-tipc_event_type(subscr_timeout) --> u32(3), !.
-
-struct_tipc_event(tipc_event(Event, name_seq(Type, Found_lower, Found_upper), PortId, Subscr)) -->
-	tipc_event_type(Event),
-	u32(Found_lower),
-	u32(Found_upper),
-	struct_tipc_portid(PortId),
-	struct_tipc_subscr(Subscr),
-	{ Subscr = subscr(name_seq(Type, _, _), _, _, _) }.
+:- multifile
+	prolog:message/3.
 
+prolog:message(error(socket_error(Message), _)) -->
+	[ 'Socket error: ~w'-[Message] ].
 
 %%	tipc_canonical_address(-CanonicalAddress, +PortId) is det.
 %
@@ -445,6 +411,14 @@ struct_tipc_event(tipc_event(Event, name_seq(Type, Found_lower, Found_upper), Po
 %	It is provided for debugging an printing purposes only. The
 %	canonical address is not used for any other purpose.
 
+integerAsU32(In, Out) :-
+	nonvar(In),
+	(   In < 0 -> Out is In + 0x100000000; Out is In).
+
+integerAsU32(In, Out) :-
+	nonvar(Out),
+	(   Out > 0x7fffffff -> In is Out - 0x100000000; In is Out).
+
 tipc_canonical_address(tipc_address(Z,C,N, Ref1), port_id(Ref, Node)) :-
        integerAsU32(Ref, Ref1),
        integerAsU32(Node, X),
@@ -477,10 +451,6 @@ user:portray(port_id(Ref, Node)) :-
 %	also permissible.
 %
 
-se_dispatch(NameSeq, Service, tipc_event(published, Service, _, subscr(NameSeq, _, _, _))) :- !.
-
-se_dispatch(_NameSeq, _Service, tipc_event(subscr_timeout, _, _, _)).
-
 tipc_address(name(T, I, 0), name_seq(T, I, I)).
 tipc_address(name_seq(T, L, U), name_seq(T, L, U)).
 tipc_address(mcast(T, L, U), name_seq(T, L, U)).
@@ -493,14 +463,16 @@ tipc_service_exists(Address) :-
 tipc_service_exists(Address, Timeout) :-
 	tipc_address(Address, NameSeq),!,
 	ITime is integer(Timeout * 1000),
-	try_finally(tipc_socket(S, seqpacket), tipc_close_socket(S)),
+	try_finally(tipc_socket(S, seqpacket),
+		    tipc_close_socket(S)),
 	tipc_connect(S, name(1,1,0)),   % connect to the topology server
-	tipc_subscribe(S, NameSeq, ITime, 2, "prolog"),
+	tipc_subscribe(S, NameSeq, ITime, 2, "swipl"),
 	repeat,
-	    tipc_receive(S, Data, _From, [as(codes)]),
-	    tipc_event(Data, Event, []),
-	    se_dispatch(NameSeq, Service, Event),
-	!, ground(Service).
+	    tipc_receive_subscr_event(S, Data),
+            (	Data == subscr_timeout
+	    -> (!, fail)
+	    ;  Data = tipc_event(published, NameSeq, _FoundSeq, _Port_id)),
+	!.
 
 %%	tipc_service_probe(?Address) is nondet.
 %%	tipc_service_probe(?Address, ?PortId) is nondet.
@@ -524,7 +496,7 @@ tipc_service_probe(Address) :-
 	NameSeq = name_seq(Type, Lower, Upper),
 	try_finally(tipc_socket(S, seqpacket), tipc_close_socket(S)),
 	tipc_connect(S, name(1,1,0)),   % connect to the topology server
-	tipc_subscribe(S, name_seq(Type, 0, -1), 0, 2, "prolog"),  % look for everything
+	tipc_subscribe(S, name_seq(Type, 0, 4294967295), 0, 2, "swipl"),  % look for everything
 	sp_collect(S, Members),
 	!, member([NameSeq, _], Members).
 
@@ -534,20 +506,18 @@ tipc_service_probe(Address, PortId) :-
 	NameSeq = name_seq(Type, Lower, Upper),
 	try_finally(tipc_socket(S, seqpacket), tipc_close_socket(S)),
 	tipc_connect(S, name(1,1,0)),   % connect to the topology server
-	tipc_subscribe(S, name_seq(Type, 0, -1), 0, 1, "prolog"),  % look for everything
+	tipc_subscribe(S, name_seq(Type, 0, 4294967295), 0, 1, "swipl"),  % look for everything
 	sp_collect(S, Members),
 	!, member([NameSeq, PortId], Members).
 
-
 sp_collect(S, Members) :-
 	findall([NameSeq ,PortId],
 	    (
 	    repeat,
-	    tipc_receive(S, Data, _From, [as(codes)]),
-	    tipc_event(Data, tipc_event(Event, NameSeq, PortId, _), []),
-	(   (Event == published) ->
-	       true ;
-	       (Event == subscr_timeout -> (!, fail)); fail)
+	    tipc_receive_subscr_event(S, Data),
+	    (	Data == subscr_timeout
+	    -> (!, fail)
+	    ; Data = tipc_event(published, _Service, NameSeq, PortId))
 	    ), Members).
 
 %%	tipc_service_port_monitor(+Addresses, :Goal) is det.
@@ -565,9 +535,10 @@ sp_collect(S, Members) :-
 %	@param Addresses is a list of name/3 or name_seq/3 addresses
 %	for the services to be monitored.
 %
-%	@param Goal is a predicate with an arity of one, that will be
-%	called when a worker's publication status changes. The Goal is
-%	called exactly once per event with the structure:
+%	@param Goal is a predicate that will be called when a
+%	worker's publication status changes. The Goal
+%	is called exactly once per event with its the last argument
+%	unified with the structure:
 %
 %	$ published(-NameSeq, -PortId) : when the worker binds
 %	its socket to the address.
@@ -590,13 +561,16 @@ sp_collect(S, Members) :-
 %	operating in the background.
 %
 
-spm_dispatch(_Goal, tipc_event(subscr_timeout, _, _, _)) :- !.
+spm_dispatch(_Goal, subscr_timeout) :- !.
 
-spm_dispatch(Goal, tipc_event(Action, NameSeq, PortId, _)) :-
+spm_dispatch(Goal, tipc_event(Action, _Subscr, NameSeq, PortId)) :-
 	Event =.. [Action, NameSeq, PortId],
 	once(call(Goal, Event)), fail.
 
 %
+:- meta_predicate
+	tipc_service_port_monitor(+, :),
+	tipc_service_port_monitor(+, :, +).
 
 tipc_service_port_monitor(Address, Goal) :-
 	tipc_service_port_monitor(Address, Goal, 0.0), !.
@@ -613,15 +587,28 @@ tipc_service_port_monitor(Addresses, Goal, Timeout) :-
 	try_finally(tipc_socket(S, seqpacket), tipc_close_socket(S)),
 	tipc_connect(S, name(1,1,0)),   % connect to the topology server
 	forall(member(NameSeq, NameSeqs),
-	       tipc_subscribe(S, NameSeq, ITime, 1, "prolog")),
-	    repeat,
-	        tipc_receive(S, Data, _From, [as(codes)]),
-	        tipc_event(Data, Event, []),
-	        spm_dispatch(Goal, Event),
+	       tipc_subscribe(S, NameSeq, ITime, 1, "swipl")),
+	repeat,
+	tipc_receive_subscr_event(S, Data),
+	spm_dispatch(Goal, Data),
         !.
 
+%%     tipc_initialize is semidet.
+%
+%      causes the TIPC service and the TIPC stack to be initialized
+%      and made ready for service. An application must call this
+%      predicate as part of its initialization prior to any use of
+%      TIPC predicates. _|Please note the change of the API.|_
+%
+%      @throws socket_error('Address family not supported by protocol')
+%      if a TIPC server is not available on the current host.
+%
 
+tipc_initialize :-
+	with_mutex(tipc_mutex,
+		   forall(tipc:tipc_stack_initialize, true)).
 
+:- multifile tipc:tipc_stack_initialize/0.
 
-
-
+tipc:tipc_stack_initialize :-
+	tipc_service_exists(name(1,1,0)).
