@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 
     As a special exception, if you link this library with other files,
     compiled with a Free Software compiler, to produce an executable, this
@@ -52,7 +52,6 @@
 	   , append/3
 	   , atom_concat/3
 	   , between/3
-	   , genarg/3
 	   , maplist/3
 	   , sub_atom/5
 	   , push_operators/1
@@ -65,10 +64,12 @@
 	verbose/0,
 	recording/2.			% items recorded
 
-pce_ifhostproperty(prolog(swi),
-		   [ (:- index(attribute(1,1,0))),
-		     (:- use_module(library(quintus), [genarg/3]))
-		   ]).
+:- if(current_predicate(index/1)).
+:- index(attribute(1,1,0)).
+:- endif.
+:- if(exists_source(library(quintus))).
+:- use_module(library(quintus), [genarg/3]).
+:- endif.
 
 		 /*******************************
 		 *	     OPERATORS		*
@@ -100,16 +101,13 @@ push_compile_operators(M) :-
 pop_compile_operators :-
 	pop_operators.
 
-:- op(100, fx, @).			% standard XPCE operators
-:- op(150, yfx, ?).
-:- op(990, xfx, :=).
 :- push_compile_operators.
 
 %verbose.
 
 pce_term_expansion(In, Out) :-
 	pce_pre_expand(In, In0),
-	(   list(In0)
+	(   is_list(In0)
 	->  maplist(map_term_expand, In0, In1),
 	    flatten(In1, Out0),
 	    (	Out0 = [X]
@@ -119,9 +117,6 @@ pce_term_expansion(In, Out) :-
 	;   do_term_expand(In0, Out)
 	).
 
-list([]).
-list([_|_]).
-
 map_term_expand(X, X) :-
 	var(X), !.
 map_term_expand(X, Y) :-
@@ -129,6 +124,8 @@ map_term_expand(X, Y) :-
 map_term_expand(X, X).
 
 
+do_term_expand(end_of_file, _) :-
+	cleanup, !, fail.
 do_term_expand(In0, Out) :-
 	pce_expandable(In0),
 	(   do_expand(In0, Out0)
@@ -184,7 +181,7 @@ head_arg(_, A) :-
 
 pce_pre_expand(X, Y) :-
 	user:pce_pre_expansion_hook(X, X1), !,
-	(   list(X1)
+	(   is_list(X1)
 	->  maplist(do_pce_pre_expand, X1, Y)
 	;   do_pce_pre_expand(X1, Y)
 	).
@@ -291,8 +288,8 @@ do_expand((:- pce_end_class),
 	findall(R, retract(attribute(ClassName, classvar, R)),  Resources),
 	findall(D, retract(attribute(ClassName, directive, D)), Directs),
 	(   attribute(ClassName, extending, true)
-	->  MetaClass = '-',
-	    Super = '-',
+	->  MetaClass = (-),
+	    Super = (-),
 	    expand_term((:- initialization(pce_extended_class(ClassName))),
 			RegisterDecl)
 	;   retract(attribute(ClassName, super, Super)),
@@ -456,7 +453,7 @@ Importing the template (pce_use_class_template/1):
 template_clause((M:send_implementation(Id, Msg, R) :- Body),
 		[ (M:send_implementation(Tid, ClassMsg, R) :- ClassBody),
 		  (M:(send_implementation(Id, Msg, R) :-
-		  	send_implementation(Tid, IClassMsg, R)))
+			send_implementation(Tid, IClassMsg, R)))
 		]) :- !,
 	atom_concat('T-', Id, Tid),
 	Msg =.. Args,
@@ -468,7 +465,7 @@ template_clause((M:send_implementation(Id, Msg, R) :- Body),
 template_clause((M:get_implementation(Id, Msg, R, V) :- Body),
 		[ (M:get_implementation(Tid, ClassMsg, R, V) :- ClassBody),
 		  (M:(get_implementation(Id, Msg, R, V) :-
-		  	get_implementation(Tid, IClassMsg, R, V)))
+			get_implementation(Tid, IClassMsg, R, V)))
 		]) :- !,
 	atom_concat('T-', Id, Tid),
 	Msg =.. Args,
@@ -552,7 +549,7 @@ use_template_send_method(Template, pce_principal:Clause) :-
 	(   Clause = pce_lazy_send_method(Sel, ClassName, NewBinder),
 	    NewBinder =.. [Functor, NewId | RestBinder]
 	;   Clause = (send_implementation(NewId, Msg, R) :-
-		     	send_implementation(Tid, IClassMsg, R)),
+			send_implementation(Tid, IClassMsg, R)),
 	    attribute(ClassName, super, SuperClass), % TBD: pce_extend_class/1
 	    arg(2, Binder, Types),
 	    type_arity(Types, Arity),
@@ -574,7 +571,7 @@ use_template_get_method(Template, pce_principal:Clause) :-
 	(   Clause = pce_lazy_get_method(Sel, ClassName, NewBinder),
 	    NewBinder =.. [Functor, NewId | RestBinder]
 	;   Clause = (get_implementation(NewId, Msg, R, V) :-
-		     	get_implementation(Tid, IClassMsg, R, V)),
+			get_implementation(Tid, IClassMsg, R, V)),
 	    attribute(ClassName, super, SuperClass), % TBD: pce_extend_class/1
 	    arg(3, Binder, Types),
 	    type_arity(Types, Arity),
@@ -944,6 +941,21 @@ pce_compiling :-
 	compiling(_, _), !.
 
 
+		 /*******************************
+		 *	      CLEANUP		*
+		 *******************************/
+
+%	cleanup
+%
+%	Cleanup the compilation data. We should  probably give a warning
+%	when not under xref and there is data left.
+
+cleanup :-
+	source_location(Path, _),
+	forall(retract(compiling(Class, Path)),
+	       retractall(attribute(Class, _, _))).
+
+
 		 /*******************************
 		 *	      CHECKS		*
 		 *******************************/
@@ -993,11 +1005,11 @@ feedback(Term) :-
 		********************************/
 
 :- multifile
-	user:term_expansion/2.
+	system:term_expansion/2.
 :- dynamic
-	user:term_expansion/2.
+	system:term_expansion/2.
 
-user:term_expansion(A, B) :-
+system:term_expansion(A, B) :-
 	pce_term_expansion(A, B).
 
 :- pop_compile_operators.
