@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 
     As a special exception, if you link this library with other files,
     compiled with a Free Software compiler, to produce an executable, this
@@ -101,7 +101,7 @@ check_loaded_class(_).
 
 pce_realise_class(ClassName) :-
 	pce_class(ClassName, MetaClassName, SuperName, _, _, _),
-	MetaClassName \== -,
+	MetaClassName \== (-),
 	create_class(ClassName, MetaClassName, SuperName, Class), !,
 	resolve_method_message(Msg),
 	send(Class, resolve_method_message, Msg),
