@@ -3,9 +3,10 @@
     Part of XPCE --- The SWI-Prolog GUI toolkit
 
     Author:        Jan Wielemaker and Anjo Anjewierden
-    E-mail:        jan@swi.psy.uva.nl
-    WWW:           http://www.swi.psy.uva.nl/projects/xpce/
-    Copyright (C): 1985-2002, University of Amsterdam
+    E-mail:        J.Wielemaker@cs.vu.nl
+    WWW:           http://www.swi-prolog.org/projects/xpce/
+    Copyright (C): 1985-2011, University of Amsterdam
+			      VU University Amsterdam
 
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
@@ -19,7 +20,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 
     As a special exception, if you link this library with other files,
     compiled with a Free Software compiler, to produce an executable, this
@@ -31,10 +32,10 @@
 
 :- module(emacs_application, []).
 :- use_module(library(pce)).
-:- (   current_prolog_flag(windows, true)
-   ->  use_module(dde_server)
-   ;   true
-   ).
+:- use_module(library(pce_history)).
+:- if(current_prolog_flag(windows, true)).
+:- use_module(dde_server).
+:- endif.
 :- require([ ignore/1
 	   , pce_help_file/2
 	   ]).
@@ -46,8 +47,10 @@
 :- pce_begin_class(emacs, application,
 		   "PceEmacs main object").
 
-variable(buffer_list,	dict,	get, "List of buffers maintained").
-variable(exit_message,	message,get, "Registered exit message").
+variable(buffer_list,	dict,	 get, "List of buffers maintained").
+variable(exit_message,	message, get, "Registered exit message").
+variable(history,	history, get, "History of visited places").
+
 
 		 /*******************************
 		 *	      CREATE		*
@@ -57,6 +60,8 @@ initialise(Emacs, Buffers:dict) :->
 	send(Emacs, send_super, initialise, emacs),
 	send(Emacs, leader, frame('PceEmacs')),
 	send(Emacs, kind, service),
+	send(Emacs, slot, history,
+	     history(message(Emacs, goto_history, @arg1, tab))),
 	send(Emacs, slot, buffer_list, Buffers),
 	new(Msg, message(Emacs, check_saved_at_exit)),
 	send(@pce, exit_message, Msg),
@@ -120,10 +125,10 @@ buffers(Emacs, Buffers:chain) :<-
 	get(Emacs?buffer_list?members, map, @arg1?object, Buffers).
 
 
-open_file(_Emacs, File:file, NewWindow:new_window=[bool]) :->
+open_file(_Emacs, File:file, How:[{here,tab,window}]) :->
 	"Open a file"::
 	new(B, emacs_buffer(File)),
-	send(B, open, NewWindow).
+	send(B, open, How).
 
 
 find_file(Emacs, Dir:[directory]) :->
@@ -131,23 +136,45 @@ find_file(Emacs, Dir:[directory]) :->
 	get(@finder, file, @on, @default, Dir, FileName),
 	send(Emacs, open_file, FileName).
 
-goto_source_location(_Emacs,
+goto_source_location(Emacs,
 		     Location:source_location,
-		     NewWindow:new_window=[bool]) :->
+		     Where:where=[{here,tab,window}],
+		     Title:title=[char_array]*) :->
 	"Visit the indicated source-location"::
+	(   Title == @nil
+	->  true
+	;   send(Emacs, location_history)
+	),
 	get(Location, file_name, File),
 	new(B, emacs_buffer(File)),
-	send(B, open, NewWindow),
+	get(B, open, Where, Frame),
 	send(B, check_modified_file),
+	get(Frame, editor, Editor),
+	get(Editor, mode, Mode),
 	(   get(Location, line_no, Line),
 	    Line \== @nil
-	->  get(B?editors, find,
-		message(@arg1, instance_of, emacs_editor), E),
-	    send(E, mark_status, inactive),
-	    send(E?mode, select_line, Line)
+	->  send(Editor, mark_status, inactive),
+	    send(Mode, select_line, Line)
+	;   true
+	),
+	(   Title == @nil
+	->  true
+	;   send(Mode, location_history, title := Title)
+	).
+
+location_history(Emacs, Title:title=[char_array]) :->
+	"Save current location into history"::
+	(   get(Emacs, current_frame, Frame),
+	    get(Frame, editor, Editor),
+	    get(Editor, mode, Mode)
+	->  send(Mode, location_history, title := Title)
 	;   true
 	).
 
+goto_history(Emacs, HE:emacs_history_entry, Where:where=[{here,tab,window}]) :->
+	"Go back to an old history location"::
+	send(Emacs, goto_source_location, HE?source_location, Where, @nil),
+	send(Emacs?history, location, HE).
 
 edit(Emacs, Location:source_location) :->
 	"Equivalent to ->goto_source_location"::
@@ -212,28 +239,17 @@ check_saved_at_exit(BM) :->
 		 *******************************/
 :- pce_group(window).
 
-free_window(BM, Pool:[name], Frame:emacs_frame) :<-
-	"Return the first non-sticky window"::
+current_frame(Emacs, Frame:emacs_frame) :<-
+	"PceEmacs frame the user is working in"::
 	(   send(@event, instance_of, event),
-	    get(@event?receiver, frame, Frame),
-	    send(Frame, instance_of, emacs_frame),
-	    (Pool == @default ; get(Frame, pool, Pool)),
-	    get(Frame, sticky_window, @off)
+	    get(@event, window, Window),
+	    get(Window, frame, Frame),
+	    send(Frame, instance_of, emacs_frame)
 	->  true
-	;   get(BM, buffer_list, Dict),
-	    new(Ed, var),
-	    get(Dict, find,
-		and(assign(new(B, var), @arg1?object),
-		    assign(Ed, ?(B?editors, find,
-				 and(message(@arg1?frame, instance_of,
-					     emacs_frame),
-				     or(@arg1?frame?pool == Pool,
-					@arg1?frame?pool == @default),
-				     @arg1?frame?sticky_window == @off)),
-			   global)),
-		_),
-	    get(Ed, frame, Frame),
-	    send(Frame, expose)
+	;   get(Emacs?members, find,
+		and(message(@arg1, instance_of, emacs_frame),
+		    message(@arg1, on_current_desktop)),
+		Frame)
 	).
 
 
@@ -274,30 +290,31 @@ customise(_Emacs) :->
 
 server_start(Emacs, Force:[bool]) :->
 	"Start server-mode (xpce-client interface)"::
-	(   (	\+ get(class(socket), send_method, listen, _)
-	    ;	\+ send(class(socket), has_feature, unix_domain)
-	    ;	get(@emacs_server, status, listen)
-	    )
-	->  (   current_prolog_flag(windows, true)
-	    ->	Goal = start_emacs_dde_server(false), % fool xref
-	        catch(Goal, _, true)
-	    ;	true
-	    )
-	;   (	send(@emacs_server_address, exists, @off)
-	    ->  (   Force \== @on,
-		    pce_catch_error(socket, send(@emacs_server, connect))
-		->  free(@emacs_server),
-		    send(Emacs, report, status, 'Server on other PceEmacs'),
-		    fail
-		;   free(@emacs_server), % will recreate!
-		    ignore(send(Emacs, report, status, 'Restarted server')),
-		    send(@emacs_server_address, remove)
-		)
-	    ;	true
-	    ),
-	    ignore(send(@emacs_server, listen))
-	).
+	server_start(Emacs, Force).
 
+:- if(current_predicate(start_emacs_dde_server/1)).
+server_start(_Emacs, _Force) :-
+	(   \+ get(class(socket), send_method, listen, _)
+	;   \+ send(class(socket), has_feature, unix_domain)
+	), !,
+	start_emacs_dde_server(false).
+:- endif.
+server_start(_Emacs, _Force) :-
+	get(@emacs_server, status, listen), !.
+server_start(Emacs, Force) :-
+	(   send(@emacs_server_address, exists, @off)
+        ->  (   Force \== @on,
+	        pce_catch_error(socket, send(@emacs_server, connect))
+	    ->  free(@emacs_server),
+		send(Emacs, report, status, 'Server on other PceEmacs'),
+		fail
+	    ;   free(@emacs_server), % will recreate!
+		ignore(send(Emacs, report, status, 'Restarted server')),
+		send(@emacs_server_address, remove)
+	    )
+	;   true
+	),
+	ignore(send(@emacs_server, listen)).
 
 :- pce_group(customise).
 
