@@ -1,11 +1,12 @@
-/*  $Id: buffer.pl,v 1.25 2004/09/21 14:22:25 jan Exp $
+/*  $Id$
 
     Part of XPCE --- The SWI-Prolog GUI toolkit
 
     Author:        Jan Wielemaker and Anjo Anjewierden
-    E-mail:        jan@swi.psy.uva.nl
-    WWW:           http://www.swi.psy.uva.nl/projects/xpce/
-    Copyright (C): 1985-2002, University of Amsterdam
+    E-mail:        J.Wielemaker@cs.vu.nl
+    WWW:           http://www.swi-prolog.org/projects/xpce/
+    Copyright (C): 1985-2011, University of Amsterdam
+			      VU University Amsterdam
 
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
@@ -19,7 +20,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 
     As a special exception, if you link this library with other files,
     compiled with a Free Software compiler, to produce an executable, this
@@ -41,14 +42,17 @@
 variable(name,		  name,		get,  "Name of this buffer").
 variable(directory,	  directory,	both, "Associated CWD").
 variable(file,		  file*,	get,  "Associated file").
+variable(prompt_reload,	  bool := @on,  both, "Prompt before reloading").
 variable(mode,		  name,		get,  "Major mode of operation").
 variable(time_stamp,	  date*,	get,  "Time-stamp for file").
-variable(ensure_newline,  bool,		both, "Add newline when done").
+variable(ensure_newline,  bool := @on,	both, "Add newline when done").
+variable(ensure_no_whitespace_errors,
+			  bool,		both, "Remove trailing whitespace when done").
+variable(tab_width,	  int := 8,	both, "Width of a tab").
 variable(auto_save_mode,  bool,		both, "Auto-save?").
 variable(auto_save_count, number,	get,  "Auto-save at expiration").
 variable(saved_caret,	  int,		both, "Saved caret on last quit").
-variable(saved_fill,	  bool,		both, "Saved fill_mode on quit").
-variable(pool,		  [name],	both, "Window pool I belong too").
+variable(saved_fill,	  bool := @off,	both, "Saved fill_mode on quit").
 variable(margin_width,	  '0..' := 0,	get,  "Margin width of editors").
 variable(coloured_generation,
 	 int := -1,
@@ -59,7 +63,17 @@ variable(xref_generation,
 	 both,
 	 "Last generation we analysed").
 
-class_variable(undo_buffer_size, int, 40000).
+class_variable(undo_buffer_size,      int, 40000).
+class_variable(ensure_no_whitespace_errors, bool, @on).
+class_variable(newline_existing_file, {posix,dos,detect}, detect).
+class_variable(newline_new_file,      {posix,dos},        posix).
+:- if(current_prolog_flag(windows, true)).
+class_variable(newline_new_file,      {posix,dos},        dos).
+class_variable(unicode_encoding,      {utf8,unicode_le,unicode_be}, unicode_le).
+:- else.
+class_variable(newline_new_file,      {posix,dos},        posix).
+class_variable(unicode_encoding,      {utf8,unicode_le,unicode_be}, utf8).
+:- endif.
 
 initialise(B, File:file*, Name:[name]) :->
 	"Create from file and name"::
@@ -72,19 +86,16 @@ initialise(B, File:file*, Name:[name]) :->
 	    default(Name, '*scratch*', BufBaseName),
 	    (	BufBaseName == '*scratch*'
 	    ->	send(B, slot, mode, prolog),
-	        send(B, pool, @default),
 		scratch_text(Text),
 		send(B, insert, 0, Text),
 		send(B, saved_caret, B?size)
-	    ;	send(B, slot, mode, fundamental),
-		send(B, pool, other)
+	    ;	send(B, slot, mode, fundamental)
 	    ),
 	    send(B, directory, directory('.'))
 	;   send(File, absolute_path),
 	    get(File, base_name, FileBaseName),
 	    default(Name, FileBaseName, BufBaseName),
 	    send(B, file, File),
-	    send(B, pool, file),
 	    send(B, auto_save_mode, @on),
 	    send(@emacs_base_names, append, FileBaseName, B),
 	    send(B, determine_initial_mode),
@@ -94,9 +105,8 @@ initialise(B, File:file*, Name:[name]) :->
 	    )
 	),
 
+	send(B, init_mode_defaults),
 	send(B, slot, auto_save_count, number(300)),
-	send(B, saved_fill, @off),
-	send(B, ensure_newline, @on),
 	send(B, name, BufBaseName).
 
 
@@ -109,20 +119,44 @@ unlink(B) :->
 	),
 	send(B, send_super, unlink).
 
-scratch_text('% This buffer is for notes you don\'t want to save.\n\
-	     % If you want to create a file, visit that file with C-x C-f,\n\
-	     % then enter the text in that file\'s own buffer.\n\n').
+report(B,
+       Kind:kind={status,inform,progress,done,warning,error,fatal},
+       Format:format=[char_array],
+       Argv:any ...) :->
+	"Report to associated editors"::
+	get(B, editors, Editors),
+	Message =.. [message, @arg1, report, Kind, Format | Argv],
+	send(Editors, for_all, Message).
+
+
+scratch_text('% This buffer is for notes you don\'t want to save.\n\c
+	      % If you want to create a file, visit that file with C-x C-f,\n\c
+	      % then enter the text in that file\'s own buffer.\n\n').
 
 :- pce_global(@emacs_interpreter_regex,
-	      new(regex('#!\\(\\S +\\)\\s '))).
+	      new(regex('#!(\\S+)\\s'))).
 :- pce_global(@emacs_mode_regex,	% -*- Mode -*-
 					% -*- mode: Mode; ... -*-
-	      new(regex('.*-\\*-\\s *\\([Mm]ode:\\s *\\(\\w+\\);.*-\\*-\\|\\(\\w+\\)\\s *-\\*-\\)'))).
+	      new(regex('.*-\\*-\\s*([Mm]ode:\\s*(\\w+);.*-\\*-|(\\w+)\\s*-\\*-)'))).
 
 
+% ->determine_initial_mode uses the following steps:
+%
+%   1. If the file is a loaded file, it is a Prolog file
+%   2. The Emacs magic sequences -*- Mode -*- or -*- mode: Mode; ... -*-
+%   3. Try @emacs_content_mode_list
+%   4. Try #! interpreter (PrologScript)
+%   5. Try the file-name
+
 determine_initial_mode(B) :->
 	"Determine initial mode"::
-	(   send(@emacs_mode_regex, match, B),
+	(   get(B, file, File), File \== @nil,
+	    get(File, name, FileName),
+	    absolute_file_name(FileName, FilePath),
+	    source_file(FilePath),
+	    \+ source_file_property(FilePath, derived_from(_,_))
+	->  send(B, slot, mode, prolog)
+	;   send(@emacs_mode_regex, match, B),
 	    member(Reg, [2,3]),
 	    get(@emacs_mode_regex, register_value, B, Reg, Mode0),
 	    get(Mode0?downcase, value, Mode),
@@ -141,11 +175,20 @@ determine_initial_mode(B) :->
 		message(@arg1?name, match, Match), Att)
 	->  send(B, slot, mode, Att?value)
 	;   send(B, slot, mode, @emacs_default_mode)
+	),
+	send(B, set_temp_file).
+
+set_temp_file(B) :->
+	"Clear ->prompt_reload if this is a temp file"::
+	get(B, file, File),
+	(   no_backup(File)
+	->  send(B, prompt_reload, @off)
+	;   true
 	).
 
 
-%	content_from_mode(+Buffer, -Mode)
-%	
+%%	content_from_mode(+Buffer, -Mode) is semidet.
+%
 %	Search Buffer with the patterns from @emacs_content_mode_list
 
 content_from_mode(B, Mode) :-
@@ -161,15 +204,18 @@ attach(B, E:editor) :->
 	get(B, editors, Editors),
 	(   send(Editors, empty)
 	->  get(B, saved_caret, Caret),
-	    get(B, saved_fill, Fill)
+	    get(B, saved_fill, Fill),
+	    get(B, tab_width, TabWidth)
 	;   get(Editors?head, caret, Caret),
-	    get(Editors?head, fill_mode, Fill)
+	    get(Editors?head, fill_mode, Fill),
+	    get(Editors?head, tab_distance, TabWidth)
 	),
 	get(B, margin_width, MW),
 	send(B, send_super, attach, E),
 	send(E, caret, Caret),
 	send(E, fill_mode, Fill),
-	send(E, margin_width, MW).
+	send(E, margin_width, MW),
+	send(E, tab_distance, TabWidth).
 
 
 detach(B, E:editor) :->
@@ -193,7 +239,7 @@ name(B, Name:name) :->
 		get(Name, append, string('<%d>', N), BufName),
 	        \+ get(@emacs_buffers, member, BufName, _),
 		!
- 	    ;   BufName = Name
+	    ;   BufName = Name
 	    ),
 	    send(B, slot, name, BufName),
 	    (	OldName \== @nil,
@@ -242,12 +288,16 @@ file(B, File:file) :->
 	(   send(directory(File?name), exists)
 	->  send(File, error, open_file, read, 'is a directory')
 	;   send(File, exists)
-	->  ignore(send(B, insert_file, 0, File)),
+	->  get(B, newline_existing_file, OpenMode),
+	    send(File, newline_mode, OpenMode),
+	    ignore(send(B, insert_file, 0, File)),
 	    send(B, reset_undo),
 	    send(B, modified, @off),
 	    send(B, slot, time_stamp, File?time)
 	;   send(B, reset_undo),
-	    send(B, modified, @off)
+	    send(B, modified, @off),
+	    get(B, newline_new_file, Mode),
+	    send(File, newline_mode, Mode)
 	),
 	send(B, slot, file, File),
 	new(F2, file(File?absolute_path)),
@@ -276,17 +326,14 @@ save(B, File:[file]) :->
 	    send(@emacs_base_names, append, File?base_name, B)
 	),
 	(   get(B, ensure_newline, @on)
-	->  get(B, size, Size),
-	    (	(   Size == 0
-		;   get(B, character, Size-1, 10)
-		)
-	    ->	true
-	    ;	send(B, append, string('\n'))
-	    )
+	->  send(B, complete_last_line)
 	;   true
 	),
-	(   get(@emacs_no_backup_list, find,
-		message(@arg1, match, SaveFile?name), _)
+	(   get(B, ensure_no_whitespace_errors, @on)
+	->  send(B, fix_whitespace_errors)
+	;   true
+	),
+	(   no_backup(SaveFile)
 	->  true
 	;   ignore(send(SaveFile, backup))
 	),
@@ -297,15 +344,99 @@ save(B, File:[file]) :->
 	;   true
 	).
 
+no_backup(File) :-
+	get(@emacs_no_backup_list, find,
+	    message(@arg1, match, File?name), _).
 
-do_save(B, SaveFile:file) :->
+
+complete_last_line(B) :->
+	"Add \\n if needed"::
+	get(B, size, Size),
+	(   (   Size == 0
+	    ;   get(B, character, Size-1, 10)
+	    )
+	->  true
+	;   send(B, append, string('\n'))
+	).
+
+fix_whitespace_errors(B) :->
+	"Remove trailing spaces and tabs from lines"::
+	new(Count, number(0)),
+	send(B, fix_trailing_space_errors, Count),
+	send(B, fix_space_tab_errors, Count),
+	(   get(Count, value, 0)
+	->  true
+	;   send(B, report, status,
+		 'Fixed %d whitespace errors', Count)
+	).
+
+fix_trailing_space_errors(B, Count:number) :->
+	"Remove trailing spaces and tabs from lines"::
+	new(Re, regex('[ \t]+\n')),
+	send(Re, for_all, B,
+	     and(message(@arg1, replace, @arg2, '\n'),
+		 message(Count, plus, 1))).
+
+fix_space_tab_errors(B, Count:number) :->
+	"Replace space+tab sequences"::
+	new(Re, regex(' +\t')),
+	send(Re, for_all, B,
+	     and(message(B, fix_space_tab, Re),
+		 message(Count, plus, 1))).
+
+fix_space_tab(B, Re:regex) :->
+	"Fix matched spaces followed by tab"::
+	(   get(B?editors, head, E)
+	->  get(Re, register_start, 0, Start),
+	    get(Re, register_end, 0, End),
+	    get(E, column, Start, StartCol),
+	    get(E, column, End, EndCol),
+	    get(E, tab_distance, TD),
+	    tabs(StartCol, EndCol, TD, Tabs),
+	    tab_atom(Tabs, Atom),
+	    send(Re, replace, B, Atom)
+	;   true
+	).
+
+tabs(SC, EC, _, 0) :-
+	SC >= EC, !.
+tabs(SC, EC, TD, N) :-
+	SC2 is ((SC+TD)//TD)*TD,
+	tabs(SC2, EC, TD, N0),
+	N is N0+1.
+
+tab_atom(N, Atom) :-
+	length(List, N),
+	maplist(=(0'\t), List),
+	atom_codes(Atom, List).
+
+
+do_save(B, SaveFile:file, Start:[int], Length:[int]) :->
 	"Do the actual saving"::
-	send(B, send_super, save, SaveFile).
+	get(B, unicode_encoding, FallBackEncoding),
+	(   pce_catch_error(io_error,
+			    send_super(B, save, SaveFile, Start, Length))
+	->  true
+	;   get(SaveFile, encoding, Encoding),
+	    Encoding \== FallBackEncoding
+	->  send(SaveFile, encoding, FallBackEncoding),
+	    send(SaveFile, bom, @on),
+	    send_super(B, save, SaveFile, Start, Length),
+	    once(user_encoding(FallBackEncoding, UserEnc)),
+	    send(B, report, warning,
+		 'Could not save using default locale; saved using %s', UserEnc)
+	;   send_super(B, save, SaveFile, Start, Length)
+	).
+
+user_encoding(utf8, 'UTF-8').
+user_encoding(unicode_le, 'UTF-16 (little endian)').
+user_encoding(unicode_be, 'UTF-16 (big endian)').
+user_encoding(Enc, Enc).
 
 
 write_region(B, File:file, Start:int, Length:int) :->
 	"Wrote region to file (start, length)"::
-	send(B, send_super, save, File, Start, Length).
+	send(B, do_save, File, Start, Length).
 
 
 save_if_modified(B, Confirm:[bool]) :->
@@ -337,7 +468,7 @@ check_auto_save(B) :->
 	->  send(B, auto_save)
 	;   true
 	).
-	
+
 
 auto_save_file(B, F:file) :<-
 	get(B, file, File), File \== @nil,
@@ -435,7 +566,7 @@ update_label(B) :->
 		EditorLabel = Name
 	    ),
 	    send(B?editors, for_all,
-		 message(@arg1?frame, label, EditorLabel))
+		 message(@arg1, label, EditorLabel))
 	;   true
 	).
 
@@ -457,7 +588,7 @@ margin_width(B, W:'0..') :->
 
 modified(B, Val:bool) :->
 	"Check the file; mark buffer-menu"::
-	send(B, send_super, modified, Val),
+	send_super(B, modified, Val),
 	(   Val == @on
 	->  send(B, check_modified_file)
 	;   send(B, delete_auto_save_file)
@@ -465,7 +596,7 @@ modified(B, Val:bool) :->
 	send(B, update_label).
 
 
-check_modified_file(B, Confirm:[bool]) :->
+check_modified_file(B, Frame:frame=[frame], Confirm:confirm=[bool]) :->
 	"Check if file has been modified after buffer"::
 	(   get(B, file, File),
 	    File \== @nil,
@@ -474,15 +605,19 @@ check_modified_file(B, Confirm:[bool]) :->
 	    get(File, time, FileStamp),
 	    \+ send(Stamp, equal, FileStamp),
 	    \+ object(@emacs_reverting)
-	->  (   confirm_reload(Confirm, File)
+	->  (   confirm_reload(B, Frame, Confirm, File)
 	    ->	send(B, revert)
 	    ;	true
 	    )
 	;   true
 	).
 
-confirm_reload(@off, _) :- !.
-confirm_reload(_, File) :-
+confirm_reload(_, _, @off, _) :- !.
+confirm_reload(B, _, @default, _) :-
+	get(B, prompt_reload, @off), !,
+	send(B, saved_caret, 0),
+	send(B?editors, for_all, message(@arg1, caret, 0)).
+confirm_reload(_, Frame, _, File) :-
 	new(D, dialog('Modified file')),
 	send(D, append,
 	     label(title,  string('File %N was modified', File))),
@@ -490,7 +625,18 @@ confirm_reload(_, File) :-
 	     button(reload_file, message(D, return, reload_file))),
 	send(D, append,
 	     button(edit_buffer, message(D, return, edit_buffer))),
-	get(D, confirm_centered, RVal),
+	(   Frame \== @default
+	->  EmacsFrame = Frame
+	;   get(@emacs, current_frame, EmacsFrame)
+	->  true
+	;   EmacsFrame = @nil
+	),
+	(   EmacsFrame \== @default
+	->  get(EmacsFrame?area, center, Position),
+	    send(D, transient_for, EmacsFrame)
+	;   Position = @default
+	),
+	get(D, confirm_centered, Position, RVal),
 	send(D, destroy),
 	RVal == reload_file.
 
@@ -499,26 +645,24 @@ confirm_reload(_, File) :-
 		 *          OPEN WINDOW		*
 		 *******************************/
 
-open(B, New:[bool], Window:emacs_frame) :<-
+open(B, How:[{here,tab,window}], Frame:emacs_frame) :<-
 	"Create window for buffer"::
-	(   New == @on
-	->  send(new(Window, emacs_frame(B)), open)
-	;   (   get(B?editors, find,
-		    and(message(@arg1, instance_of, emacs_editor),
-			@arg1?window?reuse == @on),
-		    Editor)
-	    ->  get(Editor, frame, Window),
-		send(Window, expose)
-	    ;	get(@emacs, free_window, B?pool, Window)
-	    ->	send(Window, buffer, B)
-	    ;	send(new(Window, emacs_frame(B)), open)
-	    )
+	(   How == window
+	->  send(new(Frame, emacs_frame(B)), open)
+	;   How == tab,
+	    get(@emacs, current_frame, Frame)
+	->  send(Frame, tab, B, @on),
+	    send(Frame, expose)
+	;   get(@emacs, current_frame, Frame)
+	->  send(Frame, buffer, B),
+	    send(Frame, expose)
+	;   send(new(Frame, emacs_frame(B)), open)
 	),
-	send(B, check_modified_file).
+	send(B, check_modified_file, Frame).
 
-open(B, New:[bool]) :->
+open(B, How:[{here,tab,window}]) :->
 	"Create window for buffer"::
-	get(B, open, New, _).
+	get(B, open, How, _).
 
 
 properties(Buffer) :->
@@ -532,13 +676,11 @@ properties(Buffer, V:view) :<-
 	get(Buffer, size, Size),
 	get(Buffer, line_number, Lines),
 	get(Buffer, mode, Mode),
-	get(Buffer, pool, Pool),
 	new(V, view(string('Buffer %s', Name), size(60, 8))),
 	send(V, confirm_done, @off),
 	send(V, tab_stops, vector(200)),
 	send(V, appendf, 'Buffer Name:\t%s\n', Name),
 	send(V, appendf, 'Mode:\t%s\n', Mode),
-	send(V, appendf, 'Window Pool:\t%s\n', Pool),
 	send(V, appendf, 'Modified:\t%s\n', Modified?name),
 	send(V, appendf, 'Size:\t%d characters; %d lines\n', Size, Lines-1),
 	get(Buffer, file, File),
@@ -550,7 +692,9 @@ properties(Buffer, V:view) :<-
 	),
 	(   File \== @nil
 	->  get(File, absolute_path, Path),
-	    send(V, appendf, 'Path:\t%s\n', Path)
+	    send(V, appendf, 'Path:\t%s\n', Path),
+	    send(V, appendf, 'Encoding:\t%s (BOM=%s, NL=%s)\n',
+		 File?encoding, File?bom, File?newline_mode)
 	;   send(V, appendf, 'Path:\t<No file>\n')
 	),
 	send(V, caret, 0),
@@ -562,23 +706,40 @@ properties(Buffer, V:view) :<-
 		 /*******************************
 		 *	      MODE		*
 		 *******************************/
-	
+
 mode(B, Mode:name) :->
 	"Switch to named mode"::
 	(   get(B, mode, Mode)
 	->  true
 	;   send(B, slot, mode, Mode),
+	    send(B, init_mode_defaults),
 	    send(B?editors, for_some, message(@arg1, mode, Mode))
 	).
 
+init_mode_defaults(B) :->
+	"Initialise defaults from the current mode"::
+	get(B, mode, ModeName),
+	atomic_list_concat([emacs_, ModeName, '_mode'], ClassName),
+	get(@pce, convert, ClassName, class, ModeClass),
+	(   copy_class_var(Name),
+	    get(ModeClass, class_variable, Name, CV),
+	    get(CV, value, Value),
+	    send(B, Name, Value),
+	    fail
+	;   true
+	).
+
+copy_class_var(indent_tabs).
+copy_class_var(tab_width).
+
 
 		 /*******************************
 		 *       LANGUAGE SUPPORT	*
 		 *******************************/
 
-%	emacs_buffer<-name_and_arity returns the name and arity if the 
+%	emacs_buffer<-name_and_arity returns the name and arity if the
 %	caret is in the functor of the term.  If the arity cannot be
-%	determines, arity is returned as @default.
+%	determined, arity is returned as @default.
 
 name_and_arity(TB, Pos:int, Tuple:tuple) :<-
 	"Find name and arity of term at position"::
@@ -619,5 +780,5 @@ count_args(TB, Here, NAT, A0, A) :-
 	).
 
 :- pce_end_class.
-	
-	
+
+
