@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 
     As a special exception, if you link this library with other files,
     compiled with a Free Software compiler, to produce an executable, this
@@ -78,11 +78,11 @@ current(TW, Window:window) :->
 	"Make the given window the current one"::
 	send_super(TW, current, Window),
 	get(Window, text_buffer, TB),
-	send(TB, check_modified_file, @on),
 	(   get(TW, frame, Frame),
+	    send(TB, check_modified_file, Frame),
 	    send(Frame, has_send_method, setup_mode)
 	->  send(TW?frame, setup_mode, Window)
-	;   true
+	;   send(TB, check_modified_file)
 	).
 
 frame_window(_TW, Window:window, _Name:name, _Rank:'1..', Frame:frame) :<-
@@ -172,7 +172,7 @@ tab(F, B:buffer=emacs_buffer, Expose:expose=[bool]) :->
 	    ->	send(TW, on_top, Window)
 	    ;	true
 	    )
-	;   send(TW, append, new(V, emacs_view(B)), @default, Expose),
+	;   send(TW, append, new(V, emacs_view(B)), B?name, Expose),
 	    send(B, update_label),
 	    send(F, setup_mode, V)
 	).
@@ -321,7 +321,15 @@ initialise(D) :->
 	send_super(D, initialise),
 	send(D, gap, size(0,0)),
 	send(D, pen, 0),
-	send(D, append, new(emacs_menu_bar)).
+	send(D, append, new(emacs_menu_bar)),
+	send(D, append, new(TB, tool_bar), right),
+	send(TB, reference, point(0,15)),
+	send(TB, alignment, right),
+	get(@emacs, history, History),
+	get(History, button, forward, Forward),
+	get(History, button, backward, Backward),
+	send_list(TB, append, [Backward,Forward]),
+	send_list([Backward,Forward], activate).
 
 resize(D) :->
 	send(D, layout, D?area?size).
@@ -429,7 +437,7 @@ human_accelerator(Key, Text) :-
 
 :- pce_begin_class(emacs_mini_window, dialog, "Prompt and feedback window").
 
-variable(prompter, 	 dialog_item*,  get,	"Current prompter").
+variable(prompter,	 dialog_item*,  get,	"Current prompter").
 variable(report_count,	 number,	get,	"Count to erase report").
 variable(report_type,	 name*,		both,	"Last type of report").
 
@@ -438,7 +446,8 @@ initialise(D) :->
 	send(D, slot, report_count, number(0)),
 	send(D, gap, size(10, 2)),
 	send(D, pen, 0),
-	send(D, display, label(reporter), point(0, 2)),
+	send(D, display, new(R, label(reporter)), point(0, 2)),
+	send(R, wrap, clip),
 	send(D, display, new(T, text('', right, normal)), point(100, 2)),
 	send(T, name, line),
 	get(text_item(''), height, MH),
@@ -694,7 +703,7 @@ prompt(V, Label:char_array, Default:[any], Type:[type], History:[chain],
 	fix_rval(Type, RawRval, Rval),
 	(   object(Rval),
 	    get(Rval, lock_object, @off)
-	->  send(Rval, lock_object, @on), 	% protect during deletion!
+	->  send(Rval, lock_object, @on),	% protect during deletion!
 	    free(Item),
 	    get(Rval, unlock, Rval)
 	;   free(Item)
@@ -803,7 +812,7 @@ make_idle_timer(T) :-
 	new(T, timer(2)),
 	send(T, message,
 	     new(Msg, message(T, send_hyper, editor, editor_idle_event))),
-	send(Msg, debug_class, service). 	% non-traceable
+	send(Msg, debug_class, service).	% non-traceable
 
 
 editor_idle_event(E) :->
@@ -829,7 +838,11 @@ typed(E, Id:'event|event_id') :->
 	"Handle typing via mode"::
 	send(E, start_idle_timer),
 	get(E, mode, Mode),
-	send(Mode, typed, Id, E).
+	(   send(Mode, typed, Id, E),
+	    object(Mode)                % may disappear
+	->  send(Mode, highlight_matching_bracket)
+	;   true
+	).
 
 
 caret(E, Caret:[int]) :->
@@ -989,6 +1002,15 @@ looking_at(E, Re:regex, Where:[int], End:[int]) :->
 	get(E, text_buffer, TB),
 	send(Re, match, TB, C, End).
 
+looking_at(E, Re:regex, Where:[int], End:[int], Len:int) :<-
+	"Test if regex macthes from the caret"::
+	(   Where == @default
+	->  get(E, caret, C)
+	;   C = Where
+	),
+	get(E, text_buffer, TB),
+	get(Re, match, TB, C, End, Len).
+
 :- pce_end_class.
 
 
@@ -1329,11 +1351,11 @@ typed(M, Id:'event|event_id', Editor:editor) :->
 
 					% send to mode rather than editor
 	(   get(M, focus_function, F), F \== @nil
- 	->  (   send(M, F, Id)
+	->  (   send(M, F, Id)
 	    ->  true
 	    ;   send(M, focus_function, @nil),
 		send(M, typed, Id, Editor)	  % failed: unfocus and resent
- 	    )
+	    )
 	;   get(M, bindings, Binding),
 	    send(Binding, typed, Id, M)
 	).
@@ -1460,6 +1482,46 @@ m_x_next(M, Value:any) :<-
 	send(M, m_x_index, Nidx).
 
 
+		 /*******************************
+		 *	 LOCATION HISTORY	*
+		 *******************************/
+
+location_history(M, Start:start=[int], Len:length=[int],
+		 Title:title=[char_array]) :->
+	"Add location to the editor history"::
+	(   Start == @default
+	->  get(M, caret, Caret),
+	    get(M, scan, Caret, line, 0, start, SOF)
+	;   SOF = Start
+	),
+	(   send(M, history_not_interesting, SOF)
+	->  true
+	;   get(M, text_buffer, TB),
+	    new(_, emacs_history_fragment(TB, SOF, Len, Title))
+	).
+
+history_not_interesting(M, Start:int) :->
+	"True if Start is close to the recent history mark"::
+	(   Start == 0
+	;   send(M, history_close_to_last, Start)
+	), !.
+
+history_close_to_last(M, Start:int, MaxDist:[int]) :->
+	"True if Start is within MaxDist lines from last"::
+	default(MaxDist, 10, MD),
+	get(M, text_buffer, TB),
+	get(@emacs?history, current, Current),
+	get(Current, get_hyper, fragment, text_buffer, TB),
+	get(Current, get_hyper, fragment, start, StartOfCurrent),
+	(   Start < StartOfCurrent
+	->  get(TB, count_lines, Start, StartOfCurrent, Lines)
+	;   Start > StartOfCurrent
+	->  get(TB, count_lines, StartOfCurrent, Start, Lines)
+	;   Lines = 0
+	),
+	Lines < MD.
+
+
 		 /*******************************
 		 *	      REPORT		*
 		 *******************************/
