@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 
     As a special exception, if you link this library with other files,
     compiled with a Free Software compiler, to produce an executable, this
@@ -73,10 +73,11 @@ reexports the content of these files.
 
 	    pce_catch_error/2,		% pce_error
 	    pce_open/3,
-	    pce_dispatch/1,		% +Options
-	    pce_end_dispatch/0,		% break dispatch loop
-	    pce_call/1,			% :Goal
 	    in_pce_thread/1,		% :Goal
+	    in_pce_thread_sync/1,	% :Goal
+	    set_pce_thread/0,
+	    pce_thread/1,		% -Thread
+	    pce_dispatch/0,
 
 	    pce_welcome/0,
 
@@ -85,7 +86,10 @@ reexports the content of these files.
 	    op(990, xfx, :=)
 	  ]).
 
-:- system_module.
+:- set_prolog_flag(generate_debug_info, false).
+
+:- meta_predicate
+	in_pce_thread_sync(0).
 
 		/********************************
 		*      LOAD COMMON PLATFORM	*
@@ -109,13 +113,49 @@ reexports the content of these files.
 		pce_boot(pce_keybinding),
 		pce_boot(pce_portray)
 	      ],
-	      [ qcompile(true),
+	      [ qcompile(part),		% compile boot files as pat of pce.qlf
 		silent(true)
 	      ]).
 
+%%	pce_thread(-Thread) is det.
+%
+%	True if Thread is the Prolog thread that runs the graphics
+%	message loop.
+%
+%	@see pce_dispatch/1.
+
+:- dynamic
+	pce_thread/1.
+
+start_dispatch :-
+	(   current_prolog_flag(xpce_threaded, true)
+	->  pce_dispatch([])
+	;   true
+	).
+
+%%	in_pce_thread_sync(:Goal) is semidet.
+%
+%	Same as in_pce_thread/1, but wait  for   Goal  to  be completed.
+%	Success depends on the success of executing Goal. If Goal throws
+%	an exception, this exception is re-thrown by in_pce_thread/1.
+%
+%	Possible bindings of Goal are returned,   but  be aware that the
+%	term has been _copied_. If in_pce_thread_sync/1 is called in the
+%	thread running pce, it behaves as once/1.
+
+in_pce_thread_sync(Goal) :-
+	thread_self(Me),
+	pce_thread(Me), !,
+	Goal, !.
+in_pce_thread_sync(Goal) :-
+	term_variables(Goal, Vars),
+	pce_principal:in_pce_thread_sync2(Goal-Vars, Vars).
+
+:- initialization
+	start_dispatch.
+
 set_version :-
 	current_prolog_flag(version_data, swi(Major, Minor, Patch, _)),
-	send(@pce, catch_error_signals, @on),
 	format(string(PlId),
 	       'SWI-Prolog version ~w.~w.~w', [Major, Minor, Patch]),
 	send(@prolog, system, PlId).
@@ -126,7 +166,7 @@ get_pce_version :-
 	(   current_prolog_flag(xpce_version, _)
 	->  true
 	;   get(@pce, version, name, Version),
-	    set_prolog_flag(xpce_version, Version)
+	    create_prolog_flag(xpce_version, Version, [])
 	).
 
 :- initialization get_pce_version.
@@ -174,7 +214,7 @@ prolog_edit:load :-
 		 *	    LIST HOOKS		*
 		 *******************************/
 
-%	prolog:locate_clauses(Term, Refs)
+%%	prolog:locate_clauses(Term, Refs)
 %
 %	Locate a list of clause-references from a method-specification
 %	like Class->Method.
