@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 
     As a special exception, if you link this library with other files,
     compiled with a Free Software compiler, to produce an executable, this
@@ -51,6 +51,7 @@
 :- use_module(library(debug)).
 :- use_module(library(autowin)).
 :- use_module(library(broadcast)).
+:- use_module(library(prolog_source)).
 
 version('0.1.1').
 
@@ -66,7 +67,8 @@ setting(warn_not_called,    true).
 setting(hide_system_files,  true).
 setting(hide_profile_files, true).
 
-/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+/** <module> Cross-referencer front-end
+
 XPCE based font-end of the Prolog cross-referencer.  Tasks:
 
 	* Cross-reference currently loaded program		OK
@@ -87,14 +89,20 @@ XPCE based font-end of the Prolog cross-referencer.  Tasks:
 		- Using use_module/2				OK
 		- Export header for non-module files		OK
 
-----------------
-NOTE: This is work in progress.  
+@bug	Tool produces an error if a file that has been xref'ed is
+	deleted.  Paulo Moura.
+@see	library(prolog_xref) holds the actual data-collection.
+*/
 
-Its in CVS as this makes it easier to maintain and as-is, the
-tools is now in a useable state.
-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+%%	gxref
+%
+%	Start graphical cross-referencer on loaded program.  The GUI
+%	is started in the XPCE thread.
 
 gxref :-
+	in_pce_thread(xref_gui).
+
+xref_gui :-
 	send(new(XREF, xref_frame), open),
 	send(XREF, wait),
 	send(XREF, update).
@@ -124,7 +132,7 @@ initialise(F) :->
 		  [ append(new(xref_depgraph), dependencies)
 		  ]),
 	send(F, fill_toolbar, TD).
-	
+
 tab_popup(_F, P:popup) :<-
 	"Popup for tab labels"::
 	new(P, popup),
@@ -152,7 +160,7 @@ fill_toolbar(F, TD:tool_dialog) :->
 	send(Settings, show_current, @on),
 	send(Settings, multiple_selection, @on),
 	send(F, update_setting_menu).
-	     
+
 about(_F) :->
 	version(Version),
 	send(@display, inform,
@@ -179,7 +187,7 @@ workspace(F, Which:name, Create:[bool], Expose:bool, WS:window) :<-
 
 workspace_term(file_info, prolog_file_info).
 workspace_term(header,    xref_view).
-	    
+
 browser(F, Which:name, Browser:browser) :<-
 	"Find named browser"::
 	get(F, member, browsers, Tabs),
@@ -224,7 +232,7 @@ file_header(F, File:name) :->
 	"Create import/export header"::
 	get(F, workspace, header, @on, @on, View),
 	send(View, file_header, File).
-		     
+
 :- pce_group(settings).
 
 update_setting_menu(F) :->
@@ -307,7 +315,7 @@ sources(_, Sources:prolog) :<-
 	findall(S, dep_source(S), Sources).
 
 %%	dep_source(?Src)
-%	
+%
 %	Generate all sources for the dependecy graph one-by-one.
 
 dep_source(Src) :-
@@ -342,7 +350,7 @@ node(G, File:name, Create:[bool|{always}], Pos:[point],
 	    send(G, display, new(Gr, xref_file_graph_node(File)), At),
 	    send(G, remove_intro_text)
 	).
-	
+
 update_links(G) :->
 	"Add all export links"::
 	send(G?graphicals, for_all,
@@ -482,7 +490,7 @@ popup(N, Popup:popup) :<-
 show_import_exports(N, Which:{import,export}) :->
 	"Show who I'm exporting to"::
 	get(N, device, G),
-	layout_new(G, 
+	layout_new(G,
 		   (   (   Which == export
 		       ->  send(N, create_export_links, @on)
 		       ;   send(N, create_import_links, @on)
@@ -556,7 +564,7 @@ edit_callable(C, Callable:prolog) :->
 	;   get(C?to, path, ImportFile),
 	    edit_callable(Callable, ImportFile)
 	).
-	
+
 :- pce_end_class(xref_export_connection).
 
 
@@ -586,7 +594,7 @@ event(Tag, Ev:event) :->
 
 %%	export_link(+ExportingFile, -ImportingFile, -Callables) is det.
 %%	export_link(-ExportingFile, +ImportingFile, -Callables) is det.
-%	
+%
 %	Callables are exported from ExportingFile to ImportingFile.
 
 export_link(ExportFile, ImportingFile, Callables) :-
@@ -758,7 +766,7 @@ clear(Tree) :->
 append(Tree, File:name) :->
 	"Add Prolog source file"::
 	send(Tree, append_node, new(prolog_file_node(File))).
-	
+
 append_node(Tree, Node:toc_node) :->
 	"Append a given node to the tree"::
 	get(Node, parent_id, ParentId),
@@ -768,7 +776,7 @@ append_node(Tree, Node:toc_node) :->
 		 new(Parent, prolog_directory_node(ParentId)))
 	),
 	send(Parent, son, Node).
-	
+
 sort(Tree) :->
 	forall(top_node(Name, _),
 	       (   get(Tree, node, Name, Node),
@@ -825,7 +833,7 @@ filter_files(Tree, Filter:code*) :->
 	    send(Tree, report, status, 'Filter on file name: %d hits', Count)
 	),
 	send(Tree, scroll_to, point(0,0)).
-	
+
 show_node_path(Tree, Node:node) :->
 	"Select Node and make sure all parents are expanded"::
 	send(Node, selected, @on),
@@ -857,7 +865,7 @@ initialise(DN, Dir:name, Label:[name]) :->
 	"Create a directory node"::
 	(   Label \== @default
 	->  Name = Label
-	;   alias_path(Name, Dir)
+	;   file_alias_path(Name, Dir)
 	->  true
 	;   file_base_name(Dir, Name)
 	),
@@ -866,7 +874,7 @@ initialise(DN, Dir:name, Label:[name]) :->
 parent_id(FN, ParentId:name) :<-
 	"Get id for the parent"::
 	get(FN, identifier, Path),
-	(   alias_path(_, Path)
+	(   file_alias_path(_, Path)
 	->  ParentId = alias
 	;   file_directory_name(Path, ParentId)
 	).
@@ -924,7 +932,7 @@ parent_id(FN, ParentId:name) :<-
 	"Get id for the parent"::
 	get(FN, identifier, Path),
 	file_directory_name(Path, Dir),
-	(   alias_path('.', Dir)
+	(   file_alias_path('.', Dir)
 	->  ParentId = '.'
 	;   ParentId = Dir
 	).
@@ -1139,7 +1147,7 @@ show_module_export(W, File:name, Module:name, Callable:prolog) :->
 		   send(XL, append, xref_imported_by(F, Callable)))
 	),
 	send(T, next_row).
-	
+
 show_file_export(W, Callable:prolog, ImportFiles:prolog) :->
 	get(W, prolog_file, File),
 	get(W, tabular, T),
@@ -1159,10 +1167,10 @@ qualify_from_file(Callable, _, Callable).
 
 
 %%	exported_to(+ExportFile, +Callable, -ImportFile)
-%	
+%
 %	ImportFile imports Callable from ExportFile.  The second clause
 %	deals with auto-import.
-%	
+%
 %	TBD: Make sure the autoload library is loaded before we begin.
 
 exported_to(ExportFile, Callable, ImportFile) :-
@@ -1236,7 +1244,7 @@ show_undef(W, Callable:prolog) :->
 	forall(member(P, Sorted),
 	       send(L, append, xref_predicate_text(Module:P, called_by, File))),
 	send(T, next_row).
-	
+
 
 show_not_called(W) :->
 	"Show predicates that are not called"::
@@ -1288,9 +1296,9 @@ initialise(T, Callable0:prolog,
 	;   true
 	),
 	send(T, classification, Class).
-       
+
 %%	single_qualify(+Term, -Qualified)
-%	
+%
 %	Strip redundant M: from the term, leaving at most one qualifier.
 
 single_qualify(_:Q0, Q) :-
@@ -1326,7 +1334,7 @@ classification(T, Class:[name]) :->
 
 :- pce_global(@xref_predicate_text_recogniser,
 	      new(handler_group(@arm_recogniser,
-				click_gesture(left, '', single, 
+				click_gesture(left, '', single,
 					      message(@receiver, edit))))).
 
 event(T, Ev:event) :->
@@ -1367,7 +1375,7 @@ variable(default_action, name := edit, both, "Default on click").
 
 initialise(TF, File:name) :->
 	absolute_file_name(File, Path),
-	short_file_name(Path, ShortId),
+	file_name_on_path(Path, ShortId),
 	short_file_name_to_atom(ShortId, Label),
 	send_super(TF, initialise, Label),
 	send(TF, name, Path),
@@ -1377,7 +1385,7 @@ initialise(TF, File:name) :->
 	      make_xref_file_text_recogniser).
 
 make_xref_file_text_recogniser(G) :-
-	new(C, click_gesture(left, '', single, 
+	new(C, click_gesture(left, '', single,
 			     message(@receiver, run_default_action))),
 	new(P, popup_gesture(@arg1?popup)),
 	new(D, drag_and_drop_gesture(left)),
@@ -1504,7 +1512,7 @@ initialise(IT, File:name, Imported:prolog) :->
 	send(F, name, file_text),
 	send(IT, slot, callable, Imported),
 	send(IT, show_called_by).
-	
+
 path(IT, Path:name) :<-
 	"Represented file"::
 	get(IT, member, file_text, Text),
@@ -1530,7 +1538,7 @@ called_by(IT, ByList:prolog) :<-
 	findall(By, used_in(Source, Callable, By), ByList).
 
 %%	used_in(+Source, +QCallable, -CalledBy)
-%	
+%
 %	Determine which the callers for   QCallable in Source. QCallable
 %	is qualified with the module of the exporting file (if any).
 
@@ -1717,27 +1725,11 @@ print_decls(Term, Out) :-
 
 
 		 /*******************************
-		 *	  LOGIC (MAY MOVE)	*
+		 *	  FILE-NAME LOGIC	*
 		 *******************************/
 
-%%	short_file_name(+File, -ShortId)
-%	
-%	Create a short name for a file
-
-short_file_name(Path, ShortId) :-
-	(   alias_path(Alias, Dir),
-	    atom_concat(Dir, Local, Path)
-	->  (   Alias == '.'
-	    ->  ShortId = Local
-	    ;   file_name_extension(Base, pl, Local)
-	    ->  ShortId =.. [Alias, Base]
-	    ;   ShortId =.. [Alias, Local]
-	    )
-	;   ShortId = Path
-	).
-	
 %%	short_file_name_to_atom(+ShortId, -Atom)
-%	
+%
 %	Convert a short filename into an atom
 
 short_file_name_to_atom(Atom, Atom) :-
@@ -1746,58 +1738,8 @@ short_file_name_to_atom(Term, Atom) :-
 	term_to_atom(Term, Atom).
 
 
-%%	alias_path(-Alias, ?Dir)
-%	
-%	Enumerate the defined aliases, sorting   them  starting with the
-%	longest path-name, so the first hit is immediately the best one.
-
-:- dynamic
-	alias_cache/2.
-
-alias_path(Alias, Dir) :-
-	(   alias_cache(_, _)
-	->  true
-	;   build_alias_cache
-	),
-	(   nonvar(Dir)
-	->  ensure_slash(Dir, DirSlash),
-	    alias_cache(Alias, DirSlash)
-	;   alias_cache(Alias, Dir)
-	).
-
-build_alias_cache :-
-	findall(t(DirLen, AliasLen, Alias, Dir),
-		search_path(Alias, Dir, AliasLen, DirLen), Ts),
-	sort(Ts, List0),
-	reverse(List0, List),
-	forall(member(t(_, _, Alias, Dir), List),
-	       assert(alias_cache(Alias, Dir))).
-
-search_path('.', Here, 999, DirLen) :-
-	working_directory(Here0, Here0),
-	ensure_slash(Here0, Here),
-	atom_length(Here, DirLen).
-search_path(Alias, Dir, AliasLen, DirLen) :-
-	file_search_path(Alias, _),
-	Spec =.. [Alias,'.'],
-	atom_length(Alias, AliasLen0),
-	AliasLen is 1000 - AliasLen0,	% must do reverse sort
-	absolute_file_name(Spec, Dir0,
-			   [ file_type(directory),
-			     access(read),
-			     solutions(all),
-			     file_errors(fail)
-			   ]),
-	ensure_slash(Dir0, Dir),
-	atom_length(Dir, DirLen).
-
-ensure_slash(Dir, Dir) :-
-	sub_atom(Dir, _, _, 0, /), !.
-ensure_slash(Dir0, Dir) :-
-	atom_concat(Dir0, /, Dir).
-
 %%	library_file(+Path)
-%	
+%
 %	True if Path comes from the Prolog tree and must be considered a
 %	library.
 
@@ -1806,11 +1748,11 @@ library_file(Path) :-
 	sub_atom(Path, 0, _, _, Home).
 
 %%	profile_file(+Path)
-%	
+%
 %	True if path is a personalisation file.  This is a bit hairy.
 
 profile_file(Path) :-
-	short_file_name(Path, user_profile(File)),
+	file_name_on_path(Path, user_profile(File)),
 	known_profile_file(File).
 
 known_profile_file('.plrc').
@@ -1820,7 +1762,7 @@ known_profile_file(File) :-
 	sub_atom(File, 0, _, _, 'lib/xpce/emacs').
 
 %%	sort_files(+Files, -Sorted)
-%	
+%
 %	Sort files, keeping groups comming from the same alias together.
 
 sort_files(Files0, Sorted) :-
@@ -1830,7 +1772,7 @@ sort_files(Files0, Sorted) :-
 	unkey(KSorted, Sorted).
 
 key_file(File, Key-File) :-
-	short_file_name(File, Key).
+	file_name_on_path(File, Key).
 
 
 		 /*******************************
@@ -1838,7 +1780,7 @@ key_file(File, Key-File) :-
 		 *******************************/
 
 %%	available(+File, +Callable, -HowDefined)
-%	
+%
 %	True if Callable is available in File.
 
 available(File, Called, How) :-
@@ -1870,7 +1812,7 @@ available(_, Called, How) :-
 
 
 %%	built_in_predicate(+Callable)
-%	
+%
 %	True if Callable is a built-in
 
 built_in_predicate(Goal) :-
@@ -1879,7 +1821,7 @@ built_in_predicate(Goal) :-
 
 %%	autoload_predicate(+Callable) is semidet.
 %%	autoload_predicate(+Callable, -File) is semidet.
-%	
+%
 %	True if Callable can be autoloaded.  TBD: make sure the autoload
 %	index is up-to-date.
 
@@ -1893,7 +1835,7 @@ autoload_predicate(Goal, File) :-
 
 
 %%	global_predicate(+Callable)
-%	
+%
 %	True if Callable can  be  auto-imported   from  the  global user
 %	module.
 
@@ -1901,7 +1843,7 @@ global_predicate(Goal) :-
 	predicate_property(user:Goal, _), !.
 
 %%	to_predicate_indicator(+Term, -PI)
-%	
+%
 %	Convert to a predicate indicator.
 
 to_predicate_indicator(PI, PI) :-
@@ -1911,7 +1853,7 @@ to_predicate_indicator(Callable, PI) :-
 	predicate_indicator(Callable, PI).
 
 %%	is_predicate_indicator(+PI) is semidet.
-%	
+%
 %	True if PI is a predicate indicator.
 
 is_predicate_indicator(Name/Arity) :-
@@ -1923,7 +1865,7 @@ is_predicate_indicator(Module:Name/Arity) :-
 	integer(Arity).
 
 %%	predicate_indicator(+Callable, -Name)
-%	
+%
 %	Generate a human-readable predicate indicator
 
 predicate_indicator(Module:Goal, PI) :-
@@ -1944,7 +1886,7 @@ hidden_module(M) :-
 	sub_atom(M, 0, _, _, $).
 
 %%	sort_callables(+List, -Sorted)
-%	
+%
 %	Sort list of callable terms.
 
 sort_callables(Callables, Sorted) :-
@@ -1957,7 +1899,7 @@ key_callables([], []).
 key_callables([H0|T0], [Key-H0|T]) :-
 	key_callable(H0, Key),
 	key_callables(T0, T).
-	
+
 key_callable(Callable, k(Name, Arity, Module)) :-
 	predicate_indicator(Callable, PI),
 	(   PI = Name/Arity
@@ -1970,7 +1912,7 @@ unkey([_-H|T0], [H|T]) :-
 	unkey(T0, T).
 
 %%	ord_list_to_set(+OrdList, -OrdSet)
-%	
+%
 %	Removed duplicates (after unification) from an ordered list,
 %	creating a set.
 
@@ -2028,11 +1970,11 @@ edit_callable(Callable, File) :-
 	).
 edit_callable(pce_principal:send_implementation(Id,_,_), _) :-
 	atom(Id),
-	concat_atom([Class,Method], ->, Id), !,
+	atomic_list_concat([Class,Method], ->, Id), !,
 	edit(send(Class, Method)).
 edit_callable(pce_principal:get_implementation(Id,_,_,_), _) :-
 	atom(Id),
-	concat_atom([Class,Method], <-, Id), !,
+	atomic_list_concat([Class,Method], <-, Id), !,
 	edit(get(Class, Method)).
 edit_callable('<directive>'(Line), File) :-
 	File \== @nil, !,
@@ -2054,7 +1996,7 @@ local_callable(Callable, _, Callable).
 		 *******************************/
 
 %%	file_warnings(+File:atom, -Warnings:list(atom))
-%	
+%
 %	Unify Warnings with a list  of   dubious  things  found in File.
 %	Intended to create icons.  Fails if the file is totally ok.
 
@@ -2069,7 +2011,7 @@ file_warning(File, not_called) :-
 
 
 %%	not_called(+File, -Callable)
-%	
+%
 %	Callable is a term defined in File, and for which no callers can
 %	be found.
 
@@ -2088,6 +2030,7 @@ not_called(File, NotCalled) :-		% module version
 	   ;   NotCalled = M:G,
 	       xref_called(ModFile, G),
 	       xref_module(ModFile, M)
+	   ;   generated_callable(Module:NotCalled)
 	   ).
 not_called(File, NotCalled) :-		% non-module version
 	defined(File, NotCalled),
@@ -2101,19 +2044,30 @@ not_called(File, NotCalled) :-		% non-module version
 	   ;   xref_called(AutoImportFile, NotCalled),
 	       \+ defined(AutoImportFile, NotCalled),
 	       global_predicate(NotCalled)
+	   ;   xref_hook(NotCalled)
+	   ;   xref_hook(user:NotCalled)
+	   ;   generated_callable(user:NotCalled)
 	   ).
-	   
-%%	xref_called(?Source, ?Callable) 
-%	
+
+generated_callable(M:Term) :-
+	functor(Term, Name, Arity),
+	prolog:generated_predicate(M:Name/Arity).
+
+%%	xref_called(?Source, ?Callable) is nondet.
+%
 %	True if Callable is called in   Source, after removing recursive
-%	calls.
+%	calls and calls made to predicates where the condition says that
+%	the predicate should not exist.
 
 xref_called(Source, Callable) :-
-	xref_called(Source, Callable, By),
+	xref_called_cond(Source, Callable, _).
+
+xref_called_cond(Source, Callable, Cond) :-
+	xref_called(Source, Callable, By, Cond),
 	By \= Callable.			% recursive calls
 
 %%	defined(?File, ?Callable)
-%	
+%
 %	True if Callable is defined in File and not imported.
 
 defined(File, Callable) :-
@@ -2123,20 +2077,41 @@ defined(File, Callable) :-
 	How \= (multifile).
 
 %%	undefined(+File, -Callable)
-%	
+%
 %	Callable is called in File, but no   definition can be found. If
 %	File is not a module file we   consider other files that are not
 %	module files.
 
 undefined(File, Undef) :-
 	xref_module(File, _), !,
-	xref_called(File, Undef),
-	\+ (  available(File, Undef, How),
-	      How \== plain_file
+	xref_called_cond(File, Undef, Cond),
+	\+ (   available(File, Undef, How),
+	       How \== plain_file
+	   ;   included_if_defined(Cond, Undef)
 	   ).
 undefined(File, Undef) :-
-	xref_called(File, Undef),
-	\+ available(File, Undef, _).
+	xref_called_cond(File, Undef, Cond),
+	\+ (   available(File, Undef, _)
+	   ;   included_if_defined(Cond, Undef)
+	   ).
+
+
+%%	included_if_defined(+Condition, +Callable) is semidet.
+
+included_if_defined(true, _)  :- !.
+included_if_defined(false, _) :- !, fail.
+included_if_defined(fail, _)  :- !, fail.
+included_if_defined(current_predicate(Name/Arity), Callable) :-
+	functor(Callable, Name, Arity), !.
+included_if_defined(\+ Cond, Callable) :- !,
+	\+ included_if_defined(Cond, Callable).
+included_if_defined((A,B), Callable) :- !,
+	included_if_defined(A, Callable),
+	included_if_defined(B, Callable).
+included_if_defined((A;B), Callable) :- !,
+	(   included_if_defined(A, Callable)
+	;   included_if_defined(B, Callable)
+	).
 
 
 		 /*******************************
@@ -2144,20 +2119,20 @@ undefined(File, Undef) :-
 		 *******************************/
 
 %%	file_imports(+File, -Imports)
-%	
+%
 %	Determine which modules must  be  imported   into  this  one. It
 %	considers all called predicates that are   not covered by system
 %	predicates. Next, we have three sources to resolve the remaining
 %	predicates, which are tried in the   order below. The latter two
 %	is dubious.
-%	
+%
 %		* Already existing imports
 %		* Imports from other files in the project
 %		* Imports from the (autoload) library
-%	
+%
 %	We first resolve all imports to   absolute  files. Localizing is
 %	done afterwards.  Imports is a list of
-%	
+%
 %%		use_module(FileSpec, Callables)
 
 xref_file_imports(FileSpec, Imports) :-
@@ -2170,7 +2145,7 @@ xref_file_imports(FileSpec, Imports) :-
 	keysort(ImportList, SortedByFile),
 	merge_by_key(SortedByFile, ImportsByFile),
 	maplist(make_import(File), ImportsByFile, Imports).
-	
+
 canonical_filename(FileSpec, File) :-
 	absolute_file_name(FileSpec,
 			   [ file_type(prolog),
@@ -2221,7 +2196,7 @@ ambiguous_imports([A-C|T0], [A-C|T], UnAmbig, Undef) :-
 
 
 %%	resolve(+Callable, -File)
-%	
+%
 %	Try to find files from which to resolve Callable.
 
 resolve(Callable, File) :-		% Export from module files
@@ -2236,7 +2211,7 @@ resolve(Callable, File) :-		% The Prolog autoload library
 
 
 %%	merge_by_key(+KeyedList, -ListOfKeyValues) is det.
-%	
+%
 %	Example: [a-x, a-y, b-z] --> [a-[x,y], b-[z]]
 
 merge_by_key([], []).
@@ -2250,7 +2225,7 @@ same_key(_, L, [], L).
 
 
 %%	make_import(+RefFile, +ImportList, -UseModules)
-%	
+%
 %	Glues it all together to make a list of directives.
 
 make_import(RefFile, File-Imports, (:-use_module(ShortPath, PIs))) :-
@@ -2259,13 +2234,13 @@ make_import(RefFile, File-Imports, (:-use_module(ShortPath, PIs))) :-
 	maplist(predicate_indicator, SortedImports, PIs).
 
 local_filename(File, RefFile, ShortPath) :-
-	atom(RefFile), 
+	atom(RefFile),
 	file_directory_name(File, Dir),
 	file_directory_name(RefFile, Dir), !,	% i.e. same dir
 	file_base_name(File, Base),
 	remove_extension(Base, ShortPath).
 local_filename(File, _RefFile, ShortPath) :-
-	short_file_name(File, ShortPath0),
+	file_name_on_path(File, ShortPath0),
 	remove_extension(ShortPath0, ShortPath).
 
 
@@ -2293,7 +2268,7 @@ append_row(D, File:name, Callable:prolog, Files:prolog) :->
 	send(D, append, new(FM, menu(file, cycle)), right),
 	send(FM, append, menu_item(@nil, @default, '-- Select --')),
 	forall(member(Path, Files),
-	       (   short_file_name(Path, ShortId),
+	       (   file_name_on_path(Path, ShortId),
 		   short_file_name_to_atom(ShortId, Label),
 		   send(FM, append, menu_item(Path, @default, Label))
 	       )).
@@ -2324,7 +2299,7 @@ cancel(D) :->
 :- pce_end_class(xref_disambiguate_import_dialog).
 
 %%	xref_file_exports(+File, -Exports)
-%	
+%
 %	Produce the export-header for non-module files.  Fails if the
 %	file is already a module file.
 
