@@ -1,4 +1,4 @@
-/*  $Id: swi_help.pl,v 1.28 2006/08/04 14:24:01 jan Exp $
+/*  $Id$
 
     Part of XPCE --- The SWI-Prolog GUI toolkit
 
@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 
     As a special exception, if you link this library with other files,
     compiled with a Free Software compiler, to produce an executable, this
@@ -37,6 +37,11 @@
 	    prolog_explain/1
 	  ]).
 
+/** <module> XPCE-based graphical frontend for online help
+
+This module is normally hooked into help/1 by the module swi_hooks.pl.
+*/
+
 :- use_module(library(pce)).
 :- use_module(library(pce_edit)).
 :- use_module(library(persistent_frame)).
@@ -49,8 +54,8 @@
 	   , atom_length/2
 	   , between/3
 	   , chain_list/2
-	   , concat_atom/2
-	   , concat_atom/3
+	   , atomic_list_concat/2
+	   , atomic_list_concat/3
 	   , default/3
 	   , explain/2
 	   , forall/2
@@ -72,16 +77,27 @@ resource(predicate,	image,  image('16x16/preddoc.xpm')).
 		 *	      TOPLEVEL		*
 		 *******************************/
 
+%%	prolog_help is det.
+%
+%	Open SWI-Prolog graphical reference manual.
+
 prolog_help :-
 	section(S, 'Getting started quickly', _, _),
-	concat_atom(S, -, Id),
+	atomic_list_concat(S, -, Id),
 	term_to_atom(Spec, Id),
 	prolog_help(Spec).
 
+
+%%	prolog_help(+Topic) is semidet.
+%
+%	Open SWI-Prolog graphical reference manual   on  Topic. Fails if
+%	Topic is not in the manual.
+
 prolog_help(Topic) :-
 	help_atom(Topic, Atom),
 	(   atomic(Atom)
-	->  send(@pui_help_window, give_help, Atom)
+	->  once(manual_range(Atom, _)),
+	    send(@pui_help_window, give_help, Atom)
 	;   get(@pui_help_window, member, pui_editor, Editor),
 	    send(Editor, clear),
 	    forall(member(A, Atom),
@@ -90,6 +106,11 @@ prolog_help(Topic) :-
 	),
 	send(@pui_help_window?frame, expose).
 
+
+%%	prolog_apropos(+Keyword) is det.
+%
+%	Do a keyword search on the manual through the object summaries.
+
 prolog_apropos(Keywd) :-
 	send(@pui_help_window, apropos, Keywd),
 	send(@pui_help_window?frame, expose).
@@ -209,7 +230,7 @@ fill_menu_bar(F) :->
 		    menu_item('SWI-Prolog Download (on www) ...',
 			      message(F, open_url, pl_download),
 			      end_group := @on),
-		    
+
 		    menu_item('XPCE (GUI) Manual ...',
 			      message(@prolog, manpce)),
 		    menu_item('XPCE User Guide (on www) ...',
@@ -365,7 +386,7 @@ setup_isearch(V) :->
 				message(@receiver, focus_function, @nil)),
 			    message(@receiver, send_class, '_isearch',
 				    @arg1)))).
-	
+
 
 displayed_ranges(V, Displayed:prolog) :->
 	send(V, slot, displayed_ranges, Displayed),
@@ -389,7 +410,7 @@ give_help(V, What:name,
 	    ;	send(V, clear)
 	    ),
 	    manual_file(ManFile),
-	    new(F, file(ManFile)),
+	    new(F, file(ManFile, binary)),
 	    get(V, caret, Start),
 	    send(F, open, read),
 	    append_ranges(V, F, Ranges),
@@ -416,13 +437,23 @@ give_help(V, What:name,
 	),
 	send(V, editable, @off).
 
+%%	append_ranges(+View, +File, +Ranges) is det.
+%
+%	Note that the file is opened in binary mode to allow seeking. We
+%	must delete \r from the input   to  compensate for Windows cr/lf
+%	line ends. This is all ok as long  as the contents of the manual
+%	file is ISO Latin 1.
+
 append_ranges(V, F, [H|T]) :- !,
 	append_ranges(V, F, H),
 	append_ranges(V, F, T).
 append_ranges(_, _, []) :- !.
-append_ranges(V, F, From-To) :-
+append_ranges(V, F, FromLine-ToLine) :-
+	line_start(FromLine, From),
+	line_start(ToLine, To),
 	send(F, seek, From),
 	get(F, read, To-From, Text),
+	send(Text, translate, 13, @nil),
 	send(V, insert, Text),
 	send(V, insert, @pui_ff),
 	send(V, newline).
@@ -459,7 +490,7 @@ mark_titles(TB, From, To) :-
 	->  mark_titles(TB, Idx, To)
 	;   true
 	).
-	
+
 		 /*******************************
 		 *     CROSS-REFERENCE LINKS	*
 		 *******************************/
@@ -568,27 +599,65 @@ drop(V, Id:name) :->
 		 *******************************/
 
 %%	manual_file(-File)
+%
 %	Find the database file of the manual.  If the manual cannot be
 %	found, display an error message.
 
 manual_file(File) :-
-	absolute_file_name(library('MANUAL'),
-			   [access(read)],
-			   File), !.
+	absolute_file_name(library('MANUAL'), File,
+			   [access(read)]), !.
 manual_file(_File) :-
 	send(@nil, report, error, 'Can''t find manual database MANUAL'),
 	fail.
 
+%%	line_start(Line, Start) is det.
+%
+%	True if Start is the byte position at which Line starts.
+
+:- dynamic
+	start_of_line/2.
+
+line_start(Line, Start) :-
+	start_of_line(Line, Start), !.
+line_start(Line, Start) :-
+	line_index,
+	start_of_line(Line, Start).
+
+
+%%	line_index
+%
+%	Create index holding the byte positions for the line starts
+
+line_index :-
+	start_of_line(_,_), !.
+line_index :-
+	manual_file(File),
+	open(File, read, Stream, [type(binary)]),
+	call_cleanup(line_index(Stream, 1), close(Stream)).
+
+line_index(Stream, LineNo) :-
+	byte_count(Stream, ByteNo),
+	assert(start_of_line(LineNo, ByteNo)),
+	(   at_end_of_stream(Stream)
+	->  true
+	;   LineNo2 is LineNo+1,
+	    skip(Stream, 10),
+	    line_index(Stream, LineNo2)
+	).
+
+
 		 /*******************************
 		 *     ATOMIC SPECIFICATIONS	*
 		 *******************************/
 
-%%	manual_range(+What, -From, -To)
+%%	manual_range(+What, -Range:From-To) is nondet.
+%
 %	Find the character range for the given help topic, which is of
 %	the form
-%			name/arity
-%%			function()
-%			n.m..
+%
+%		* name/arity
+%		* function()
+%		* n.m..
 
 manual_range(What, Ranges) :-
 	atom_codes(What, S),
@@ -727,11 +796,11 @@ alnum(C) -->
 help_atom(Name/Arity, Atom) :- !,
 	(   var(Arity)
 	->  atom_concat(Name, '/_', Atom)
-	;   concat_atom([Name, /, Arity], Atom)
+	;   atomic_list_concat([Name, /, Arity], Atom)
 	).
 help_atom(S1-S0, Atom) :- !,
 	help_atom(S1, A0),
-	concat_atom([A0, '.', S0], Atom).
+	atomic_list_concat([A0, '.', S0], Atom).
 help_atom(C, C) :-
 	integer(C), !.
 help_atom(F, F) :-
@@ -792,7 +861,7 @@ section_apropos(V, Pattern) :-
 	send(V, insert_section,
 	     string('Sections from "%s":', Pattern)),
 	forall(member(Index-Title, Names),
-	       (   concat_atom(Index, '.', Jump),
+	       (   atomic_list_concat(Index, '.', Jump),
 		   append_apropos(V, Jump, Title)
 	       )).
 
@@ -883,8 +952,8 @@ source(File, Line) -->			% XPCE method
 	  get(Location, file_name, File),
 	  get(Location, line_no, Line)
 	}.
-	
-	
+
+
 		 /*******************************
 		 *       TABLE-OF-CONTENTS	*
 		 *******************************/
@@ -962,7 +1031,7 @@ function_in_section(Section, Func) :-
 
 subsection(Sec, Sub) :-
 	append(Sec, [_], Sub).
-	
+
 select_range(PT, Ranges:prolog) :->
 	"Select selections in range"::
 	(   Ranges == []
@@ -974,7 +1043,7 @@ select_range(PT, Ranges:prolog) :->
 	    get(PT, selection, Nodes),
 	    send(PT, normalise, Nodes, y)
 	).
-	    
+
 in_range(F-T, Node, How) :-
 	get(Node, start, S),
 	get(Node, end, E),
@@ -1005,7 +1074,7 @@ initialise(N, Index:prolog) :->
 	->  CanExpand = @on
 	;   CanExpand = @off
 	),
-	concat_atom(Index, '.', Id),
+	atomic_list_concat(Index, '.', Id),
 	send_super(N, initialise, Title?capitalise, Id,
 		   resource(manual), resource(book),
 		   CanExpand),
@@ -1029,7 +1098,7 @@ select_range(N, Ranges:prolog) :->
 	;   true
 	).
 
-:- pce_end_class(pui_section_node). 
+:- pce_end_class(pui_section_node).
 
 
 :- pce_begin_class(pui_predicate_node, toc_file,
