@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 
     As a special exception, if you link this library with other files,
     compiled with a Free Software compiler, to produce an executable, this
@@ -39,7 +39,6 @@
 :- consult([ clause,
 	     util,
 	     source,
-	     break,
 	     gui,
 	     stack
 	   ]).
@@ -53,6 +52,7 @@
 		 *******************************/
 
 :- thread_local
+	finished_frame/1,
 	last_action/1,
 	show_unify_as/2.
 
@@ -67,6 +67,13 @@ user:prolog_trace_interception(Port, Frame, CHP, Action) :-
 	    Action = continue
 	).
 
+:- multifile
+	user:prolog_event_hook/1.
+
+user:prolog_event_hook(frame_finished(Frame)) :-
+	retractall(finished_frame(Frame)),
+	fail.
+
 
 %%	map_action(+GuiAction, +Frame, -Action) is det.
 %
@@ -77,9 +84,8 @@ user:prolog_trace_interception(Port, Frame, CHP, Action) :-
 
 map_action(creep, _, continue) :-
 	traceall.
-map_action(skip, Frame, continue) :-
-	trace,
-	prolog_skip_frame(Frame).
+map_action(skip, _Frame, skip) :-
+	trace.
 map_action(into, _, continue) :-
 	visible(+unify),
 	traceall.
@@ -97,6 +103,7 @@ map_action(halt, _, continue) :-
 	halt.
 map_action(finish, _, continue) :-
 	get_tracer(selected_frame, Frame),
+	asserta(finished_frame(Frame)),
 	trace,
 	prolog_skip_frame(Frame).
 
@@ -113,7 +120,6 @@ traceall :-
 %	Toplevel of the tracer interception.  Runs in debugged thread.
 
 intercept(Port, Frame, CHP, Action) :-
-	send_tracer(current_break(@nil)),
 	debug('*** do_intercept(~w, ~w, ~w, _) ...~n', [Port, Frame, CHP]),
 	do_intercept(Port, Frame, CHP, Action0),
 	fix_action(Port, Action0, Action),
@@ -147,12 +153,12 @@ do_intercept(call, Frame, CHP, Action) :-
 	    action(Action)
 	).
 do_intercept(exit, Frame, CHP, Action) :-
-	(   %last_action(finish)
-	    prolog_frame_attribute(Frame, goal, Goal),
+	(   prolog_frame_attribute(Frame, goal, Goal),
 	    \+(( predicate_property(Goal, nodebug)
 	       ; predicate_property(Goal, foreign)
 	      )),
-	    \+(( last_action(skip),
+	    \+(( prolog_frame_attribute(Frame, skipped, true),
+		 \+ finished_frame(Frame),
 		 prolog_skip_level(L,L),
 		 L \== very_deep,
 		 prolog_frame_attribute(Frame, level, FL),
@@ -208,23 +214,6 @@ do_intercept(unify, Frame, CHP, Action) :-
 	predicate_name(user:Goal, Pred),
 	send_tracer(report(status, '%s: %s', How?label_name, Pred)),
 	action(Action).
-do_intercept(break(PC), Frame, CHP, Action) :-
-	prolog_frame_attribute(Frame, goal, Goal),
-	prolog_frame_attribute(Frame, clause, ClauseRef),
-	'$fetch_vm'(ClauseRef, PC, NPC, _VMI),
-	predicate_name(user:Goal, Pred),
-	send_tracer(report(status, 'Break in: %s', Pred)),
-	send_tracer(current_break(tuple(ClauseRef, PC))),
-	prolog_show_frame(Frame,
-			  [ pc(NPC),
-			    choice(CHP),
-			    port(call),
-			    style(break),
-			    stack,
-			    source,
-			    bindings
-			  ]),
-	action(Action).
 do_intercept(cut_call(PC), Frame, CHP, Action) :-
 	prolog_frame_attribute(Frame, goal, Goal),
 	predicate_name(user:Goal, Pred),
@@ -268,7 +257,11 @@ show(StartFrame, CHP, Up, Port) :-
 	show(StartFrame, CHP, Up, Port, Port),
 	prolog_frame_attribute(StartFrame, goal, Goal),
 	predicate_name(user:Goal, Pred),
-	send_tracer(report(status, '%s: %s', Port?label_name, Pred)).
+	(   Port == redo,
+	    prolog_frame_attribute(StartFrame, skipped, true)
+	->  send_tracer(report(status, '%s: %s (skipped)', Port?label_name, Pred))
+	;   send_tracer(report(status, '%s: %s', Port?label_name, Pred))
+	).
 
 show(StartFrame, CHP, Up, Port, Style) :-
 	find_frame(Up, StartFrame, Port, PC, Frame),
@@ -407,7 +400,14 @@ show_source(Frame, Attributes) :-
 		)
 	    ;	prolog_frame_attribute(GUI, Frame, goal, Goal),
 		qualify(Goal, QGoal),
+		\+ predicate_property(QGoal, foreign),
 		(   clause(QGoal, _Body, ClauseRef)
+		->  subgoal_position(GUI, ClauseRef, unify, File, CharA, CharZ),
+		    send_tracer(GUI, show_range(File, CharA, CharZ, Style))
+		;   functor(Goal, Functor, Arity),
+		    functor(GoalTemplate, Functor, Arity),
+		    qualify(GoalTemplate, QGoalTemplate),
+		    clause(QGoalTemplate, _TBody, ClauseRef)
 		->  subgoal_position(GUI, ClauseRef, unify, File, CharA, CharZ),
 		    send_tracer(GUI, show_range(File, CharA, CharZ, Style))
 		;   find_source(QGoal, File, Line),
@@ -430,6 +430,7 @@ qualify(Goal, user:Goal).
 
 clause_position(PC) :- integer(PC), !.
 clause_position(exit).
+clause_position(unify).
 clause_position(choice(_)).
 
 %%	subgoal_position(+GUI, +Clause, +PortOrPC, -File, -CharA, -CharZ) is det.
@@ -455,10 +456,10 @@ subgoal_position(_, ClauseRef, exit, File, CharA, CharZ) :- !,
 	pce_clause_info(ClauseRef, File, TPos, _),
 	arg(2, TPos, CharA),
 	CharZ is CharA + 1.		% i.e. select the dot.
-subgoal_position(_, ClauseRef, fail, File, CharA, CharZ) :- !,
-	subgoal_position(ClauseRef, exit, File, CharA, CharZ).
-subgoal_position(_, ClauseRef, exception, File, CharA, CharZ) :- !,
-	subgoal_position(ClauseRef, exit, File, CharA, CharZ).
+subgoal_position(GUI, ClauseRef, fail, File, CharA, CharZ) :- !,
+	subgoal_position(GUI, ClauseRef, exit, File, CharA, CharZ).
+subgoal_position(GUI, ClauseRef, exception, File, CharA, CharZ) :- !,
+	subgoal_position(GUI, ClauseRef, exit, File, CharA, CharZ).
 subgoal_position(_, ClauseRef, PC, File, CharA, CharZ) :-
 	pce_clause_info(ClauseRef, File, TPos, _),
 	(   '$clause_term_position'(ClauseRef, PC, List)
@@ -634,14 +635,14 @@ stack_frames(Depth, F, PC, Frames) :-
 	;   RestFrames = []
 	).
 
-%%	choice_frames(+Max, +CHP, +MinLevel-MaxLevel, -Frames) is det.
+%%	choice_frames(+Max, +CHP, +MinMaxLevel, -Frames) is det.
 %
 %	Frames is a list of frames that hold choice-points.
 %
 %	@param Max is the maximum number of choicepoints returned
 %	@param CHP is the initial choicepoint
-%	@param MinLevel-MaxLevel is the depth-range we consider.
-%	       Currently, MaxLevel is ignored (see in_range/2).
+%	@param MinMaxLevel is a pair holding the depth-range we
+%	consider. Currently, MaxLevel is ignored (see in_range/2).
 
 choice_frames(_, none, _, _, []) :- !.
 choice_frames(Max, CHP, Range, Seen, [frame(Frame, choice(CH))|Frames]) :-
@@ -757,12 +758,14 @@ show_args_pc(foreign).
 show_arguments(GUI, Frame, _Attributes) :-
 	get_tracer(GUI, member(bindings), Browser),
 	in_debug_thread(GUI, frame_arguments(Frame, Args)),
+	debug('Frame arguments = ~w~n', [Args]),
 	send(Browser, bindings, Args).
 
 %%	frame_arguments(+Frame, -Args)
 %
-%	Return arguments of the frame as [I:I=Value, ...], compatible with
-%	the normal binding list. Must run in context of debugged thread.
+%	Return arguments of the frame  as [[I:I]=Value, ...], compatible
+%	with the normal binding list. Must   run  in context of debugged
+%	thread.
 
 frame_arguments(Frame, Args) :-
 	prolog_frame_attribute(Frame, goal, Goal),
@@ -772,7 +775,7 @@ frame_arguments(Frame, Args) :-
 	),
 	frame_arguments(1, Arity, Frame, Args).
 
-frame_arguments(I, Arity, Frame, [I:I=Value|T]) :-
+frame_arguments(I, Arity, Frame, [[I:I]=Value|T]) :-
 	I =< Arity, !,
 	prolog_frame_attribute(Frame, argument(I), Value),
 	NI is I + 1,
