@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
 #include <h/kernel.h>
@@ -138,9 +138,7 @@ static int XopenNesting = 0;
 
 static status
 XopenFont(FontObj f, DisplayObj d)
-{ status rc;
-
-  if ( isDefault(d) )
+{ if ( isDefault(d) )
     d = CurrentDisplay(f);
 
   makeBuiltinFonts();
@@ -297,7 +295,7 @@ getSizeFont(FontObj f)
 }
 
 
-Bool
+BoolObj
 getFixedWidthFont(FontObj f)
 { if ( isDefault(f->fixed_width) )
   { getXrefObject(f, CurrentDisplay(NIL));
@@ -312,7 +310,7 @@ getFixedWidthFont(FontObj f)
 }
 
 
-Bool
+BoolObj
 getB16Font(FontObj f)
 { if ( isDefault(f->iswide) )
     XopenFont(f, CurrentDisplay(NIL));
