@@ -1,4 +1,4 @@
-/*  $Id: error.c,v 1.43 2004/10/02 15:50:55 jan Exp $
+/*  $Id$
 
     Part of XPCE --- The SWI-Prolog GUI toolkit
 
@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
 #include <h/kernel.h>
@@ -79,14 +79,17 @@ displayError(Error e, int argc, Any *argv)
 
     sendv(argv[0], NAME_report, argc+2, av);
   } else
-  { char buf[FORMATSIZE];
+  { string s;
+
+    str_writefv(&s, (CharArray) e->format, argc, argv);
 
-    swritefv(buf, NULL, (CharArray) e->format, argc, argv);
-	
     if ( e->kind == NAME_inform || e->kind == NAME_status )
-      Cprintf("[PCE: %s", buf);
+      Cprintf("[PCE: ");
     else
-      Cprintf("[PCE %s: %s", strName(e->kind), buf);
+      Cprintf("[PCE %s: ", strName(e->kind));
+
+    Cputstr(&s);
+    str_unalloc(&s);
 
 #ifndef O_RUNTIME
     if ( e->kind == NAME_fatal ||
@@ -103,18 +106,22 @@ displayError(Error e, int argc, Any *argv)
 #endif
     Cprintf("]\n");
   }
-  
+
   succeed;
 }
 
 
 static StringObj
 getFormatError(Error e, int argc, const Any argv[])
-{ char buf[FORMATSIZE];
+{ string s;
+  StringObj sobj;
 
-  swritefv(buf, NULL, (CharArray)e->format, argc, argv);
+  str_writefv(&s, (CharArray)e->format, argc, argv);
 
-  answer(CtoString(buf));
+  sobj = StringToString(&s);
+  str_unalloc(&s);
+
+  answer(sobj);
 }
 
 
@@ -170,7 +177,7 @@ static getdecl get_error[] =
 #define rc_error NULL
 /*
 static classvardecl rc_error[] =
-{ 
+{
 };
 */
 
@@ -331,6 +338,11 @@ static const error_def errors[] =
   { NAME_noApplicationContext,	ET_FATAL,
     "%N: Failed to create X-application-context" },
 					/* Colour/Cursor/Font, etc */
+  { NAME_noLocaleSupport,	ET_WARNING,
+    "%N: X11 does not support locale %s" },
+  { NAME_cannotSetLocale,	ET_WARNING,
+    "%N: X11: cannot set locale modifiers" },
+
   { NAME_noNamedColour,		ET_WARNING,
     "%N: No colour named %s; using black" },
   { NAME_noNamedCursor,		EF_REPORT,
@@ -381,8 +393,10 @@ static const error_def errors[] =
     "%N: Failed to convert %s.  Trying program default" },
   { NAME_cannotConvertProgramDefault,	ET_FATAL,
     "%N: Failed to convert program default %O" },
-  { NAME_winMetafile, 		ET_WARNING,
+  { NAME_winMetafile,		ET_WARNING,
     "%O: API operation %s failed: %s" },
+  { NAME_x11Threads,		ET_WARNING,
+    "%O: Cannot change X11 threading: display is already open" },
 
 					/* Save/Load */
   { NAME_newSaveVersion,	ET_IGNORED|EF_PRINT,
@@ -548,6 +562,8 @@ static const error_def errors[] =
     "%IReference-count of %O drops below zero" },
   { NAME_stringTooLong,		ET_FATAL,
     "%O: string too long (%d; max = 134217727)" },
+  { NAME_typeNameTooLong,	ET_FATAL,
+    "%O: type name too long (max = %d)" },
   { NAME_maxRecordSize,		0,
     "%O: max record-size is %d" },
 
@@ -559,9 +575,9 @@ static const error_def errors[] =
     "%O%s%s: Unterminated argument list on call from C?" },
 
 					/* consistency-check (object) */
-  { NAME_checkedObjects, 	ET_INFORM|EF_PRINT,
+  { NAME_checkedObjects,	ET_INFORM|EF_PRINT,
     "%IChecked %d objects" },
-  { NAME_noExtension, 		ET_WARNING|EF_PRINT,
+  { NAME_noExtension,		ET_WARNING|EF_PRINT,
     "%O: No attribute of extension %s" },
   { NAME_noProperObject,	ET_WARNING|EF_PRINT,
     "%O: Not a proper object" },
@@ -664,7 +680,7 @@ static const error_def errors[] =
     "%O: Signal trapped: %s" },
   { NAME_createFailed,		0,
     "%O: Failed to ->create" },
-  { NAME_noCharacter,		0,
+  { NAME_noCharacter,		ET_WARNING|EF_REPORT,
     "%O: No character and @event is not printable" },
   { NAME_noKeyBinding,		0,
     "%O: No key_binding named %s" },
@@ -690,7 +706,9 @@ static const error_def errors[] =
   { NAME_notInitialised,	0,
     "%O: Object is not initialised"
   },
-
+  { NAME_unknownEncoding,	0,
+    "%O: Unknown encoding: %s"
+  },
 					/* Tables/layout managament */
   { NAME_spannedRow,	0,	/* tables */
     "%O: Table contains row-spanned cell %O in sort range"
