@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
 :- module(test_text, [test_text/0]).
@@ -40,7 +40,8 @@ test_text :-
 		    atom_to_term,
 		    number_codes,
 		    number_chars,
-		    sub_atom
+		    sub_atom,
+		    atomic_list_concat
 		  ]).
 
 :- begin_tests(char_code).
@@ -205,3 +206,12 @@ test(neg, C = '\235\') :-		% test signed char handling
 	sub_atom('Azi\235\', _, 1, 0, C).
 
 :- end_tests(sub_atom).
+
+
+:- begin_tests(atomic_list_concat).
+
+test(int64, X == 'x-9223372036854775808') :-
+	N is -(1<<63),
+	atomic_list_concat([x, N], X).
+
+:- end_tests(atomic_list_concat).
