@@ -3,9 +3,10 @@
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        J.Wielemaker@uva.nl
+    E-mail:        J.Wielemaker@cs.vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2008, University of Amsterdam
+    Copyright (C): 1985-2011, University of Amsterdam
+			      VU University Amsterdam
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -19,7 +20,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -34,13 +35,9 @@ For compiled arithmetic, the compiler generates WAM codes that execute a
 stack machine.  This module maintains an array of arithmetic  functions.
 These  functions are addressed by the WAM instructions using their index
 in this array.
-
-The  current  version  of  this  module  also  supports  Prolog  defined
-arithmetic  functions.   In  the  current  version these can only return
-numbers.  This should be changed to return arbitrary Prolog  terms  some
-day.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
+/*#define O_DEBUG 1*/
 #include "pl-incl.h"
 #undef abs			/* avoid abs() problem with MSVC++ */
 #include <math.h>
@@ -49,7 +46,7 @@ day.
 #include <limits.h>
 #ifdef HAVE_FLOAT_H
 #include <float.h>
-#ifdef __WINDOWS__
+#ifdef _MSC_VER
 #define isnan(x) _isnan(x)
 #endif
 #endif
@@ -71,16 +68,12 @@ day.
 #define M_E (2.7182818284590452354)
 #endif
 
-#ifdef __WINDOWS__
+#ifdef _MSC_VER
 #define LL(x) x ## i64
 #else
 #define LL(x) x ## LL
 #endif
 
-#ifdef HAVE___TRY
-#include <excpt.h>
-#endif
-
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 On some machines, notably  FreeBSD  upto   version  3.x,  floating point
 operations raise signals rather then leaving an error condition and this
@@ -106,44 +99,7 @@ problem.
 #endif
 #endif
 
-typedef int (*ArithF)();
-
-#define TABLE_REF_MASK		0x1UL
-#define isTableRef(p)		((uintptr_t)(p) & TABLE_REF_MASK)
-#define makeTableRef(p)		((void*)((uintptr_t)(p) | TABLE_REF_MASK))
-#define unTableRef(s, p)	(*((s*)((uintptr_t)(p) & ~TABLE_REF_MASK)))
-
-#define return_next_table(t, v, clean) \
-	{ for((v) = (v)->next; isTableRef(v) && (v); (v) = unTableRef(t, v)) \
-	  if ( (v) == (t)NULL ) \
-	  { clean; \
-	    succeed; \
-	  } \
-	  ForeignRedoPtr(v); \
-	}
-
-
-struct arithFunction
-{ ArithFunction next;		/* Next of chain */
-  functor_t	functor;	/* Functor defined */
-  ArithF	function;	/* Implementing function */
-  Module	module;		/* Module visibility module */
-  int		level;		/* Level of the module */
-#if O_PROLOG_FUNCTIONS
-  Procedure	proc;		/* Prolog defined functions */
-#endif
-#if O_COMPILE_ARITH
-  code		index;		/* Index of function */
-#endif
-};
-
-#define arithFunctionTable	(GD->arith.table)
-#define function_array		(&GD->arith.functions)
-#define FunctionFromIndex(n)	fetchBuffer(function_array, n, ArithFunction)
-
-static ArithFunction	isCurrentArithFunction(functor_t, Module);
-static int		registerFunction(ArithFunction f, int index);
-static int		getCharExpression(term_t t, Number r ARG_LD);
+static int		getCharExpression(Word p, Number r ARG_LD);
 static int		ar_minus(Number n1, Number n2, Number r);
 
 
@@ -161,7 +117,7 @@ clearInteger(Number n)
 }
 
 
-typedef struct
+typedef struct between_state
 { number low;
   number high;
   int hinf;
@@ -228,13 +184,14 @@ PRED_IMPL("between", 3, between, PL_FA_NONDETERMINISTIC)
 	  succeed;
 	}
 
-	state = allocHeap(sizeof(*state));
+	state = allocForeignState(sizeof(*state));
 	cpNumber(&state->low, &l);
 	cpNumber(&state->high, &h);
 	state->hinf = hinf;
 	clearInteger(&l);
 	clearInteger(&h);
 	ForeignRedoPtr(state);
+	/*NOTREACHED*/
       }
     case FRG_REDO:
       { state = CTX_PTR;
@@ -248,13 +205,15 @@ PRED_IMPL("between", 3, between, PL_FA_NONDETERMINISTIC)
 	     cmpNumbers(&state->low, &state->high) == 0 )
 	  goto cleanup;
 	ForeignRedoPtr(state);
+	/*NOTREACHED*/
       }
     case FRG_CUTTED:
       { state = CTX_PTR;
       cleanup:
 	clearInteger(&state->low);
 	clearInteger(&state->high);
-	freeHeap(state, sizeof(*state));
+	freeForeignState(state, sizeof(*state));
+	/*FALLTHROUGH*/
       }
     default:;
       return rc;
@@ -457,37 +416,37 @@ compareNumbers(term_t n1, term_t n2, int what ARG_LD)
 }
 
 static
-PRED_IMPL("<", 2, lt, PL_FA_TRANSPARENT)
+PRED_IMPL("<", 2, lt, PL_FA_ISO)
 { PRED_LD
   return compareNumbers(A1, A2, LT PASS_LD);
 }
 
 static
-PRED_IMPL(">", 2, gt, PL_FA_TRANSPARENT)
+PRED_IMPL(">", 2, gt, PL_FA_ISO)
 { PRED_LD
   return compareNumbers(A1, A2, GT PASS_LD);
 }
 
 static
-PRED_IMPL("=<", 2, leq, PL_FA_TRANSPARENT)
+PRED_IMPL("=<", 2, leq, PL_FA_ISO)
 { PRED_LD
   return compareNumbers(A1, A2, LE PASS_LD);
 }
 
 static
-PRED_IMPL(">=", 2, geq, PL_FA_TRANSPARENT)
+PRED_IMPL(">=", 2, geq, PL_FA_ISO)
 { PRED_LD
   return compareNumbers(A1, A2, GE PASS_LD);
 }
 
 static
-PRED_IMPL("=\\=", 2, neq, PL_FA_TRANSPARENT)
+PRED_IMPL("=\\=", 2, neq, PL_FA_ISO)
 { PRED_LD
   return compareNumbers(A1, A2, NE PASS_LD);
 }
 
 static
-PRED_IMPL("=:=", 2, eq, PL_FA_TRANSPARENT)
+PRED_IMPL("=:=", 2, eq, PL_FA_ISO)
 { PRED_LD
   return compareNumbers(A1, A2, EQ PASS_LD);
 }
@@ -582,92 +541,21 @@ freeArithLocalData(PL_local_data_t *ld)
 		*********************************/
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-isCurrentArithFunction(functor_t f, Module m)
-    Find existing arithmetic function definition for f using m as
-    context.
-
-    The one we are looking for is the function that is in the most
-    local module.  As the entries are sorted such that more specific
-    functions are before global functions, we can pick the first
-    one that is in our module path.
+isCurrentArithFunction(functor_t f)
+    Find existing arithmetic function definition for f.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
+static inline ArithF
+isCurrentArithFunction(functor_t f)
+{ size_t index = indexFunctor(f);
 
-static ArithFunction
-isCurrentArithFunction(functor_t f, Module m)
-{ ArithFunction a;
-
-  for(a = arithFunctionTable[functorHashValue(f, ARITHHASHSIZE)];
-      !isTableRef(a) && a; a = a->next)
-  { if ( a->functor == f )
-    { if ( isSuperModule(a->module, m) )
-	return a;
-    }
+  if ( index < GD->arith.functions_allocated )
+  { return GD->arith.functions[index];
   }
 
   return NULL;
 }
 
-#if O_PROLOG_FUNCTIONS
-
-/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-Activating a Prolog predicate as function below the arithmetic functions
-is/0, >, etc. `f' is the arithmetic function   to  be called. `t' is the
-base term-reference of an array holding  the proper number of arguments.
-`r' is the result of the evaluation.
-
-This calling convention is somewhat  unnatural,   but  fits  best in the
-calling convention required by ar_func_n() below.
-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
-
-static int
-prologFunction(ArithFunction f, term_t av, Number r ARG_LD)
-{ Definition def = getProcDefinition(f->proc);
-  int arity = def->functor->arity;
-  fid_t fid;
-  qid_t qid;
-  int rval;
-
-  if ( !(fid=PL_open_foreign_frame()) ||
-       !(qid=PL_open_query(NULL, PL_Q_PASS_EXCEPTION, f->proc, av)) )
-    return FALSE;
-
-  if ( PL_next_solution(qid) )
-  { rval = valueExpression(av+arity-1, r PASS_LD);
-    PL_close_query(qid);
-    PL_discard_foreign_frame(fid);
-  } else
-  { term_t except;
-
-    if ( (except = PL_exception(qid)) )
-    { rval = FALSE;
-      PL_close_query(qid);
-    } else
-    { PL_close_query(qid);
-
-#ifdef O_LIMIT_DEPTH
-      if ( depth_reached > depth_limit )
-      { rval = FALSE;
-      } else
-#endif
-      { term_t goal;
-
-	rval = ( (goal = PL_new_term_ref()) &&
-		 PL_cons_functor_v(goal, def->functor->functor, av) &&
-		 PL_error(NULL, 0,
-			  "Aritmetic function must succeed or throw exception",
-			  ERR_FAILED, goal)
-	       );
-      }
-    }
-
-    PL_close_foreign_frame(fid);
-  }
-
-  return rval;
-}
-
-#endif /* O_PROLOG_FUNCTIONS */
 
 int
 check_float(double f)
@@ -728,184 +616,251 @@ check_float(double f)
 }
 
 
-int
-eval_expression(term_t t, Number r, int recursion ARG_LD)
-{ ArithFunction f;
-  functor_t functor;
-  Word p = valTermRef(t);
-  word w;
+		 /*******************************
+		 *	     EVALULATE		*
+		 *******************************/
 
-  deRef(p);
-  w = *p;
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+valueExpression() evaluates an `evaluable term'.
 
-  switch(tag(w))
-  { case TAG_INTEGER:
-      get_integer(w, r);
-      succeed;
-    case TAG_FLOAT:
-      r->value.f = valFloat(w);
-      r->type = V_FLOAT;
-      succeed;
-    case TAG_VAR:
-      return PL_error(NULL, 0, NULL, ERR_INSTANTIATION);
-    case TAG_ATOM:
-      functor = lookupFunctorDef(w, 0);
-      break;
-    case TAG_COMPOUND:
-      functor = functorTerm(w);
-      break;
-    default:
-      return PL_error(NULL, 0, NULL, ERR_TYPE, ATOM_number, t);
-  }
+This new implementation avoids using the C-stack   to be able to process
+more deeply nested terms and to be able  to recover in the unlikely case
+that terms are still too deeply nested.
 
-  if ( !(f = isCurrentArithFunction(functor,
-				    contextModule(environment_frame))))
-  { if ( functor == FUNCTOR_dot2 )	/* handle "a" (make function) */
-      return getCharExpression(t, r PASS_LD);
-    else
-      return PL_error(NULL, 0, NULL, ERR_NOT_EVALUABLE, functor);
-  }
+If finds a term, it starts processing at the last argument, working back
+to the start. It it finds  the   functor  itself it evaluates the pushed
+arguments. Using this technique we push as  few as possible arguments on
+terms that are nested on the left (as   in (1+2)+3, while we only push a
+single pointer for each recursion level in the evaluable term.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-  if ( recursion == 100 && !PL_is_acyclic(t) )
-    return PL_error(NULL, 0, "cyclic term", ERR_TYPE, ATOM_expression, t);
+static int
+pushForMark(segstack *stack, Word p, int wr)
+{ word w = ((word)p)|wr;
 
-#if O_PROLOG_FUNCTIONS
-  if ( f->proc )
-  { fid_t fid;
+  return pushSegStack(stack, w, word);
+}
 
-    if ( (fid=PL_open_foreign_frame()) )
-    { int rval, n, arity = arityFunctor(functor);
-      term_t h0 = PL_new_term_refs(arity+1); /* one extra for the result */
+static void
+popForMark(segstack *stack, Word *pp, int *wr)
+{ word w;
 
-      if ( !h0 )
-	return FALSE;
-
-      for(n=0; n<arity; n++)
-      { number n1;
-
-	_PL_get_arg(n+1, t, h0+n);
-	if ( eval_expression(h0+n, &n1, recursion+1 PASS_LD) )
-	{ _PL_put_number(h0+n, &n1);
-	  clearNumber(&n1);
-	} else
-	{ PL_close_foreign_frame(fid);
-	  fail;
-	}
-      }
-
-      rval = prologFunction(f, h0, r PASS_LD);
-      PL_close_foreign_frame(fid);
-
-      return rval;
-    }
-
-    return FALSE;
-  }
-#endif
-
-  DEBUG(3, Sdprintf("Starting __try ...\n"));
-
-  { int rval;
-
-#ifdef HAVE___TRY
-#ifndef EXCEPTION_EXECUTE_HANDLER	/* lcc */
-#define EXCEPTION_EXECUTE_HANDLER 1
-#endif
-    __try
-    {
-#else
-    LD->in_arithmetic++;
-#endif
-    switch(arityFunctor(functor))
-    { case 0:
-	rval = (*f->function)(r);
-        break;
-      case 1:
-      { term_t a = PL_new_term_ref();
-	number n1;
-
-	_PL_get_arg(1, t, a);
-	if ( eval_expression(a, &n1, recursion+1 PASS_LD) )
-	{ rval = (*f->function)(&n1, r);
-	  clearNumber(&n1);
-	} else
-	  rval = FALSE;
-
-	PL_reset_term_refs(a);
-	break;
-      }
-      case 2:
-      { term_t a = PL_new_term_ref();
-	number n1, n2;
-
-	_PL_get_arg(1, t, a);
-	if ( eval_expression(a, &n1, recursion+1 PASS_LD) )
-	{ _PL_get_arg(2, t, a);
-	  if ( eval_expression(a, &n2, recursion+1 PASS_LD) )
-	  { rval = (*f->function)(&n1, &n2, r);
-	    clearNumber(&n2);
-	  } else
-	  { rval = FALSE;
-	  }
-	  clearNumber(&n1);
-	} else
-	  rval = FALSE;
-
-	PL_reset_term_refs(a);
-	break;
-      }
-      case 3:
-      { term_t a = PL_new_term_ref();
-	number n1, n2, n3;
-
-	_PL_get_arg(1, t, a);
-	if ( eval_expression(a, &n1, recursion+1 PASS_LD) )
-	{ _PL_get_arg(2, t, a);
-	  if ( eval_expression(a, &n2, recursion+1 PASS_LD) )
-	  { _PL_get_arg(3, t, a);
-	    if ( eval_expression(a, &n3, recursion+1 PASS_LD) )
-	    { rval = (*f->function)(&n1, &n2, &n3, r);
-	      clearNumber(&n3);
-	    } else
-	    { rval = FALSE;
-	    }
-	    clearNumber(&n2);
-	  } else
-	  { rval = FALSE;
-	  }
-	  clearNumber(&n1);
-	} else
-	  rval = FALSE;
-
-	PL_reset_term_refs(a);
-	break;
-      }
-      default:
-	sysError("Illegal arity for arithmic function");
-        rval = FALSE;
-    }
-#if defined(HAVE___TRY)
-    } __except(EXCEPTION_EXECUTE_HANDLER)
-    { warning("Floating point exception");
-#ifndef O_RUNTIME
-      Sfprintf(Serror, "[PROLOG STACK:\n");
-      backTrace(NULL, 10);
-      Sfprintf(Serror, "]\n");
-#endif
-      rval = abortProlog(ABORT_RAISE);
-    }
-#else /*HAVE___TRY*/
-    LD->in_arithmetic--;
-#endif /*HAVE___TRY*/
-
-    return rval;
-  }
+  popSegStack(stack, &w, word);
+  *wr = w & (word)0x1;
+  *pp = (Word)(w & ~(word)0x1);
 }
 
 
 int
-valueExpression(term_t t, Number r ARG_LD)
-{ return eval_expression(t, r, 0 PASS_LD);
+valueExpression(term_t expr, number *result ARG_LD)
+{ segstack term_stack;
+  segstack arg_stack;
+  Word term_buf[16];
+  number arg_buf[16];
+  number *n = result;
+  number n_tmp;
+  int walk_ref = FALSE;
+  Word p = valTermRef(expr);
+  Word start;
+  int known_acyclic = FALSE;
+  int pushed = 0;
+
+  deRef(p);
+  start = p;
+  LD->in_arithmetic++;
+
+  for(;;)
+  { switch(tag(*p))
+    { case TAG_INTEGER:
+	get_integer(*p, n);
+        break;
+      case TAG_FLOAT:
+	n->value.f = valFloat(*p);
+        n->type = V_FLOAT;
+	break;
+      case TAG_VAR:
+	PL_error(NULL, 0, NULL, ERR_INSTANTIATION);
+        goto error;
+      case TAG_REFERENCE:
+      { if ( !pushForMark(&term_stack, p, walk_ref) )
+	{ PL_no_memory();
+	  goto error;
+	}
+	walk_ref = TRUE;
+	deRef(p);
+	continue;
+      }
+      case TAG_ATOM:
+      { functor_t functor = lookupFunctorDef(*p, 0);
+	ArithF f;
+
+        if ( (f = isCurrentArithFunction(functor)) )
+	{ if ( (*f)(n) != TRUE )
+	    goto error;
+	} else
+	{ PL_error(NULL, 0, NULL, ERR_NOT_EVALUABLE, functor);
+	  goto error;
+	}
+	break;
+      }
+      case TAG_COMPOUND:
+      { Functor term = valueTerm(*p);
+	int arity;
+
+	if ( term->definition == FUNCTOR_dot2 )
+	{ if ( getCharExpression(p, n PASS_LD) != TRUE )
+	    goto error;
+	  break;
+	}
+
+	if ( p == start )
+	{ initSegStack(&term_stack, sizeof(Word), sizeof(term_buf), term_buf);
+	  initSegStack(&arg_stack, sizeof(number), sizeof(arg_buf), arg_buf);
+	}
+
+	if ( !pushForMark(&term_stack, p, walk_ref) )
+	{ PL_no_memory();
+	  goto error;
+	}
+	if ( ++pushed > 100 && !known_acyclic )
+	{ int rc;
+
+	  if ( (rc=is_acyclic(start PASS_LD)) == TRUE )
+	  { known_acyclic = TRUE;
+	  } else
+	  { if ( rc == MEMORY_OVERFLOW )
+	      PL_error(NULL, 0, NULL, ERR_NOMEM);
+	    else
+	      PL_error(NULL, 0, "cyclic term", ERR_TYPE, ATOM_expression, expr);
+	    goto error;
+	  }
+	}
+	walk_ref = FALSE;
+	n = &n_tmp;
+	arity = arityFunctor(term->definition);
+	p = &term->arguments[arity-1];
+	continue;
+      }
+      default:
+	PL_error(NULL, 0, NULL, ERR_PTR_TYPE, ATOM_number, p);
+        goto error;
+    }
+
+    if ( p == start )
+    { LD->in_arithmetic--;
+      assert(n == result);
+
+      return TRUE;
+    }
+
+    if ( walk_ref )
+      popForMark(&term_stack, &p, &walk_ref);
+
+    if ( !pushSegStack(&arg_stack, n_tmp, number) )
+    { PL_no_memory();
+      goto error;
+    }
+
+    while ( tagex(*--p) == (TAG_ATOM|STG_GLOBAL) )
+    { functor_t functor = *p;
+      ArithF f;
+
+      DEBUG(1, Sdprintf("Eval %s/%d\n",
+			stringAtom(nameFunctor(functor)),
+			arityFunctor(functor)));
+
+      if ( (f = isCurrentArithFunction(functor)) )
+      { int arity = arityFunctor(functor);
+
+	switch(arity)
+	{ case 1:
+	  { int rc;
+	    number *a0 = topOfSegStack(&arg_stack);
+
+	    rc = (*f)(a0, n);
+	    clearNumber(a0);
+	    if ( rc == TRUE )
+	    { *a0 = *n;
+	    } else
+	    { popTopOfSegStack(&arg_stack);
+	      goto error;
+	    }
+
+	    break;
+	  }
+	  case 2:
+	  { int rc;
+	    void *a[2];
+
+	    topsOfSegStack(&arg_stack, 2, a);
+	    rc = (*f)((number*)a[0], (number*)a[1], n);
+	    clearNumber((number*)a[0]);
+	    clearNumber((number*)a[1]);
+	    popTopOfSegStack(&arg_stack);
+
+	    if ( rc == TRUE )
+	    { number *n1 = a[1];
+	      *n1 = *n;
+	    } else
+	    { popTopOfSegStack(&arg_stack);
+	      goto error;
+	    }
+
+	    break;
+	  }
+	  case 3:
+	  { int rc;
+	    void *a[3];
+
+	    topsOfSegStack(&arg_stack, 3, a);
+	    rc = (*f)((number*)a[0], (number*)a[1], (number*)a[2], n);
+	    clearNumber((number*)a[0]);
+	    clearNumber((number*)a[1]);
+	    clearNumber((number*)a[2]);
+	    popTopOfSegStack(&arg_stack);
+	    popTopOfSegStack(&arg_stack);
+
+	    if ( rc == TRUE )
+	    { number *n2 = a[2];
+	      *n2 = *n;
+	    } else
+	    { popTopOfSegStack(&arg_stack);
+	      goto error;
+	    }
+
+	    break;
+	  }
+	  default:
+	    assert(0);
+	}
+
+	popForMark(&term_stack, &p, &walk_ref);
+	if ( p == start )
+	{ LD->in_arithmetic--;
+	  *result = *n;
+
+	  return TRUE;
+	}
+	if ( walk_ref )
+	  popForMark(&term_stack, &p, &walk_ref);
+      } else
+      { PL_error(NULL, 0, NULL, ERR_NOT_EVALUABLE, functor);
+	goto error;
+      }
+    }
+  }
+
+error:
+  if ( p != start )
+  { number n;
+
+    clearSegStack(&term_stack);
+    while( popSegStack(&arg_stack, &n, number) )
+      clearNumber(&n);
+  }
+  LD->in_arithmetic--;
+
+  return FALSE;
 }
 
 
@@ -936,15 +891,16 @@ arithChar(Word p ARG_LD)
   }
 
   PL_error(NULL, 0, NULL, ERR_TYPE,
-	   ATOM_character, wordToTermRef(p));
+	   ATOM_character, pushWordAsTermRef(p));
+  popTermRef();
 
   return EOF;
 }
 
 
 static int
-getCharExpression(term_t t, Number r ARG_LD)
-{ Word a, p = valTermRef(t);
+getCharExpression(Word p, Number r ARG_LD)
+{ Word a;
   int chr;
 
   deRef(p);
@@ -955,8 +911,11 @@ getCharExpression(term_t t, Number r ARG_LD)
 
   a = argTermP(*p, 1);
   if ( !isNil(*a) )
-    return PL_error(".", 2, "\"x\" must hold one character", ERR_TYPE,
-		    ATOM_nil, wordToTermRef(a));
+  { PL_error(".", 2, "\"x\" must hold one character", ERR_TYPE,
+	     ATOM_nil, pushWordAsTermRef(a));
+    popTermRef();
+    return FALSE;
+  }
 
   r->value.i = chr;
   r->type = V_INTEGER;
@@ -1366,7 +1325,16 @@ ar_shift(Number n1, Number n2, Number r, int dir)
       if ( dir < 0 )			/* shift left (<<) */
       {
 #ifdef O_GMP				/* msb() is 0..63 */
-	if ( msb64(n1->value.i) + shift >= (int)(sizeof(int64_t)*8-1) )
+        int bits = shift;
+
+	if ( n1->value.i >= 0 )
+	  bits += msb64(n1->value.i);
+	else if ( n1->value.i == PLMININT )
+	  bits += sizeof(int64_t)*8;
+	else
+	  bits += msb64(-n1->value.i);
+
+	if ( bits >= (int)(sizeof(int64_t)*8-1) )
 	{ promoteToMPZNumber(n1);
 	  goto mpz;
 	} else
@@ -1375,7 +1343,7 @@ ar_shift(Number n1, Number n2, Number r, int dir)
 	}
       } else
       { if ( shift >= (long)sizeof(int64_t)*8 )
-	  r->value.i = 0;
+	  r->value.i = (r->value.i >= 0 ? 0 : -1);
 	else
 	  r->value.i = n1->value.i >> shift;
       }
@@ -2096,9 +2064,14 @@ mul64(int64_t x, int64_t y, int64_t *r)
 
     As division is pretty expensive, we make a quick test to see whether
     we are in the danger zone.
+
+    Finally, we must avoid INT64_MIN/-1 :-(
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 #define MU64_SAFE_MAX (LL(1)<<30)
+#ifndef INT64_MIN
+#define INT64_MIN (LL(1)<<63)
+#endif
 
 static int
 mul64(int64_t x, int64_t y, int64_t *r)
@@ -2133,11 +2106,15 @@ mul64(int64_t x, int64_t y, int64_t *r)
     prod = (int64_t)(ax*ay);
     if ( sign < 0 )
       prod = -prod;
-    if ( (ax < MU64_SAFE_MAX && ay < MU64_SAFE_MAX) || prod/y == x )
+    if ( (ax < MU64_SAFE_MAX && ay < MU64_SAFE_MAX) )
     { *r = prod;
-
       return TRUE;
     }
+    if ( !(y==LL(-1) && prod == INT64_MIN) && prod/y == x )
+    { *r = prod;
+      return TRUE;
+    }
+
     return FALSE;
   }
 }
@@ -2276,25 +2253,15 @@ ar_negation(Number n1, Number r)
 }
 
 
-static int
-domainErrorNumber(const char *f, int a, Number n, atom_t error)
-{ GET_LD
-  term_t t = PL_new_term_ref();
-
-  PL_unify_number(t, n);
-  return PL_error(f, a, NULL, ERR_DOMAIN, error, t);
-}
-
-
 static int
 notLessThanZero(const char *f, int a, Number n)
-{ return domainErrorNumber(f, a, n, ATOM_not_less_than_zero);
+{ return PL_error(f, a, NULL, ERR_AR_DOMAIN, ATOM_not_less_than_zero, n);
 }
 
 
 static int
 mustBePositive(const char *f, int a, Number n)
-{ return domainErrorNumber(f, a, n, ATOM_not_less_than_one);
+{ return PL_error(f, a, NULL, ERR_AR_DOMAIN, ATOM_not_less_than_one, n);
 }
 
 
@@ -2432,7 +2399,7 @@ ar_u_minus(Number n1, Number r)
 	promoteToMPZNumber(n1);
 	r->type = V_MPZ;
 #else
-  	if ( !promoteToFloatNumber(n1) )
+	if ( !promoteToFloatNumber(n1) )
 	  return FALSE;
 	r->type = V_FLOAT;
 #endif
@@ -2818,7 +2785,7 @@ seed_from_dev(const char *dev ARG_LD)
 #ifdef S_ISCHR
   int fd;
 
-  if ( (fd=open("/dev/urandom", O_RDONLY)) )
+  if ( (fd=open(dev, O_RDONLY)) )
   { struct stat buf;
 
     if ( fstat(fd, &buf) == 0 && S_ISCHR(buf.st_mode) )
@@ -2835,12 +2802,12 @@ seed_from_dev(const char *dev ARG_LD)
       }
 
       if ( rd >= MIN_RAND_SEED_LEN )
-      { DEBUG(1, Sdprintf("Seed random using %ld bytes from /dev/random\n",
-			  (long)n));
+      { DEBUG(1, Sdprintf("Seed random using %ld bytes from %s\n",
+			  (long)rd, dev));
 
 	LD->gmp.persistent++;
 	mpz_init(seed);
-	mpz_import(seed, n, 1, sizeof(char), 0, 0, seedarray);
+	mpz_import(seed, rd, 1, sizeof(char), 0, 0, seedarray);
 	gmp_randseed(LD->arith.random.state, seed);
 	mpz_clear(seed);
 	LD->gmp.persistent--;
@@ -2857,16 +2824,63 @@ seed_from_dev(const char *dev ARG_LD)
 }
 
 
+
+static int
+seed_from_crypt_context(ARG1_LD)
+{
+#ifdef _MSC_VER
+  HCRYPTPROV hCryptProv;
+  char *user_name = "seed_random";
+  BYTE seedarray[RAND_SEED_LEN];
+  mpz_t seed;
+
+
+  if ( CryptAcquireContext(&hCryptProv, user_name, NULL, PROV_RSA_FULL, 0) )
+  { CryptGenRandom(hCryptProv, sizeof(seedarray), seedarray);
+  } else if ( (GetLastError() == NTE_BAD_KEYSET) &&
+	      CryptAcquireContext(&hCryptProv, user_name, NULL,
+				  PROV_RSA_FULL, CRYPT_NEWKEYSET) )
+  { CryptGenRandom(hCryptProv, sizeof(seedarray), seedarray);
+  } else
+  { return FALSE;
+  }
+
+  LD->gmp.persistent++;
+  mpz_init(seed);
+  mpz_import(seed, RAND_SEED_LEN, 1, sizeof(BYTE), 0, 0, seedarray);
+  gmp_randseed(LD->arith.random.state, seed);
+  mpz_clear(seed);
+  LD->gmp.persistent--;
+
+  return TRUE;
+#else
+#ifdef O_PLMT
+  (void)__PL_ld;
+#endif
+  return FALSE;
+#endif
+}
+
+
 static void
 seed_random(ARG1_LD)
 { if ( !seed_from_dev("/dev/urandom" PASS_LD) &&
-       !seed_from_dev("/dev/random" PASS_LD) )
-  { LD->gmp.persistent++;
-    gmp_randseed_ui(LD->arith.random.state,
-		    (unsigned long)time(NULL));
+       !seed_from_dev("/dev/random" PASS_LD) &&
+       !seed_from_crypt_context(PASS_LD1) )
+  { double t[1] = { WallTime() };
+    unsigned long key = 0;
+    unsigned long *p = (unsigned long*)t;
+    unsigned long *e = (unsigned long*)&t[1];
+
+    for(; p<e; p++)
+      key ^= *p;
+
+    LD->gmp.persistent++;
+    gmp_randseed_ui(LD->arith.random.state, key);
     LD->gmp.persistent--;
   }
 }
+
 #else /* O_GMP */
 
 static void
@@ -2882,7 +2896,7 @@ init_random(ARG1_LD)
 #ifdef O_GMP
   if ( !LD->arith.random.initialised )
   { LD->gmp.persistent++;
-    gmp_randinit_default(LD->arith.random.state);
+    gmp_randinit_mt(LD->arith.random.state);
     LD->arith.random.initialised = TRUE;
     seed_random(PASS_LD1);
     LD->gmp.persistent--;
@@ -2908,6 +2922,7 @@ PRED_IMPL("set_random", 1, set_random, 0)
 
       if ( PL_get_atom(arg, &a) && a == ATOM_random )
       { seed_random(PASS_LD1);
+	return TRUE;
       } else
       { number n;
 
@@ -2920,30 +2935,74 @@ PRED_IMPL("set_random", 1, set_random, 0)
 	  case V_INTEGER:
 	    gmp_randseed_ui(LD->arith.random.state,
 			    (unsigned long)n.value.i);
-	    break;
+	    return TRUE;
 	  case V_MPZ:
 	    gmp_randseed(LD->arith.random.state, n.value.mpz);
-	    break;
+	    return TRUE;
 #else
 	  case V_INTEGER:
           { unsigned int seed = (unsigned int)n.value.i;
 	    setRandom(&seed);
-	    break;
+	    return TRUE;
 	  }
 #endif
 	  default:
-	    PL_error(NULL, 0, NULL, ERR_TYPE, ATOM_seed, a);
+	    return PL_error(NULL, 0, NULL, ERR_TYPE, ATOM_seed, a);
 	}
       }
+#ifdef O_GMP
+    } else if ( name == ATOM_state )
+    { number n;
+
+      if ( !PL_get_number(arg, &n) ||
+	   n.type != V_MPZ )
+	return PL_error(NULL, 0, NULL, ERR_TYPE, ATOM_state, arg);
+
+      mpz_set(LD->arith.random.state[0]._mp_seed, n.value.mpz);
+      clearNumber(&n);
+
+      return TRUE;
+#endif /*O_GMP*/
     } else
-    { PL_error(NULL, 0, NULL, ERR_DOMAIN, ATOM_random_option, A1);
+    { return PL_error(NULL, 0, NULL, ERR_DOMAIN, ATOM_random_option, A1);
     }
   } else
-  { PL_error(NULL, 0, NULL, ERR_TYPE, ATOM_random_option, A1);
+  { return PL_error(NULL, 0, NULL, ERR_TYPE, ATOM_random_option, A1);
+  }
+}
+
+
+#ifdef O_GMP
+static
+PRED_IMPL("random_property", 1, random_property, 0)
+{ PRED_LD
+  atom_t name;
+  int arity;
+
+  init_random(PASS_LD1);
+
+  if ( PL_get_name_arity(A1, &name, &arity) && arity == 1 )
+  { term_t arg = PL_new_term_ref();
+
+    _PL_get_arg(1, A1, arg);
+    if ( name == ATOM_state )
+    { int rc;
+      number seed;
+
+      seed.type = V_MPZ;
+      mpz_init(seed.value.mpz);
+      LD->arith.random.state[0]._mp_seed[0]._mp_size =
+      LD->arith.random.state[0]._mp_seed[0]._mp_alloc;					      mpz_set(seed.value.mpz, LD->arith.random.state[0]._mp_seed);
+      rc = PL_unify_number(arg, &seed);
+      clearNumber(&seed);
+
+      return rc;
+    }
   }
 
-  succeed;
+  return FALSE;
 }
+#endif
 
 
 static int
@@ -2986,6 +3045,31 @@ ar_random(Number n1, Number r)
   }
 }
 
+#ifndef UINT64_MAX
+#define UINT64_MAX (~(uint64_t)0)
+#endif
+
+static int
+ar_random_float(Number r)
+{ GET_LD
+
+  init_random(PASS_LD1);
+
+#ifdef O_GMP
+{ mpf_t rop;
+  mpf_init2(rop, sizeof(double)*8);
+  mpf_urandomb(rop, LD->arith.random.state, sizeof(double)*8);
+  r->value.f = mpf_get_d(rop);
+  mpf_clear(rop);
+}
+#else
+  r->value.f = _PL_Random()/(float)UINT64_MAX;
+#endif
+
+  r->type = V_FLOAT;
+  succeed;
+}
+
 
 static int
 ar_pi(Number r)
@@ -3028,7 +3112,7 @@ ar_cputime(Number r)
 		*********************************/
 
 static
-PRED_IMPL("is", 2, is, PL_FA_TRANSPARENT)	/* -Value is +Expr */
+PRED_IMPL("is", 2, is, PL_FA_ISO)	/* -Value is +Expr */
 { PRED_LD
   AR_CTX
   number arg;
@@ -3047,132 +3131,31 @@ PRED_IMPL("is", 2, is, PL_FA_TRANSPARENT)	/* -Value is +Expr */
 }
 
 
-#if O_PROLOG_FUNCTIONS
-/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-Functions are module-sensitive. We has the arithmetic functions on their
-functor and keep them sorted in   the chain, most-specific module first.
-See also isCurrentArithFunction()
-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+/** current_arithmetic_function(?Term) is nondet.
+
+True if Term is evaluable.
+*/
 
 static
-PRED_IMPL("$arithmetic_function", 2, arithmetic_function, PL_FA_TRANSPARENT)
-{ GET_LD
-  Procedure proc;
-  functor_t fd;
-  FunctorDef fdef;
-  ArithFunction f, a, *ap;
-  Module m = NULL;
-  term_t head = PL_new_term_ref();
-  size_t v;
-  int index, rc;
-
-  PL_strip_module(A1, &m, head);
-  if ( !PL_get_functor(head, &fd) )
-    return PL_error(NULL, 0, NULL,
-		    ERR_TYPE, ATOM_callable, head);
-  fdef = valueFunctor(fd);
-  if ( fdef->arity < 1 )
-    return PL_error(NULL, 0, NULL,
-		    ERR_DOMAIN, ATOM_not_less_than_one, head);
-  if ( !PL_get_integer_ex(A2, &index) )
-    fail;
-
-  proc = lookupProcedure(fd, m);
-  fd = lookupFunctorDef(fdef->name, fdef->arity - 1);
-  if ( (f = isCurrentArithFunction(fd, m)) && f->module == m )
-    succeed;				/* already registered */
-
-  v = functorHashValue(fd, ARITHHASHSIZE);
-  f = allocHeap(sizeof(struct arithFunction));
-  f->functor  = fd;
-  f->function = NULL;
-  f->module   = m;
-  f->level    = m->level;
-  f->proc     = proc;
-
-  startCritical;
-  for(ap = &arithFunctionTable[v], a=*ap; ; ap = &a->next, a = *ap)
-  { if ( !a || isTableRef(a) || f->level >= a->level )
-    { f->next = a;
-      *ap = f;
-      break;
-    }
-  }
-  rc = registerFunction(f, index);
-  if ( !endCritical ) rc = FALSE;
-
-  return rc;
-}
-
-word
-pl_current_arithmetic_function(term_t f, control_t h)
-{ GET_LD
-  ArithFunction a;
-  Module m = NULL;
-  term_t head = PL_new_term_ref();
-
-  switch( ForeignControl(h) )
-  { case FRG_FIRST_CALL:
-    { functor_t fd;
-
-      PL_strip_module(f, &m, head);
-
-      if ( PL_is_variable(head) )
-      { a = arithFunctionTable[0];
-        break;
-      } else if ( PL_get_functor(head, &fd) )
-      {	return isCurrentArithFunction(fd, m) ? TRUE : FALSE;
-      } else
-        return PL_error(NULL, 0, NULL,
-			ERR_TYPE, ATOM_callable, f);
-    }
-    case FRG_REDO:
-      PL_strip_module(f, &m, head);
-
-      a = ForeignContextPtr(h);
-      break;
-    case FRG_CUTTED:
-    default:
-      succeed;
-  }
-
-  for( ; a; a = a->next )
-  { Module m2;
-
-    while( isTableRef(a) )
-    { a = unTableRef(ArithFunction, a);
-      if ( !a )
-        fail;
-    }
-
-    for(m2 = m; m2; )
-    { if ( m2 == a->module && a == isCurrentArithFunction(a->functor, m) )
-      { if ( PL_unify_functor(f, a->functor) )
-	  return_next_table(ArithFunction, a, ;);
-      }
-      if ( m2->supers )			/* TBD: multiple supers! */
-      { m2 = m2->supers->value;
-      } else
-	fail;
-    }
-  }
-
-  fail;
-}
-
-
-static
-PRED_IMPL("$prolog_arithmetic_function", 2, prolog_arithmetic_function,
-	  PL_FA_NONDETERMINISTIC|PL_FA_TRANSPARENT)
+PRED_IMPL("current_arithmetic_function", 1, current_arithmetic_function,
+	  PL_FA_NONDETERMINISTIC)
 { PRED_LD
-  int i, mx;
-  term_t tmp;
-  fid_t fid;
+  unsigned int i;
+  term_t head = A1;
 
   switch( CTX_CNTRL )
   { case FRG_FIRST_CALL:
-      i = 0;
-      break;
+    { functor_t fd;
+
+      if ( PL_is_variable(head) )
+      { i = 0;
+        break;
+      } else if ( PL_get_functor(head, &fd) )
+      {	return isCurrentArithFunction(fd) ? TRUE : FALSE;
+      } else
+        return PL_error(NULL, 0, NULL,
+			ERR_TYPE, ATOM_callable, head);
+    }
     case FRG_REDO:
       i = (int)CTX_INT;
       break;
@@ -3181,37 +3164,18 @@ PRED_IMPL("$prolog_arithmetic_function", 2, prolog_arithmetic_function,
       succeed;
   }
 
-  tmp = PL_new_term_ref();
-  mx = (int)entriesBuffer(function_array, ArithFunction);
+  for(; i<GD->arith.functions_allocated; i++)
+  { if ( GD->arith.functions[i] )
+    { functor_t f = functorArithFunction(i);
 
-  if ( !(fid = PL_open_foreign_frame()) )
-    return FALSE;
-
-  for( ; i<mx; i++ )
-  { ArithFunction f = FunctionFromIndex(i);
-
-    if ( PL_put_functor(tmp, f->functor) &&
-	 f->proc &&
-	 PL_unify_term(A1,
-		       PL_FUNCTOR, FUNCTOR_colon2,
-		         PL_ATOM, f->module->name,
-		         PL_TERM, tmp) &&
-	 PL_unify_integer(A2, (intptr_t)f->index) )
-    { if ( ++i == mx )
-	succeed;
-      ForeignRedoInt(i);
+      if ( PL_unify_functor(head, f) )
+	ForeignRedoInt(i+1);
     }
-
-    if ( exception_term )
-      return FALSE;
-
-    PL_rewind_foreign_frame(fid);
   }
 
   fail;
 }
 
-#endif /* O_PROLOG_FUNCTIONS */
 
 typedef struct
 { functor_t	functor;
@@ -3251,6 +3215,9 @@ static const ar_funcdef ar_funcdefs[] = {
   ADD(FUNCTOR_backslash1,	ar_negation),
 
   ADD(FUNCTOR_random1,		ar_random),
+#ifdef O_GMP
+  ADD(FUNCTOR_random_float0,	ar_random_float),
+#endif
 
   ADD(FUNCTOR_integer1,		ar_integer),
   ADD(FUNCTOR_round1,		ar_integer),
@@ -3291,67 +3258,54 @@ static const ar_funcdef ar_funcdefs[] = {
 
 #undef ADD
 
-static int
-registerFunction(ArithFunction f, int index)
-{ int i = (int)entriesBuffer(function_array, ArithFunction);
+static size_t
+registerFunction(functor_t f, ArithF func)
+{ size_t index = indexFunctor(f);
 
-  if ( index )
-  { if ( index != i )
-    { fatalError("Mismatch in arithmetic function index (%d != %d)",
-		 index, i);
-      fail;				/* NOTREACHED */
+  DEBUG(1, Sdprintf("Register functor %ld\n", (long)index));
+
+  while ( index >= GD->arith.functions_allocated )
+  { if ( GD->arith.functions_allocated == 0 )
+    { size_t size = 256;
+
+      GD->arith.functions = allocHeapOrHalt(size*sizeof(ArithF));
+      memset(GD->arith.functions, 0, size*sizeof(ArithF));
+      GD->arith.functions_allocated = size;
+    } else
+    { size_t size = GD->arith.functions_allocated*2;
+      ArithF *new = allocHeapOrHalt(size*sizeof(ArithF));
+      size_t half = GD->arith.functions_allocated*sizeof(ArithF);
+      ArithF *old = GD->arith.functions;
+
+      DEBUG(0, Sdprintf("Re-sized function-table to %ld\n", (long)size));
+
+      memcpy(new, old, half);
+      memset(addPointer(new,half), 0, half);
+      GD->arith.functions = new;
+      GD->arith.functions_allocated = size;
+      freeHeap(old, half);
     }
   }
 
-  f->index = i;
-  addBuffer(function_array, f, ArithFunction);
+  GD->arith.functions[index] = func;
 
-  succeed;
+  return index;
 }
 
 
 static void
 registerBuiltinFunctions()
-{ GET_LD
-  int n, size = sizeof(ar_funcdefs)/sizeof(ar_funcdef);
-  ArithFunction f = allocHeap(size * sizeof(struct arithFunction));
+{ int n, size = sizeof(ar_funcdefs)/sizeof(ar_funcdef);
   const ar_funcdef *d;
 
-					/* grow to desired size immediately */
-  growBuffer(function_array, size * sizeof(ArithFunction));
-  memset(f, 0, size * sizeof(struct arithFunction));
-
-  for(d = ar_funcdefs, n=0; n<size; n++, f++, d++)
-  { int v = (int)functorHashValue(d->functor, ARITHHASHSIZE);
-
-    f->functor  = d->functor;
-    f->function = d->function;
-    f->module   = MODULE_system;
-    f->level    = 0;			/* level of system module */
-    f->next     = arithFunctionTable[v];
-    arithFunctionTable[v] = f;
-    registerFunction(f, 0);
-    DEBUG(1, Sdprintf("Registered %s/%d at %d, index=%d\n",
-		      stringAtom(nameFunctor(f->functor)),
-		      arityFunctor(f->functor),
-		      v,
-		      f->index));
-  }
+  for(d = ar_funcdefs, n=0; n<size; n++, d++)
+    registerFunction(d->functor, d->function);
 }
 
 
 void
 initArith(void)
-{ initBuffer(function_array);
-					/* link the table to enumerate */
-  { ArithFunction *f;
-    int n;
-
-    for(n=0, f = arithFunctionTable; n < (ARITHHASHSIZE-1); n++, f++)
-      *f = makeTableRef(f+1);
-  }
-
-  registerBuiltinFunctions();
+{ registerBuiltinFunctions();
 
 #ifdef O_INHIBIT_FP_SIGNALS
   fpsetmask(fpgetmask() & ~(FP_X_DZ|FP_X_INV|FP_X_OFL));
@@ -3361,12 +3315,17 @@ initArith(void)
 
 void
 cleanupArith(void)
-{ discardBuffer(function_array);
-
+{
 #ifdef O_INHIBIT_FP_SIGNALS
   fpresetsticky(FP_X_DZ|FP_X_INV|FP_X_OFL);
   fpsetmask(FP_X_DZ|FP_X_INV|FP_X_OFL);
 #endif
+
+  if ( GD->arith.functions )
+  { freeHeap(GD->arith.functions, GD->arith.functions_allocated*sizeof(ArithF));
+    GD->arith.functions = 0;
+    GD->arith.functions_allocated = 0;
+  }
 }
 
 
@@ -3377,19 +3336,29 @@ cleanupArith(void)
 		*********************************/
 
 int
-indexArithFunction(functor_t fdef, Module m)
-{ ArithFunction f;
+indexArithFunction(functor_t f)
+{ size_t index = indexFunctor(f);
 
-  if ( !(f = isCurrentArithFunction(fdef, m)) )
-    return -1;
+  if ( index < GD->arith.functions_allocated )
+  { if ( GD->arith.functions[index] )
+      return (int)index;
+  }
 
-  return (int)f->index;
+  return -1;
 }
 
 
 functor_t
-functorArithFunction(int n)
-{ return FunctionFromIndex(n)->functor;
+functorArithFunction(unsigned int i)
+{ FunctorDef fd = fetchFunctorArray(i);
+
+  return fd->functor;
+}
+
+
+static ArithF
+FunctionFromIndex(int index)
+{ return GD->arith.functions[index];
 }
 
 
@@ -3403,42 +3372,28 @@ bool
 ar_func_n(int findex, int argc ARG_LD)
 { number result;
   int rval;
-  ArithFunction f = FunctionFromIndex(findex);
+  ArithF f = FunctionFromIndex(findex);
   Number argv = argvArithStack(argc PASS_LD);
 
-  if ( f->proc )			/* Prolog implementation */
-  { intptr_t lsafe = (char*)lTop - (char*)lBase;
+  DEBUG(0, if ( !f )
+	     fatalError("No function at index %d", findex));
 
-    lTop = (LocalFrame)argFrameP(lTop, 1); /* needed for is/2.  See A_IS */
-    { fid_t fid = PL_open_foreign_frame();
-      term_t h0 = PL_new_term_refs(argc+1);
-      int n;
-
-      for(n=0; n<argc; n++)
-	_PL_put_number(h0+n, argv+n);
-
-      rval = prologFunction(f, h0, &result PASS_LD);
-      PL_close_foreign_frame(fid);
-    }
-    lTop = addPointer(lBase, lsafe);
-  } else
-  { switch(argc)
-    { case 0:
-	rval = (*f->function)(&result);
-        break;
-      case 1:
-	rval = (*f->function)(argv, &result);
-        break;
-      case 2:
-	rval = (*f->function)(argv, argv+1, &result);
-	break;
-      case 3:
-	rval = (*f->function)(argv, argv+1, argv+2, &result);
-        break;
-      default:
-	rval = FALSE;
-        sysError("Too many arguments to arithmetic function");
-    }
+  switch(argc)
+  { case 0:
+      rval = (*f)(&result);
+      break;
+    case 1:
+      rval = (*f)(argv, &result);
+      break;
+    case 2:
+      rval = (*f)(argv, argv+1, &result);
+      break;
+    case 3:
+      rval = (*f)(argv, argv+1, argv+2, &result);
+      break;
+    default:
+      rval = FALSE;
+      sysError("Too many arguments to arithmetic function");
   }
 
   popArgvArithStack(argc PASS_LD);
@@ -3500,18 +3455,20 @@ PL_eval_expression_to_int64_ex(term_t t, int64_t *val)
 		 *******************************/
 
 BeginPredDefs(arith)
-  PRED_DEF("is",   2, is,  PL_FA_TRANSPARENT|PL_FA_ISO)
-  PRED_DEF("<",	   2, lt,  PL_FA_TRANSPARENT|PL_FA_ISO)
-  PRED_DEF(">",	   2, gt,  PL_FA_TRANSPARENT|PL_FA_ISO)
-  PRED_DEF("=<",   2, leq, PL_FA_TRANSPARENT|PL_FA_ISO)
-  PRED_DEF(">=",   2, geq, PL_FA_TRANSPARENT|PL_FA_ISO)
-  PRED_DEF("=\\=", 2, neq, PL_FA_TRANSPARENT|PL_FA_ISO)
-  PRED_DEF("=:=",  2, eq,  PL_FA_TRANSPARENT|PL_FA_ISO)
-  PRED_DEF("$prolog_arithmetic_function", 2, prolog_arithmetic_function,
-	   PL_FA_NONDETERMINISTIC|PL_FA_TRANSPARENT)
-  PRED_DEF("$arithmetic_function", 2, arithmetic_function, PL_FA_TRANSPARENT)
+  PRED_DEF("is",   2, is,  PL_FA_ISO)
+  PRED_DEF("<",	   2, lt,  PL_FA_ISO)
+  PRED_DEF(">",	   2, gt,  PL_FA_ISO)
+  PRED_DEF("=<",   2, leq, PL_FA_ISO)
+  PRED_DEF(">=",   2, geq, PL_FA_ISO)
+  PRED_DEF("=\\=", 2, neq, PL_FA_ISO)
+  PRED_DEF("=:=",  2, eq,  PL_FA_ISO)
+  PRED_DEF("current_arithmetic_function", 1, current_arithmetic_function,
+	   PL_FA_NONDETERMINISTIC)
   PRED_DEF("succ", 2, succ, 0)
   PRED_DEF("plus", 3, plus, 0)
   PRED_DEF("between", 3, between, PL_FA_NONDETERMINISTIC)
   PRED_DEF("set_random", 1, set_random, 0)
+#ifdef O_GMP
+  PRED_DEF("random_property", 1, random_property, 0)
+#endif
 EndPredDefs
