@@ -19,10 +19,9 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
-/*#define O_SECURE 1*/
 /*#define O_DEBUG 1*/
 #include "pl-incl.h"
 
@@ -64,6 +63,9 @@ PRED_IMPL("$new_findall_bag", 1, new_findall_bag, 0)
 { PRED_LD
   findall_bag *bag = allocHeap(sizeof(*bag));
 
+  if ( !bag )
+    return PL_no_memory();
+
   memset(bag, 0, sizeof(*bag));
   bag->magic = FINDALL_MAGIC;
   bag->answers.unit_size = sizeof(Record);
@@ -86,7 +88,10 @@ PRED_IMPL("$add_findall_bag", 2, add_findall_bag, 0)
     return FALSE;
 
   r = compileTermToHeap(A2, R_NOLOCK);
-  pushSegStack(&bag->answers, &r);
+  if ( !pushRecordSegStack(&bag->answers, r) )
+  { freeRecord(r);
+    return PL_no_memory();
+  }
   bag->gsize += r->gsize;
   bag->solutions++;
 
@@ -101,50 +106,43 @@ PRED_IMPL("$add_findall_bag", 2, add_findall_bag, 0)
 
 
 static inline void
-freeBag(findall_bag *bag ARG_LD)
+freeBag(findall_bag *bag)
 { bag->magic = 0;
   clearSegStack(&bag->answers);
   freeHeap(bag, sizeof(*bag));
 }
 
 
-/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-Note that we much lock  L_AGC,  as   otherwise  AGC  may  get in between
-popSegStack and copyRecordToGlobal(). Alternatively  we   could  make an
-interface that gets the top of the stack, processed it and then pops it.
-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
-
 static
 PRED_IMPL("$collect_findall_bag", 3, collect_findall_bag, 0)
 { PRED_LD
   findall_bag *bag;
-  Record r;
+  Record *rp;
   term_t list = PL_copy_term_ref(A3);
   term_t answer = PL_new_term_ref();
+  size_t space;
+  int rc;
 
   if ( !get_bag(A1, &bag PASS_LD) )
     return FALSE;
+  space = bag->gsize + bag->solutions*3;
 
-  if ( bag->gsize + bag->solutions*3 > spaceStack(global)/sizeof(word) )
-  { garbageCollect(NULL, NULL);
-
-    if ( bag->gsize + bag->solutions*3 > spaceStack(global)/sizeof(word) )
-      return outOfStack(&LD->stacks.global, STACK_OVERFLOW_RAISE);
+  if ( !hasGlobalSpace(space) )
+  { if ( (rc=ensureGlobalSpace(space, ALLOW_GC)) != TRUE )
+      return raiseStackOverflow(rc);
   }
 
-  PL_LOCK(L_AGC);
-  while(popSegStack(&bag->answers, &r))
-  { copyRecordToGlobal(answer, r PASS_LD);
+  while ( (rp=topOfSegStack(&bag->answers)) )
+  { Record r = *rp;
+    copyRecordToGlobal(answer, r, ALLOW_GC PASS_LD);
     PL_cons_list(list, answer, list);
+    PL_LOCK(L_AGC);			/* needed for markAtomsAnswers() */
+    popTopOfSegStack(&bag->answers);
+    PL_UNLOCK(L_AGC);
 
     freeRecord(r);
   }
-
-  assert(LD->bags.bags == bag);
-  LD->bags.bags = bag->parent;
-  PL_UNLOCK(L_AGC);
-
-  freeBag(bag PASS_LD);
+  DEBUG(CHK_SECURE, assert(emptySegStack(&bag->answers)));
 
   return PL_unify(A2, list);
 }
@@ -156,59 +154,32 @@ PRED_IMPL("$destroy_findall_bag", 1, destroy_findall_bag, 0)
   findall_bag *bag;
 
   if ( PL_get_pointer(A1, (void**)&bag) && bag->magic == FINDALL_MAGIC )
-  { Record r;
+  { Record *rp;
+
+    while ( (rp=topOfSegStack(&bag->answers)) )
+    { Record r = *rp;
+
+      PL_LOCK(L_AGC);
+      popTopOfSegStack(&bag->answers);
+      PL_UNLOCK(L_AGC);
 
-    while(popSegStack(&bag->answers, &r))
       freeRecord(r);
+    }
 
-    PL_LOCK(L_AGC);
     assert(LD->bags.bags == bag);
+    PL_LOCK(L_AGC);
     LD->bags.bags = bag->parent;
     PL_UNLOCK(L_AGC);
 
-    freeBag(bag PASS_LD);
+    freeBag(bag);
+    succeed;
   }
 
-  succeed;
+  assert(0);
+  fail;
 }
 
 
-/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-bind_bagof_keys(+Vars, +List:Key-Terms) is det
-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
-
-static
-PRED_IMPL("$bind_bagof_keys", 2, bind_bagof_keys, 0)
-{ PRED_LD
-  term_t list = PL_copy_term_ref(A2);
-  term_t head = PL_new_term_ref();
-  term_t a = PL_new_term_ref();
-  term_t key = PL_new_term_ref();
-  term_t vars;
-  int i, arity;
-
-  if ( !PL_get_name_arity(A1, NULL, &arity) )
-    fail;
-  vars = PL_new_term_refs(arity);
-  for(i=1; i<=arity; i++)
-  { _PL_get_arg(i, A1, vars+i-1);
-  }
-
-  while(PL_get_list(list, head, list))
-  { if ( !PL_get_arg(1, head, key) )
-      fail;
-
-    for(i=1; i<=arity; i++)
-    { _PL_get_arg(i, key, a);
-      if ( PL_is_variable(a) )
-      { PL_unify(a, vars+i-1);
-      }
-    }
-  }
-
-  succeed;
-}
-
 		 /*******************************
 		 *	  ATOM-GC SUPPORT	*
 		 *******************************/
@@ -239,5 +210,4 @@ BeginPredDefs(bag)
   PRED_DEF("$add_findall_bag", 2, add_findall_bag, 0)
   PRED_DEF("$collect_findall_bag", 3, collect_findall_bag, 0)
   PRED_DEF("$destroy_findall_bag", 1, destroy_findall_bag, 0)
-  PRED_DEF("$bind_bagof_keys", 2, bind_bagof_keys, 0)
 EndPredDefs
