@@ -3,9 +3,10 @@
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        jan@swi.psy.uva.nl
+    E-mail:        J.Wielemaker@cs.vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2002, University of Amsterdam
+    Copyright (C): 1985-2011, University of Amsterdam
+			      VU University Amsterdam
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -19,60 +20,24 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
-/*#define O_DEBUG 1*/
+#define O_DEBUG 1
 #ifdef __WINDOWS__
 
-/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-   Extension of SWI-Prolog:
-   Primitives to support interprocess communication via DDE, for those
-   platforms that support DDEML.  Currently, this is the Windows family (3.1
-   and above) and Unix platforms with Bristol's Windows support.
-
-   Eventually, this should turn into a full DDE capability.  For the
-   present, I'm just implementing the client side of conversation
-   management, and only providing request transactions, as follows:
-
-   open_dde_conversation(+Service, +Topic, -Handle)
-   Open a conversation with a server supporting the given service name and
-   topic (atoms).  If successful, Handle may be used to send transactions to
-   the server.  If no willing server is found, fails.
-
-   close_dde_conversation(+Handle)
-   Close the conversation associated with Handle.  All opened conversations
-   should be closed when they're no longer needed, although the system
-   will close any that remain open on process termination.
-
-   dde_request(+Handle, +Item, -Value)
-   Request a value from the server.  Item is an atom that identifies the
-   requested data, and Value will be an atom (CF_TEXT data in DDE parlance)
-   representing that data, if the request is successful.  If unsuccessful,
-   Value will be unified with a term of form error(reason), identifying the
-   problem.
-
-   It could be argued that the atoms above should be strings; I may go that
-   way sometime.
-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
-
 #define INCLUDE_DDEML_H
 #include "pl-incl.h"
 
 #if O_DDE
 #include <string.h>
 
-#ifdef __WATCOMC__			/* at least version 9.5 */
-#define WATCOM_DDEACCESS_BUG 1
-#endif
-
 #define FASTBUFSIZE 512			/* use local buffer upto here */
 #define MAX_CONVERSATIONS 32		/* Max. # of conversations */
 #define TIMEOUT_VERY_LONG 0x7fffffff;	/* largest positive int */
 
 static HCONV conv_handle[MAX_CONVERSATIONS];
 static HCONV server_handle[MAX_CONVERSATIONS];
-static DWORD ddeInst;			/* Instance of this process */
 
 static Module	 MODULE_dde;		/* win_dde */
 static functor_t FUNCTOR_dde_connect3;
@@ -82,12 +47,16 @@ static functor_t FUNCTOR_dde_request4;
 static functor_t FUNCTOR_dde_execute3;
 static functor_t FUNCTOR_error1;
 
+#define LOCK()   PL_LOCK(L_DDE)
+#define UNLOCK() PL_UNLOCK(L_DDE)
+
 static const char *
 dde_error_message(int errn)
-{ const char *err;
+{ GET_LD
+  const char *err;
 
   if ( errn <= 0 )
-    errn = DdeGetLastError(ddeInst);
+    errn = DdeGetLastError(LD->os.dde_instance);
 
   switch(errn)
   { case DMLERR_ADVACKTIMEOUT:
@@ -112,7 +81,7 @@ dde_error_message(int errn)
 }
 
 
-static word
+static int
 dde_warning(const char *cmd)
 { const char *err = dde_error_message(-1);
 
@@ -121,7 +90,7 @@ dde_warning(const char *cmd)
 
 
 static int
-unify_hsz(term_t term, HSZ hsz)
+unify_hsz(DWORD ddeInst, term_t term, HSZ hsz)
 { wchar_t buf[FASTBUFSIZE];
   int len;
 
@@ -137,7 +106,7 @@ unify_hsz(term_t term, HSZ hsz)
       int rc;
 
       if ( !(b2 = malloc((len+1)*sizeof(wchar_t))) )
-	return PL_error(NULL, 0, NULL, ERR_RESOURCE, ATOM_memory);
+	return PL_no_memory();
 
       DdeQueryStringW(ddeInst, hsz, b2, len+1, CP_WINUNICODE);
       rc = PL_unify_wchars(term, PL_ATOM, len, b2);
@@ -161,7 +130,7 @@ unify_hdata(term_t t, HDDEDATA data)
   if ( !(len=DdeGetData(data, buf, sizeof(buf), 0)) )
     return dde_warning("data handle");
 
-  DEBUG(0, Sdprintf("DdeGetData() returned %ld bytes\n", (long)len));
+  DEBUG(1, Sdprintf("DdeGetData() returned %ld bytes\n", (long)len));
 
   if ( len == sizeof(buf) )
   { if ( (len=DdeGetData(data, NULL, 0, 0)) > 0 )
@@ -169,7 +138,7 @@ unify_hdata(term_t t, HDDEDATA data)
       int rval;
 
       if ( !(b2 = malloc(len)) )
-	return PL_error(NULL, 0, NULL, ERR_RESOURCE, ATOM_memory);
+	return PL_no_memory();
 
       DdeGetData(data, b2, len, 0);
       rval = PL_unify_wchars(t, PL_ATOM, len/sizeof(wchar_t)-1, (wchar_t*)b2);
@@ -186,20 +155,23 @@ unify_hdata(term_t t, HDDEDATA data)
 
 
 static int
-get_hsz(term_t data, HSZ *rval)
+get_hsz(DWORD ddeInst, term_t data, HSZ *rval)
 { wchar_t *s;
   size_t len;
 
   if ( PL_get_wchars(data, &len, &s, CVT_ALL|CVT_EXCEPTION) )
-  { HSZ h = DdeCreateStringHandleW(ddeInst, s, CP_WINUNICODE);
+  { HSZ h;
 
-    if ( s[len] )
-      Sdprintf("OOPS, s[%d] != 0\n", len);
+    assert(s[len] == 0);			/* Must be 0-terminated */
 
-    if ( h )
-    { *rval = h;
+    DEBUG(2, Sdprintf("Get HSZ for %Ws ...\n", s));
+    if ( (h=DdeCreateStringHandleW(ddeInst, s, CP_WINUNICODE)) )
+    { DEBUG(2, Sdprintf("\tHSZ = %d\n", (int)h));
+      *rval = h;
       succeed;
     }
+
+    return PL_error(NULL, 0, WinError(), ERR_SYSCALL, "DdeCreateStringHandleW");
   }
 
   fail;
@@ -210,15 +182,19 @@ static int
 allocServerHandle(HCONV handle)
 { int i;
 
+  LOCK();
   for(i=0; i<MAX_CONVERSATIONS; i++)
   { if ( !server_handle[i] )
     { server_handle[i] = handle;
-      return i;
+      break;
     }
   }
+  UNLOCK();
+
+  if ( i<MAX_CONVERSATIONS )
+    return i;
 
   PL_error(NULL, 0, NULL, ERR_RESOURCE, ATOM_max_dde_handles);
-
   return -1;
 }
 
@@ -227,7 +203,7 @@ static int
 findServerHandle(HCONV handle)
 { int i;
 
-  for(i=0; i<MAX_CONVERSATIONS; i++)
+  for(i=0; i<MAX_CONVERSATIONS; i++)		/* can be unlocked */
   { if ( server_handle[i] == handle )
       return i;
   }
@@ -240,6 +216,8 @@ static HDDEDATA CALLBACK
 DdeCallback(UINT type, UINT fmt, HCONV hconv, HSZ hsz1, HSZ hsz2,
             HDDEDATA hData, DWORD dwData1, DWORD dwData2)
 { GET_LD
+  DWORD ddeInst = LD->os.dde_instance;
+
   switch(type)
   {  case XTYP_CONNECT:
      { fid_t cid = PL_open_foreign_frame();
@@ -247,8 +225,8 @@ DdeCallback(UINT type, UINT fmt, HCONV hconv, HSZ hsz1, HSZ hsz2,
        predicate_t pred = PL_pred(FUNCTOR_dde_connect3, MODULE_dde);
        int rval;
 
-       if ( unify_hsz(argv+0, hsz2) &&			/* topic */
-	    unify_hsz(argv+1, hsz1) &&			/* service */
+       if ( unify_hsz(ddeInst, argv+0, hsz2) &&		/* topic */
+	    unify_hsz(ddeInst, argv+1, hsz1) &&		/* service */
 	    PL_unify_integer(argv+2, dwData2 ? 1 : 0) )	/* same instance */
        { rval = PL_call_predicate(MODULE_dde, TRUE, pred, argv);
        } else
@@ -266,8 +244,8 @@ DdeCallback(UINT type, UINT fmt, HCONV hconv, HSZ hsz1, HSZ hsz2,
 	 term_t argv = PL_new_term_refs(3);
 	 predicate_t pred = PL_pred(FUNCTOR_dde_connect_confirm3, MODULE_dde);
 
-	 if ( unify_hsz(argv+0, hsz2) &&			/* topic */
-	      unify_hsz(argv+1, hsz1) &&			/* service */
+	 if ( unify_hsz(ddeInst, argv+0, hsz2) &&		/* topic */
+	      unify_hsz(ddeInst, argv+1, hsz1) &&		/* service */
 	      PL_unify_integer(argv+2, plhandle) )
 	   PL_call_predicate(MODULE_dde, TRUE, pred, argv);
 
@@ -298,10 +276,10 @@ DdeCallback(UINT type, UINT fmt, HCONV hconv, HSZ hsz1, HSZ hsz2,
        term_t argv = PL_new_term_refs(3);
        predicate_t pred = PL_pred(FUNCTOR_dde_execute3, MODULE_dde);
 
-       DEBUG(0, Sdprintf("Got XTYP_EXECUTE request\n"));
+       DEBUG(1, Sdprintf("Got XTYP_EXECUTE request\n"));
 
        PL_put_integer(argv+0, plhandle);
-       unify_hsz(     argv+1, hsz1);
+       unify_hsz(ddeInst, argv+1, hsz1);
        unify_hdata(   argv+2, hData);
        if ( PL_call_predicate(MODULE_dde, TRUE, pred, argv) )
 	 rval = (void *) DDE_FACK;
@@ -319,8 +297,8 @@ DdeCallback(UINT type, UINT fmt, HCONV hconv, HSZ hsz1, HSZ hsz2,
 	 int plhandle = findServerHandle(hconv);
 
 	 PL_put_integer( argv+0, plhandle);
-	 unify_hsz(	 argv+1, hsz1);	/* topic */
-	 unify_hsz(      argv+2, hsz2);	/* item */
+	 unify_hsz(ddeInst, argv+1, hsz1);	/* topic */
+	 unify_hsz(ddeInst, argv+2, hsz2);	/* item */
 
 	 if ( PL_call_predicate(MODULE_dde, TRUE, pred, argv) )
 	 { wchar_t *s;
@@ -346,19 +324,12 @@ DdeCallback(UINT type, UINT fmt, HCONV hconv, HSZ hsz1, HSZ hsz2,
 }
 
 
-static word
-dde_initialise()
-{ if ( ddeInst == (DWORD)NULL )
-  { if (DdeInitializeW(&ddeInst, (PFNCALLBACK)DdeCallback,
-		       APPCLASS_STANDARD|CBF_FAIL_ADVISES|CBF_FAIL_POKES|
-		       CBF_SKIP_REGISTRATIONS|CBF_SKIP_UNREGISTRATIONS,
-		       0L)
-	!= DMLERR_NO_ERROR)
-    { ddeInst = (DWORD) -1;
-      return dde_warning("initialise");
-    }
+static void
+dde_init_constants(void)
+{ static int done = FALSE;
 
-    MODULE_dde = lookupModule(PL_new_atom("win_dde"));
+  if ( !done )				/* no worries if this happens twice */
+  { MODULE_dde = lookupModule(PL_new_atom("win_dde"));
 
     FUNCTOR_dde_connect3  =
 	lookupFunctorDef(PL_new_atom("$dde_connect"), 3);
@@ -372,20 +343,65 @@ dde_initialise()
 	lookupFunctorDef(PL_new_atom("$dde_execute"), 3);
     FUNCTOR_error1        =
         lookupFunctorDef(ATOM_error, 1);
-  }
 
-  succeed;
+    done = TRUE;
+  }
+}
+
+static void
+dde_uninitialise(void *closure)
+{ GET_LD
+  DWORD ddeInst;
+
+  if ( (ddeInst=LD->os.dde_instance) )
+  { LD->os.dde_instance = 0;
+
+    DdeUninitialize(ddeInst);
+  }
 }
 
 
-word
-pl_dde_register_service(term_t topic, term_t onoff)
+static DWORD
+dde_initialise(void)
+{ GET_LD
+  DWORD ddeInst;
+
+  dde_init_constants();
+
+  if ( !(ddeInst=LD->os.dde_instance) )
+  { if ( DdeInitializeW(&ddeInst, (PFNCALLBACK)DdeCallback,
+			APPCLASS_STANDARD|CBF_FAIL_ADVISES|CBF_FAIL_POKES|
+			CBF_SKIP_REGISTRATIONS|CBF_SKIP_UNREGISTRATIONS,
+			0L) == DMLERR_NO_ERROR)
+    { LD->os.dde_instance = ddeInst;
+#ifdef O_PLMT
+      PL_thread_at_exit(dde_uninitialise, NULL, FALSE);
+#endif
+    } else
+    { dde_warning("initialise");
+    }
+
+    DEBUG(1, Sdprintf("Thread %d: created ddeInst %d\n",
+		      PL_thread_self(), ddeInst));
+
+  }
+
+  return ddeInst;
+}
+
+
+static
+PRED_IMPL("$dde_register_service", 2, dde_register_service, 0)
 { HSZ t;
   int a;
+  DWORD ddeInst;
 
-  TRY(dde_initialise());
+  term_t topic = A1;
+  term_t onoff = A2;
 
-  if ( !get_hsz(topic, &t) )
+  if ( !(ddeInst=dde_initialise()) )
+    return FALSE;
+  if ( !get_hsz(ddeInst, topic, &t) )
     fail;
   if ( !PL_get_bool(onoff, &a) )
     return PL_error(NULL, 0, NULL, ERR_TYPE, ATOM_bool, onoff);
@@ -403,35 +419,55 @@ pl_dde_register_service(term_t topic, term_t onoff)
   }
 }
 
-
-word
-pl_open_dde_conversation(term_t service, term_t topic, term_t handle)
-{ GET_LD
+static
+PRED_IMPL("open_dde_conversation", 3, open_dde_conversation, 0)
+{ PRED_LD
   UINT i;
-  HSZ Hservice, Htopic;
+  HSZ Hservice = 0, Htopic = 0;
+  int rc = TRUE;
+  DWORD ddeInst;
 
-  if ( !dde_initialise() )
+  term_t service = A1;
+  term_t topic   = A2;
+  term_t handle  = A3;
+
+  if ( !(ddeInst=dde_initialise()) )
     fail;
 
-  if ( !get_hsz(service, &Hservice) ||
-       !get_hsz(topic, &Htopic) )
-    fail;
+  if ( !get_hsz(ddeInst, service, &Hservice) ||
+       !get_hsz(ddeInst, topic, &Htopic) )
+  { rc = FALSE;
+    goto out;
+  }
 
   /* Establish a connection and get a handle for it */
-  for (i=0; i < MAX_CONVERSATIONS; i++)   /* Find an open slot */
+  LOCK();
+  for (i=0; i < MAX_CONVERSATIONS; i++)		/* Find an open slot */
   { if (conv_handle[i] == (HCONV)NULL)
+    { conv_handle[i] = (HCONV)~0;		/* reserve it */
       break;
+    }
   }
+  UNLOCK();
   if (i == MAX_CONVERSATIONS)
-    return PL_error(NULL, 0, NULL, ERR_RESOURCE, ATOM_max_dde_handles);
+  { rc = PL_error(NULL, 0, NULL, ERR_RESOURCE, ATOM_max_dde_handles);
+    goto out;
+  }
 
   if ( !(conv_handle[i] = DdeConnect(ddeInst, Hservice, Htopic, 0)) )
-    fail;
+  { rc = dde_warning("connect");
+    goto out;
+  }
 
-  DdeFreeStringHandle(ddeInst, Hservice);
-  DdeFreeStringHandle(ddeInst, Htopic);
+  rc = PL_unify_integer(handle, i);
 
-  return PL_unify_integer(handle, i);
+out:
+  if ( Hservice )
+    DdeFreeStringHandle(ddeInst, Hservice);
+  if ( Htopic )
+    DdeFreeStringHandle(ddeInst, Htopic);
+
+  return rc;
 }
 
 
@@ -451,10 +487,12 @@ get_conv_handle(term_t handle, int *theh)
 }
 
 
-word
-pl_close_dde_conversation(term_t handle)
+static
+PRED_IMPL("close_dde_conversation", 1, close_dde_conversation, 0)
 { int hdl;
 
+  term_t handle = A1;
+
   if ( !get_conv_handle(handle, &hdl) )
     fail;
 
@@ -470,9 +508,8 @@ NOTE: Windows-XP gives the wrong value for valuelen below. Hence we will
 use nul-terminated strings.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-word
-pl_dde_request(term_t handle, term_t item,
-	       term_t value, term_t timeout)
+static
+PRED_IMPL("dde_request", 4, dde_request, 0)
 { int hdl;
   int rval;
   HSZ Hitem;
@@ -481,9 +518,18 @@ pl_dde_request(term_t handle, term_t item,
   long tmo;
   static UINT fmt[] = {CF_UNICODETEXT, CF_TEXT};
   int fmti;
+  DWORD ddeInst;
+
+  term_t handle  = A1;
+  term_t item    = A2;
+  term_t value   = A3;
+  term_t timeout = A4;
+
+  if ( !(ddeInst=dde_initialise()) )
+    return FALSE;
 
   if ( !get_conv_handle(handle, &hdl) ||
-       !get_hsz(item, &Hitem) ||
+       !get_hsz(ddeInst, item, &Hitem) ||
        !PL_get_long_ex(timeout, &tmo) )
     fail;
 
@@ -502,10 +548,10 @@ pl_dde_request(term_t handle, term_t item,
   { LPBYTE valuedata;
 
     if ( (valuedata = DdeAccessData(Hvalue, &valuelen)) )
-    { DEBUG(0, Sdprintf("valuelen = %ld; format = %d\n", valuelen, fmti));
+    { DEBUG(1, Sdprintf("valuelen = %ld; format = %d\n", valuelen, fmti));
       if ( fmt[fmti] == CF_TEXT )
-      { DEBUG(0, Sdprintf("ANSI text\n"));
-	rval = PL_unify_string_chars(value, valuedata);
+      { DEBUG(1, Sdprintf("ANSI text\n"));
+	rval = PL_unify_string_chars(value, (char*)valuedata);
       } else
 	rval = PL_unify_wchars(value, PL_STRING, -1, (wchar_t*)valuedata);
       DdeUnaccessData(Hvalue);
@@ -514,26 +560,27 @@ pl_dde_request(term_t handle, term_t item,
     { return dde_warning("access_data");
     }
   } else
-  { const char *errmsg = dde_error_message(-1);
-
-    return PL_unify_term(value,
-			 PL_FUNCTOR, FUNCTOR_error1, /* error(Message) */
-			 PL_CHARS,   errmsg);
+  { return dde_warning("request");
   }
 }
 
 
-
-word
-pl_dde_execute(term_t handle, term_t command, term_t timeout)
+static
+PRED_IMPL("dde_execute", 3, dde_execute, 0)
 { int hdl;
   wchar_t *cmdstr;
   size_t cmdlen;
   HDDEDATA Hvalue, data;
   DWORD result;
+  DWORD ddeInst;
   long tmo;
 
-  if ( !get_conv_handle(handle, &hdl) ||
+  term_t handle  = A1;
+  term_t command = A2;
+  term_t timeout = A3;
+
+  if ( !(ddeInst=dde_initialise()) ||
+       !get_conv_handle(handle, &hdl) ||
        !PL_get_wchars(command, &cmdlen, &cmdstr, CVT_ALL|CVT_EXCEPTION) ||
        !PL_get_long_ex(timeout, &tmo) )
     fail;
@@ -557,17 +604,24 @@ pl_dde_execute(term_t handle, term_t command, term_t timeout)
 }
 
 
-word
-pl_dde_poke(term_t handle, term_t item, term_t data, term_t timeout)
+static
+PRED_IMPL("dde_poke", 4, dde_poke, 0)
 { int hdl;
   wchar_t *datastr;
   size_t datalen;
   HDDEDATA Hvalue;
   HSZ Hitem;
   long tmo;
+  DWORD ddeInst;
 
-  if ( !get_conv_handle(handle, &hdl) ||
-       !get_hsz(item, &Hitem) )
+  term_t handle  = A1;
+  term_t item    = A2;
+  term_t data    = A3;
+  term_t timeout = A4;
+
+  if ( !(ddeInst=dde_initialise()) ||
+       !get_conv_handle(handle, &hdl) ||
+       !get_hsz(ddeInst, item, &Hitem) )
     fail;
   if ( !PL_get_wchars(data, &datalen, &datastr, CVT_ALL|CVT_EXCEPTION) )
     fail;
@@ -582,24 +636,24 @@ pl_dde_poke(term_t handle, term_t item, term_t data, term_t timeout)
 				conv_handle[hdl], Hitem, CF_UNICODETEXT,
 				XTYP_POKE, (DWORD)tmo, NULL);
 
-#if 0
-  if ( !Hvalue )
-  { char *txt;
-
-    if ( !PL_get_nchars(data, &datalen, &txt, CVT_ALL|CVT_EXCEPTION|REP_MB) )
-      fail;
-
-    Hvalue = DdeClientTransaction(txt, datalen+1,
-				  conv_handle[hdl], Hitem, CF_TEXT,
-				  XTYP_POKE, (DWORD)tmo, NULL);
-  }
-#endif
-
   if ( !Hvalue )
     return dde_warning("poke");
 
   succeed;
 }
 
+		 /*******************************
+		 *      PUBLISH PREDICATES	*
+		 *******************************/
+
+BeginPredDefs(dde)
+  PRED_DEF("$dde_register_service",  2, dde_register_service,   0)
+  PRED_DEF("open_dde_conversation",  3, open_dde_conversation,  0)
+  PRED_DEF("close_dde_conversation", 1, close_dde_conversation, 0)
+  PRED_DEF("dde_request",	     4, dde_request,		0)
+  PRED_DEF("dde_execute",	     3, dde_execute,		0)
+  PRED_DEF("dde_poke",		     4, dde_poke,		0)
+EndPredDefs
+
 #endif /*O_DDE*/
 #endif /*__WINDOWS__*/
