@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
 /*#define O_DEBUG 1*/
@@ -30,7 +30,7 @@
 
 #ifdef O_PROFILE
 
-#define current    (LD->profile.current) 	/* current call-node */
+#define current    (LD->profile.current)	/* current call-node */
 #define roots      (LD->profile.roots)		/* roots (<spontaneous>) */
 #define nodes      (LD->profile.nodes)		/* # nodes created */
 #define accounting (LD->profile.accounting)	/* We are accounting */
@@ -53,7 +53,7 @@ static PL_prof_type_t *types[MAX_PROF_TYPES] = { &prof_default_type };
 #define PROFNODE_MAGIC 0x7ae38f24
 
 typedef struct call_node
-{ intptr_t 	    magic;		/* PROFNODE_MAGIC */
+{ intptr_t	    magic;		/* PROFNODE_MAGIC */
   struct call_node *parent;
   void *            handle;		/* handle to procedure-id */
   PL_prof_type_t   *type;
@@ -390,7 +390,7 @@ PRED_IMPL("$prof_sibling_of", 2, prof_sibling_of, PL_FA_NONDETERMINISTIC)
 
 static int
 identify_def(term_t t, void *handle)
-{ return unify_definition(t, handle, 0, 0);
+{ return unify_definition(MODULE_user, t, handle, 0, GP_QUALIFY);
 }
 
 
@@ -462,7 +462,7 @@ typedef struct
 
 
 static void
-free_relatives(prof_ref *r ARG_LD)
+free_relatives(prof_ref *r)
 { prof_ref *n;
 
   for( ; r; r=n)
@@ -480,7 +480,7 @@ static void
 add_parent_ref(node_sum *sum,
 	       call_node *self,
 	       void *handle, PL_prof_type_t *type,
-	       int cycle ARG_LD)
+	       int cycle)
 { prof_ref *r;
 
   sum->calls += self->calls;
@@ -497,7 +497,7 @@ add_parent_ref(node_sum *sum,
     }
   }
 
-  r = allocHeap(sizeof(*r));
+  r = allocHeapOrHalt(sizeof(*r));
   r->calls = self->calls;
   r->redos = self->redos;
   r->ticks = self->ticks;
@@ -511,8 +511,7 @@ add_parent_ref(node_sum *sum,
 
 
 static void
-add_recursive_ref(node_sum *sum, call_node *self, uintptr_t count,
-		  int cycle ARG_LD)
+add_recursive_ref(node_sum *sum, uintptr_t count, int cycle)
 { prof_ref *r;
 
   for(r=sum->callers; r; r=r->next)
@@ -523,7 +522,7 @@ add_recursive_ref(node_sum *sum, call_node *self, uintptr_t count,
     }
   }
 
-  r = allocHeap(sizeof(*r));
+  r = allocHeapOrHalt(sizeof(*r));
   memset(r, 0, sizeof(*r));
   r->calls = count;
   r->handle = DEF_RECURSIVE;
@@ -534,8 +533,7 @@ add_recursive_ref(node_sum *sum, call_node *self, uintptr_t count,
 
 
 static void
-add_sibling_ref(node_sum *sum, call_node *self, call_node *sibling,
-		int cycle ARG_LD)
+add_sibling_ref(node_sum *sum, call_node *sibling, int cycle)
 { prof_ref *r;
 
   for(r=sum->callees; r; r=r->next)
@@ -549,7 +547,7 @@ add_sibling_ref(node_sum *sum, call_node *self, call_node *sibling,
     }
   }
 
-  r = allocHeap(sizeof(*r));
+  r = allocHeapOrHalt(sizeof(*r));
   r->calls = sibling->calls;
   r->redos = sibling->redos;
   r->ticks = sibling->ticks;
@@ -578,15 +576,15 @@ sumProfile(call_node *n, void *handle, PL_prof_type_t *type,
     }
 
     if ( n->parent )
-      add_parent_ref(sum, n, n->parent->handle, n->parent->type, seen PASS_LD);
+      add_parent_ref(sum, n, n->parent->handle, n->parent->type, seen);
     else
-      add_parent_ref(sum, n, DEF_SPONTANEOUS, NULL, seen PASS_LD);
+      add_parent_ref(sum, n, DEF_SPONTANEOUS, NULL, seen);
 
     if ( n->recur )
-      add_recursive_ref(sum, n, n->recur, seen PASS_LD);
+      add_recursive_ref(sum, n->recur, seen);
 
     for(s=n->siblings; s; s = s->next)
-      add_sibling_ref(sum, n, s, seen PASS_LD);
+      add_sibling_ref(sum, s, seen);
 
     seen++;
   }
@@ -707,8 +705,8 @@ PRED_IMPL("$prof_procedure_data", 7, prof_procedure_data, PL_FA_TRANSPARENT)
 	 unify_relatives(A7, sum.callees PASS_LD)
        );
 
-  free_relatives(sum.callers PASS_LD);
-  free_relatives(sum.callees PASS_LD);
+  free_relatives(sum.callers);
+  free_relatives(sum.callees);
 
   return rc ? TRUE : FALSE;
 }
@@ -732,7 +730,7 @@ PRED_IMPL("$prof_statistics", 4, prof_statistics, 0)
 		 *******************************/
 
 bool
-resetProfiler()
+resetProfiler(void)
 { GET_LD
   stopProfiler();
 
@@ -801,6 +799,7 @@ profile(intptr_t count, PL_local_data_t *__PL_ld)
 #else /*__WINDOWS__*/
 profile(int sig)
 { GET_LD
+  (void)sig;
 
 #define count 1
 
@@ -855,7 +854,7 @@ prof_call(void *handle, PL_prof_type_t *type ARG_LD)
       }
     }
 
-    node = allocHeap(sizeof(*node));
+    node = allocHeapOrHalt(sizeof(*node));
     memset(node, 0, sizeof(*node));
     nodes++;
 
@@ -906,7 +905,7 @@ prof_call(void *handle, PL_prof_type_t *type ARG_LD)
     }
   }
 
-  node = allocHeap(sizeof(*node));
+  node = allocHeapOrHalt(sizeof(*node));
   memset(node, 0, sizeof(*node));
   nodes++;
   node->magic = PROFNODE_MAGIC;
@@ -944,7 +943,7 @@ code.  Considering this is development only, we'll leave this for now.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 void
-profExit(struct call_node *node ARG_LD)
+profResumeParent(struct call_node *node ARG_LD)
 { call_node *n;
 
   if ( node && node->magic != PROFNODE_MAGIC )
@@ -960,6 +959,15 @@ profExit(struct call_node *node ARG_LD)
 }
 
 
+void
+profExit(struct call_node *node ARG_LD)
+{ if ( !node || node->magic != PROFNODE_MAGIC )
+    return;
+
+  profResumeParent(node->parent PASS_LD);
+}
+
+
 void
 profRedo(struct call_node *node ARG_LD)
 { if ( node && node->magic != PROFNODE_MAGIC )
@@ -1015,7 +1023,7 @@ PL_prof_exit(void *node)
 { GET_LD
   struct call_node *n = node;
 
-  profExit(n->parent PASS_LD);
+  profResumeParent(n->parent PASS_LD);
 }
 
 
@@ -1039,7 +1047,7 @@ collectSiblingsNode(call_node *n)
 
 
 static void
-collectSiblingsTime()
+collectSiblingsTime(void)
 { GET_LD
 
   if ( !sum_ok )
@@ -1055,10 +1063,14 @@ collectSiblingsTime()
 
 static void
 freeProfileNode(call_node *node ARG_LD)
-{ call_node *n;
+{ call_node *n, *next;
 
-  for(n=node->siblings; n; n=n->next)
-  { freeProfileNode(n PASS_LD);
+  assert(node->magic == PROFNODE_MAGIC);
+
+  for(n=node->siblings; n; n=next)
+  { next = n->next;
+
+    freeProfileNode(n PASS_LD);
   }
 
   node->magic = 0;
@@ -1068,17 +1080,19 @@ freeProfileNode(call_node *node ARG_LD)
 
 
 static void
-freeProfileData()
+freeProfileData(void)
 { GET_LD
-  call_node *n;
-
-  for(n=roots; n; n=n->next)
-  { freeProfileNode(n PASS_LD);
-  }
+  call_node *n, *next;
 
+  n = roots;
   roots = NULL;
   current = NULL;
 
+  for(; n; n=next)
+  { next = n->next;
+    freeProfileNode(n PASS_LD);
+  }
+
   assert(nodes == 0);
 }
 
@@ -1089,7 +1103,7 @@ freeProfileData()
 		 *******************************/
 
 void
-stopItimer()
+stopItimer(void)
 {
 }
 
