@@ -20,7 +20,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
 #include <math.h>
@@ -559,6 +559,9 @@ errorWarning(const char *id_str, term_t id_term, ReadData _PL_rd)
 { GET_LD
   term_t ex;
 
+  if ( Sferror(rb.stream) )		/* Stream error; will be reported */
+    fail;				/* elsewhere */
+
   LD->exception.processing = TRUE;	/* allow using spare stack */
 
   ex = makeErrorTerm(id_str, id_term, _PL_rd);
@@ -688,7 +691,7 @@ clearBuffer(ReadData _PL_rd)
 static void
 growToBuffer(int c, ReadData _PL_rd)
 { if ( rb.base == rb.fast )		/* long clause: jump to use malloc() */
-  { rb.base = PL_malloc(FASTBUFFERSIZE * 2);
+  { rb.base = PL_malloc_atomic(FASTBUFFERSIZE * 2);
     memcpy(rb.base, rb.fast, FASTBUFFERSIZE);
   } else
     rb.base = PL_realloc(rb.base, rb.size*2);
@@ -850,8 +853,11 @@ raw_read_quoted(int q, ReadData _PL_rd)
     addToBuffer(c, _PL_rd);
   }
   if (c == EOF)
-  { eofinstr:
-      rawSyntaxError("end_of_file_in_string");
+  {
+  eofinstr:
+    if ( Sferror(rb.stream) )
+      return FALSE;
+    rawSyntaxError("end_of_file_in_string");
   }
   addToBuffer(c, _PL_rd);
 
@@ -920,6 +926,8 @@ raw_read2(ReadData _PL_rd ARG_LD)
   handle_c:
     switch(c)
     { case EOF:
+		if ( Sferror(rb.stream) )
+		  return NULL;
 		if ( isStringStream(rb.stream) ) /* do not require '. ' when */
 		{ addToBuffer(' ', _PL_rd);     /* reading from a string */
 		  addToBuffer('.', _PL_rd);
@@ -979,6 +987,8 @@ raw_read2(ReadData _PL_rd ARG_LD)
 		  { if ( cbuf )
 		      discardBuffer(cbuf);
 		    setErrorLocation(pos, _PL_rd);
+		    if ( Sferror(rb.stream) )
+		      return NULL;
 		    rawSyntaxError("end_of_file_in_block_comment");
 		  }
 		  if ( cbuf )
@@ -1001,6 +1011,8 @@ raw_read2(ReadData _PL_rd ARG_LD)
 			if ( cbuf )
 			  discardBuffer(cbuf);
 		        setErrorLocation(pos, _PL_rd);
+			if ( Sferror(rb.stream) )
+			  return NULL;
 			rawSyntaxError("end_of_file_in_block_comment");
 		      case '*':
 			if ( last == '/' )
@@ -1976,7 +1988,8 @@ str_number(cucharp in, ucharp *end, Number value, int escape)
   }
 					/* floating point numbers */
   if ( *in == '.' && isDigit(in[1]) )
-  { value->type = V_FLOAT;
+  { clearNumber(value);
+    value->type = V_FLOAT;
 
     in++;
     while( isDigit(*in) )
@@ -1985,7 +1998,8 @@ str_number(cucharp in, ucharp *end, Number value, int escape)
 
   if ( (*in == 'e' || *in == 'E') &&
        ((isSign(in[1]) && isDigit(in[2])) || isDigit(in[1])) )
-  { value->type = V_FLOAT;
+  { clearNumber(value);
+    value->type = V_FLOAT;
 
     in++;
     if ( isSign(*in) )
@@ -3415,8 +3429,18 @@ pl_raw_read2(term_t from, term_t term)
 
   init_read_data(&rd, in PASS_LD);
   if ( !(s = raw_read(&rd, &e PASS_LD)) )
-  { rval = PL_raise_exception(rd.exception);
-    goto out;
+  { if ( Sferror(in) )
+    { rval = streamStatus(in);
+    } else
+    { rval = PL_raise_exception(rd.exception);
+      PL_release_stream(in);
+    }
+
+    free_read_data(&rd);
+
+    return rval;
+  } else
+  { PL_release_stream(in);
   }
 
 					/* strip the input from blanks */
@@ -3437,13 +3461,7 @@ pl_raw_read2(term_t from, term_t term)
   txt.canonical = FALSE;
 
   rval = PL_unify_text(term, 0, &txt, PL_ATOM);
-
-out:
   free_read_data(&rd);
-  if ( Sferror(in) )
-    return streamStatus(in);
-  else
-    PL_release_stream(in);
 
   return rval;
 }
