@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
 /*#define O_DEBUG 1*/
@@ -52,27 +52,12 @@ static int allocStacks(size_t local, size_t global, size_t trail);
 static void initSignals(void);
 static void gcPolicy(Stack s, int policy);
 
-#undef I
-#define I TAGEX_INDIRECT
-
-const unsigned int tagtypeex[] =
-{
-	    /* var     attvar  int    float   atom   string    term     ref */
-/* static */	0,	0,	0,	0,	0,	0,	0,	0,
-/* heap */	0,	0,      I,	I,	0,	I,	0,	0,
-/* global */	0,	0,      I,	I,	0,	I,	0,	0,
-/* local */	0,	0,	0,	0,	0,	0,	0,	0
-};
-
-#undef I
-
-void
+int
 setupProlog(void)
 { GET_LD
   DEBUG(1, Sdprintf("Starting Heap Initialisation\n"));
 
   LD->critical = 0;
-  LD->aborted = ABORT_NONE;
   LD->signal.pending = 0;
 
   startCritical;
@@ -129,13 +114,13 @@ setupProlog(void)
   initFiles();
   initIO();
   initCharConversion();
-  DEBUG(1, Sdprintf("Term ...\n"));
-  resetTerm();
   GD->io_initialised = TRUE;
 
-  endCritical;
+  if ( !endCritical )
+    return FALSE;
 
   DEBUG(1, Sdprintf("Heap Initialised\n"));
+  return TRUE;
 }
 
 
@@ -146,7 +131,9 @@ initPrologLocalData(ARG1_LD)
   depth_limit   = (uintptr_t)DEPTH_NO_LIMIT;
 #endif
 
+  LD->break_level = -1;
   LD->prolog_flag.write_attributes = PL_WRT_ATTVAR_IGNORE;
+
   updateAlerted(LD);
 }
 
@@ -578,7 +565,9 @@ unprepareSignal(int sig)
 #ifdef SIGHUP
 static void
 hupHandler(int sig)
-{ PL_halt(2);
+{ (void)sig;
+
+  PL_halt(2);
 }
 #endif
 
@@ -586,6 +575,7 @@ hupHandler(int sig)
 static void
 sig_exception_handler(int sig)
 { GET_LD
+  (void)sig;
 
   if ( LD && LD->signal.exception )
   { record_t ex = LD->signal.exception;
@@ -597,7 +587,7 @@ sig_exception_handler(int sig)
     PL_erase(ex);
     exception_term = exception_bin;
 
-    SECURE(checkData(valTermRef(exception_term)));
+    DEBUG(CHK_SECURE, checkData(valTermRef(exception_term)));
   }
 }
 
@@ -605,6 +595,7 @@ sig_exception_handler(int sig)
 static void
 agc_handler(int sig)
 { GET_LD
+  (void)sig;
 
   if ( GD->statistics.atoms >= GD->atoms.non_garbage + GD->atoms.margin &&
        !gc_status.blocked )
@@ -614,7 +605,9 @@ agc_handler(int sig)
 
 static void
 gc_handler(int sig)
-{ garbageCollect();
+{ (void)sig;
+
+  garbageCollect();
 }
 
 
@@ -622,6 +615,7 @@ static void
 free_clauses_handler(int sig)
 { GET_LD
   ClauseRef cref;
+  (void)sig;
 
   if ( (cref=LD->freed_clauses) )
   { LD->freed_clauses = NULL;
@@ -632,7 +626,9 @@ free_clauses_handler(int sig)
 
 static void
 abort_handler(int sig)
-{ abortProlog(ABORT_RAISE);
+{ (void)sig;
+
+  abortProlog();
 }
 
 
@@ -644,7 +640,7 @@ with EINTR and thus make them interruptable for thread-signals.
 #ifdef SIG_ALERT
 static void
 alert_handler(int sig)
-{
+{ (void)sig;
 }
 #endif
 
@@ -658,7 +654,13 @@ initSignals(void)
 #endif
 
   for( ; sn->name; sn++)
-  { if ( sn->flags )
+  {
+#ifdef HAVE_BOEHM_GC
+    if ( sn->sig == GC_get_suspend_signal() ||
+	 sn->sig == GC_get_thr_restart_signal() )
+      sn->flags = 0;
+#endif
+    if ( sn->flags )
     { SigHandler sh = prepareSignal(sn->sig);
       sh->flags |= sn->flags;
     }
@@ -913,8 +915,8 @@ out:
 
 int
 endCritical__LD(ARG1_LD)
-{ if ( LD->aborted )
-    return abortProlog(ABORT_THROW);
+{ if ( exception_term )
+    return FALSE;
 
   return TRUE;
 }
@@ -1135,7 +1137,6 @@ emptyStack(Stack s)
 void
 emptyStacks(void)
 { GET_LD
-  int i;
 
   environment_frame = NULL;
   fli_context       = NULL;
@@ -1147,23 +1148,27 @@ emptyStacks(void)
   emptyStack((Stack)&LD->stacks.argument);
 
   LD->mark_bar          = gTop;
-  PL_open_foreign_frame();
-  exception_bin         = PL_new_term_ref();
-  exception_printed     = PL_new_term_ref();
-  LD->exception.tmp     = PL_new_term_ref();
-  LD->exception.pending = PL_new_term_ref();
-  LD->trim.dummy        = PL_new_term_ref();
+  if ( lTop && gTop )
+  { int i;
+
+    PL_open_foreign_frame();
+    exception_bin         = PL_new_term_ref();
+    exception_printed     = PL_new_term_ref();
+    LD->exception.tmp     = PL_new_term_ref();
+    LD->exception.pending = PL_new_term_ref();
+    LD->trim.dummy        = PL_new_term_ref();
 #ifdef O_ATTVAR
-  LD->attvar.head	= PL_new_term_ref();
-  LD->attvar.tail       = PL_new_term_ref();
-  DEBUG(3, Sdprintf("attvar.tail at %p\n", valTermRef(LD->attvar.tail)));
+    LD->attvar.head	= PL_new_term_ref();
+    LD->attvar.tail       = PL_new_term_ref();
+    DEBUG(3, Sdprintf("attvar.tail at %p\n", valTermRef(LD->attvar.tail)));
 #endif
 #ifdef O_GVAR
-  destroyGlobalVars();
+    destroyGlobalVars();
 #endif
-  for(i=0; i<TMP_PTR_SIZE; i++)
-    LD->tmp.h[i] = PL_new_term_ref();
-  LD->tmp.top = 0;
+    for(i=0; i<TMP_PTR_SIZE; i++)
+      LD->tmp.h[i] = PL_new_term_ref();
+    LD->tmp.top = 0;
+  }
 }
 
 
@@ -1176,8 +1181,7 @@ Malloc/realloc/free based stack allocation
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 static void
-init_stack(Stack s, char *name,
-	   size_t size, size_t limit, size_t minfree, size_t spare)
+init_stack(Stack s, char *name, size_t size, size_t limit, size_t spare)
 { GET_LD
 
   s->name	= name;
@@ -1222,13 +1226,13 @@ allocStacks(size_t local, size_t global, size_t trail)
   lBase = (LocalFrame) addPointer(gBase, iglobal);
 
   init_stack((Stack)&LD->stacks.global,
-	     "global",   iglobal, global,  minglobal, 512*SIZEOF_VOIDP);
+	     "global",   iglobal, global, 512*SIZEOF_VOIDP);
   init_stack((Stack)&LD->stacks.local,
-	     "local",    ilocal,  local,   minlocal, 512*SIZEOF_VOIDP);
+	     "local",    ilocal,  local,  512*SIZEOF_VOIDP);
   init_stack((Stack)&LD->stacks.trail,
-	     "trail",    itrail,  trail,   mintrail, 256*SIZEOF_VOIDP);
+	     "trail",    itrail,  trail,  256*SIZEOF_VOIDP);
   init_stack((Stack)&LD->stacks.argument,
-	     "argument", argument, argument, minargument, 0);
+	     "argument", argument, argument, 0);
 
   LD->stacks.local.min_free = LOCAL_MARGIN;
 
@@ -1252,6 +1256,9 @@ stack_malloc(size_t size)
   if ( mem )
   { size_t *sp = mem;
     *sp++ = size;
+#ifdef SECURE_GC
+    memset(sp, 0xFB, size);
+#endif
 
     PL_LOCK(L_MISC);
     GD->statistics.stack_space += size;
@@ -1268,6 +1275,13 @@ stack_realloc(void *old, size_t size)
   size_t osize = *--sp;
   void *mem;
 
+#ifdef SECURE_GC
+  if ( (mem = stack_malloc(size)) )
+  { memcpy(mem, old, (size>osize?osize:size));
+    stack_free(old);
+    return mem;
+  }
+#else
   if ( (mem = realloc(sp, size+sizeof(size_t))) )
   { sp = mem;
     *sp++ = size;
@@ -1277,6 +1291,7 @@ stack_realloc(void *old, size_t size)
     PL_UNLOCK(L_MISC);
     return sp;
   }
+#endif
 
   return NULL;
 }
@@ -1290,6 +1305,9 @@ stack_free(void *mem)
   GD->statistics.stack_space -= osize;
   PL_UNLOCK(L_MISC);
 
+#ifdef SECURE_GC
+  memset(sp, 0xFB, osize+sizeof(size_t));
+#endif
   free(sp);
 }
 
@@ -1301,7 +1319,8 @@ trim_stack(Stack s)
     ssize_t room = roomStackP(s);
 
     if ( room < reduce )
-    { DEBUG(0, Sdprintf("Only %d spare for %s-stack\n", room, s->name));
+    { DEBUG(MSG_SPARE_STACK,
+	    Sdprintf("Only %d spare for %s-stack\n", room, s->name));
       reduce = room;
     }
 
@@ -1396,9 +1415,10 @@ trimStacks(int resize ARG_LD)
     }
   }
 
-  SECURE({ scan_global(FALSE);
-	   checkStacks(NULL);
-	 });
+  DEBUG(CHK_SECURE,
+	{ scan_global(FALSE);
+	  checkStacks(NULL);
+	});
 }
 
 
@@ -1448,7 +1468,10 @@ freePrologLocalData(PL_local_data_t *ld)
   discardBuffer(&ld->fli._discardable_buffer);
 
   for(i=0; i<BUFFER_RING_SIZE; i++)
-    discardBuffer(&ld->fli._buffer_ring[i]);
+  { discardBuffer(&ld->fli._buffer_ring[i]);
+    initBuffer(&ld->fli._buffer_ring[i]);	/* Used by debug-print */
+						/* on shutdown */
+  }
 
   freeVarDefs(ld);
 
@@ -1465,7 +1488,15 @@ freePrologLocalData(PL_local_data_t *ld)
   freeArithLocalData(ld);
 #ifdef O_PLMT
   simpleMutexDelete(&ld->signal.sig_lock);
+  if ( ld->prolog_flag.table )
+  { PL_LOCK(L_PLFLAG);
+    destroyHTable(ld->prolog_flag.table);
+    PL_UNLOCK(L_PLFLAG);
+  }
 #endif
+
+  if ( ld->qlf.getstr_buffer )
+    free(ld->qlf.getstr_buffer);
 }
 
 
