@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -32,6 +32,33 @@ The    file    pl-extend.c,    copied    by    the    installation    to
 embedded application.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
+/*
+  Preprocessor symbols that may be defined during the compilation
+  of this file:
+
+  #ifdef __WINDOWS__
+    This file (swipl-ld.c) is being compiled by a Windows compiler
+    (either MSC or gcc/MinGW).  The resulting swipl-ld* executable
+    will thus run under Windows.
+  #endif
+
+  #ifdef HOST_TOOLCHAIN_MSC
+    The generated swipl-ld* executable will use the Microsoft C/C++
+    compiler to produce shared objects and executables.
+  #endif
+
+  #ifdef HOST_OS_WINDOWS
+    The generated swipl-ld* executable will produce shared objects
+    and executable that will run under Windows.
+  #endif
+ */
+
+#ifndef __WINDOWS__
+#if defined(_MSC_VER) || defined(__MINGW32__)
+#define __WINDOWS__ 1
+#endif
+#endif
+
 #define UNQUOTED_PREFIX "\1"
 
 #include "pl-incl.h"
@@ -39,45 +66,70 @@ embedded application.
 #ifdef __WINDOWS__
 #include <process.h>
 #include <io.h>
+
 #if (_MSC_VER < 1400)
 #define off_t intptr_t
 #endif
+
+#ifdef _MSC_VER
+#define HOST_TOOLCHAIN_MSC
+
 #define popen _popen
 #define pclose _pclose
+
 #define O_WRONLY _O_WRONLY
 #define O_RDONLY _O_RDONLY
 #define O_CREAT _O_CREAT
 #define O_TRUNC _O_TRUNC
 #define O_BINARY _O_BINARY
+#endif
+
+#endif /*__WINDOWS__*/
+
+#if defined(__WINDOWS__) || defined(HOST_OS_WINDOWS)
+#define DEF_PROG_PL "swipl.exe"
+#define PROG_OUT "a.exe"
+#else
+#define DEF_PROG_PL "swipl"
+#define PROG_OUT "a.out"
+#endif
 
 #ifndef PROG_PL
-#define PROG_PL "swipl.exe"
+#define PROG_PL DEF_PROG_PL
 #endif
-#ifdef __MINGW32__
-#define PROG_CC C_CC
-#define PROG_CXX C_CC				/* TBD */
-#define PROG_LD C_CC
-#define EXT_OBJ "o"
-#define OPT_DEBUG "-g"
-#else
-#define PROG_LD "link.exe"
-#define SO_LD "link.exe"
+
+
+#if defined(HOST_TOOLCHAIN_MSC)
+
 #define PROG_CC "cl.exe /MD"
 #define PROG_CXX "cl.exe /MD /GX"
+/* PROG_CPP is defined in config.h: we need to redefine it. */
+#undef PROG_CPP
 #define PROG_CPP "cl.exe -P"
-#define LIB_PL_DEBUG "swiplD.lib"
-#define EXT_OBJ "obj"
-#define OPT_DEBUG "/DEBUG"
-#endif /*__MINGW32__*/
-#define PROG_OUT "plout.exe"
-#else /*__WINDOWS__*/
 
-#ifndef PROG_PL
-#define PROG_PL "swipl"
+#define PROG_LD "link.exe"
+/* SO_LD is defined in config.h: we need to redefine it. */
+#undef SO_LD
+#define SO_LD "link.exe"
+
+#define EXT_OBJ "obj"
+
+#define LIB_PL_DEBUG "swiplD.lib"
+#define OPT_DEBUG "/DEBUG"
+
+#else /* !defined(HOST_TOOLCHAIN_MSC) */
+
+#define PROG_CC C_CC
+#define PROG_CXX C_CC " -x c++"
+/* PROG_CPP is defined in config.h */
+
+#define PROG_LD C_CC
+
+/* SO_LD is defined in config.h. */
+#ifndef PROG_LD
+#define PROG_LD C_CC
 #endif
-#define PROG_CC "cc"
-#define PROG_CXX "c++"
-#define PROG_OUT "a.out"
+
 #define EXT_OBJ "o"
 #define OPT_DEBUG "-g"
 
@@ -85,7 +137,8 @@ embedded application.
 #define SO_LDFLAGS "-shared"
 #endif
 
-#endif /*__WINDOWS__*/
+#endif /* !defined(HOST_TOOLCHAIN_MSC) */
+
 
 #include <stdio.h>
 #include <stdlib.h>
@@ -130,7 +183,7 @@ embedded application.
 #define streq(s, q)     (strcmp((s), (q)) == 0)
 #endif
 #define strprefix(s, p) (strncmp((s), (p), strlen(p)) == 0)
-#ifdef __WINDOWS__
+#if defined(__WINDOWS__)
 #define strfeq(s, q)	(stricmp((s), (q)) == 0)
 #else
 #define strfeq(s, q)	streq(s, q)
@@ -180,6 +233,7 @@ static arglist libdirs;			/* -L library directories */
 static arglist includedirs;		/* -I include directories */
 
 static char *pllib;			/* -lswipl, swipl.lib, ... */
+static char *pllibs = "";		/* Requirements to link to pllib */
 
 static char *pl;			/* Prolog executable */
 static char *cc;			/* CC executable */
@@ -201,9 +255,13 @@ static char *ctmp;			/* base executable */
 static char *pltmp;			/* base saved state */
 static char *out;			/* final output */
 static int  opt_o=FALSE;		/* -o out given */
+static int  opt_E=FALSE;		/* -E given */
 
-static int nostate = FALSE;		/* do not make a state */
+static int build_defaults = FALSE;	/* don't ask Prolog for parameters*/
+
+static int nostate = TRUE;		/* do not make a state */
 static int nolink = FALSE;		/* do not link */
+static int nolibswipl = FALSE;		/* do not link with -lswipl */
 static int shared = FALSE;		/* -shared: make a shared-object/DLL */
 static char *soext;			/* extension of shared object */
 static int embed_shared = FALSE;	/* -dll/-embed-shared: embed Prolog */
@@ -518,7 +576,7 @@ typedef struct
 
 static extdef extdefs[] =
 { { EXT_OBJ,	&ofiles },
-#ifdef __WINDOWS__
+#if defined(HOST_TOOLCHAIN_MSC)
   { "lib",	&libs },
 #else
   { "a",	&libs },
@@ -559,7 +617,9 @@ dispatchFile(const char *name)
 
     for( ; d->extension; d++ )
     { if ( strfeq(d->extension, ext) )
-      { appendArgList(d->list, name);
+      { if ( d->list == &plfiles )
+	  nostate = FALSE;
+	appendArgList(d->list, name);
 	return TRUE;
       }
     }
@@ -583,10 +643,9 @@ usage()
 	  "usage: %s -help\n"
 	  "       %s [options] inputfile ...\n"
 	  "       %s -shared -o out inputfile ...\n"
-#ifdef __WINDOWS__
+#if defined(HOST_OS_WINDOWS)
 	  "       %s -dll -o out inputfile ...\n"
 #endif
-	  "       %s -E cppargument ...\n"
 	  "\n"
 	  "options:\n"
 	  "       -o out           define output file\n"
@@ -602,10 +661,15 @@ usage()
 	  "       -c++ compiler    compiler for C++ source files\n"
 	  "\n"
 	  "       -c               only compile C/C++ files, do not link\n"
+	  "       -S               emit assembler, do not link\n"
+	  "       -E               only run preprocessor, do not link\n"
+          "       -build-defaults  use default parameters, don't ask Prolog\n"
 	  "       -nostate         just relink the kernel\n"
+	  "       -state           add a Prolog saved state\n"
+	  "       -nolibswipl      do not link with -lswipl\n"
 	  "       -shared          create target for load_foreign_library/2\n"
 	  "       -embed-shared    embed Prolog in a shared object/DLL\n"
-#ifdef __WINDOWS__
+#if defined(HOST_OS_WINDOWS)
 	  "       -dll             synonym for -embed-shared\n"
 #endif
 	  "       -fpic            compile small position-independent code\n"
@@ -632,8 +696,7 @@ usage()
 	  "       -llib            library (C/C++)\n",
 	plld,
         plld,
-        plld,
-#ifdef __WINDOWS__
+#if defined(HOST_OS_WINDOWS)
         plld,
 #endif
         plld);
@@ -666,10 +729,15 @@ parseOptions(int argc, char **argv)
     { nolink++;
       appendArgList(&coptions, opt);
       appendArgList(&cppoptions, opt);
+    } else if ( streq(opt, "-E") )		/* -E */
+    { nolink++;
+      opt_E = TRUE;
+      appendArgList(&coptions, opt);
+      appendArgList(&cppoptions, opt);
     } else if ( streq(opt, "-g") )		/* -g */
     { appendArgList(&coptions, OPT_DEBUG);
       appendArgList(&cppoptions, OPT_DEBUG);
-#ifdef _MSC_VER					/* MSVC DEBUG OPTIONS */
+#if defined(HOST_TOOLCHAIN_MSC)			/* MSVC DEBUG OPTIONS */
       appendArgList(&coptions, "/ZI");
       appendArgList(&coptions, "/Od");
       appendArgList(&cppoptions, "/ZI");
@@ -696,12 +764,18 @@ parseOptions(int argc, char **argv)
     } else if ( strprefix(opt, "-W") )		/* -W* */
     { appendArgList(&coptions, opt);
       appendArgList(&cppoptions, opt);
+    } else if ( streq(opt, "-build-defaults") )	/* -build-defaults */
+    { build_defaults = TRUE;
     } else if ( streq(opt, "-nostate") )	/* -nostate */
     { nostate = TRUE;
+    } else if ( streq(opt, "-state") )		/* -state */
+    { nostate = FALSE;
+    } else if ( streq(opt, "-nolibswipl") )	/* -nolibswipl */
+    { nolibswipl = TRUE;
     } else if ( streq(opt, "-dll") ||		/* -dll */
 		streq(opt, "-embed-shared") )   /* -embed-shared */
     { embed_shared = TRUE;
-#ifdef _MSC_VER
+#if defined(HOST_TOOLCHAIN_MSC)
       appendArgList(&ldoptions, "/DLL");
 #else
 #ifdef SO_pic
@@ -867,13 +941,6 @@ tmpPath(char **store, const char *base)
 }
 
 
-#ifndef SO_LD
-#define SO_LD cc
-#endif
-#ifndef PROG_LD
-#define PROG_LD cc
-#endif
-
 static void
 fillDefaultOptions()
 { char tmp[1024];
@@ -891,7 +958,7 @@ fillDefaultOptions()
       ld = cxx;
   }
 
-#ifdef _MSC_VER
+#if defined(HOST_TOOLCHAIN_MSC)
   if ( strcmp(LIB_PL_DEBUG,pllib) == 0 )
     ensureOption(&coptions, "/MDd");
   else ensureOption(&coptions, "/MD");
@@ -908,7 +975,7 @@ fillDefaultOptions()
   free(ctmp);
   ctmp = strdup(tmp);
 #endif
-#if defined(__WINDOWS__) || defined(__CYGWIN__)
+#if defined(HOST_OS_WINDOWS) || /* FIXME: remove this */ defined(__CYGWIN__)
 /* Saved states have the .exe extension under Windows */
   replaceExtension(pltmp, embed_shared ? "dll" : "exe", tmp);
   free(pltmp);
@@ -933,7 +1000,7 @@ fillDefaultOptions()
   defaultProgram(&plinitfile, "none");
   defaultProgram(&plsysinit,  "none");
 
-#ifdef __WINDOWS__
+#if defined(HOST_TOOLCHAIN_MSC)
   sprintf(tmp, "%s/lib", plbase);
 #else
   sprintf(tmp, "%s/lib/%s", plbase, plarch);
@@ -980,12 +1047,8 @@ getPrologOptions()
 	  defaultPath(&plbase, v);
 	else if ( streq(name, "PLARCH") )
 	  defaultPath(&plarch, v);
-#ifdef __CYGWIN__
 	else if ( streq(name, "PLLIBS") )	/* Always required. */
-#else
-	else if ( streq(name, "PLLIBS") && !shared )
-#endif
-	  addOptionString(v);
+	  pllibs = strdup(v);
 	else if ( streq(name, "PLLIB") )
 	  defaultProgram(&pllib, v);
 	else if ( streq(name, "PLLDFLAGS") )
@@ -1015,7 +1078,7 @@ getPrologOptions()
 
     pclose(fd);
 
-#ifdef __WINDOWS__
+#if defined(__WINDOWS__) && defined(HOST_OS_WINDOWS)
     sprintf(buf, "%s/bin/%s", plbase, PROG_PL);
 #else
     sprintf(buf, "%s/bin/%s/%s", plbase, plarch, PROG_PL);
@@ -1111,8 +1174,9 @@ compileFile(const char *compiler, arglist *options, const char *cfile)
       strcpy(ext, EXT_OBJ);
   }
 
-  prependArgList(args, "-c");
-#ifdef __WINDOWS__
+  if ( !opt_E )
+    prependArgList(args, "-c");
+#if defined(HOST_OS_WINDOWS)
   appendArgList(args, "-D__WINDOWS__");
   appendArgList(args, "-D_WINDOWS");
 #endif
@@ -1121,8 +1185,10 @@ compileFile(const char *compiler, arglist *options, const char *cfile)
     appendArgList(args, "-D__SWI_EMBEDDED__");
   concatArgList(args, "-I", &includedirs);
 
-  appendArgList(args, "-o");
-  appendArgList(args, ofile);
+  if ( opt_o || !opt_E )
+  { appendArgList(args, "-o");
+    appendArgList(args, ofile);
+  }
   appendArgList(args, cfile);
 
   callprog(compiler, args);
@@ -1143,7 +1209,7 @@ compileObjectFiles()
     compileFile(cxx, &cppoptions, cppfiles.list[n]);
 }
 
-#ifdef __WINDOWS__
+#if defined(HOST_TOOLCHAIN_MSC)
 char *
 os_path(char *out, const char *in)
 { for(; *in; in++)
@@ -1184,36 +1250,42 @@ void
 linkBaseExecutable()
 { char *cout = out;
 
-#ifndef __WINDOWS__				/* bit of a hack ... */
+#if !defined(HOST_OS_WINDOWS)			/* bit of a hack ... */
   if ( embed_shared )
   { linkSharedObject();
     return;
   }
 #endif
 
-#ifdef __WINDOWS__
+#if defined(HOST_TOOLCHAIN_MSC)
 { char tmp[MAXPATHLEN];
   sprintf(tmp, "/out:%s", cout);
   prependArgList(&ldoptions, tmp);
 }
   concatArgList(&ldoptions, "", &ofiles);	/* object files */
   exportlibdirs();
-  appendArgList(&ldoptions, pllib);		/* -lpl */
+  if ( !nolibswipl )
+  { appendArgList(&ldoptions, pllib);		/* -lswipl */
+    addOptionString(pllibs);
+  }
   concatArgList(&ldoptions, "", &libs);		/* libraries */
   concatArgList(&ldoptions, "", &lastlibs);	/* libraries */
-#else /*__WINDOWS__*/
+#else /* !defined(HOST_TOOLCHAIN_MSC) */
   prependArgList(&ldoptions, cout);
   prependArgList(&ldoptions, "-o");		/* -o ctmp */
   concatArgList(&ldoptions, "", &ofiles);	/* object files */
   concatArgList(&ldoptions, "-L", &libdirs);    /* library directories */
-  appendArgList(&ldoptions, pllib);		/* -lpl */
+  if ( !nolibswipl )
+  { appendArgList(&ldoptions, pllib);		/* -lswipl */
+    addOptionString(pllibs);
+  }
   concatArgList(&ldoptions, "", &libs);		/* libraries */
   concatArgList(&ldoptions, "", &lastlibs);	/* libraries */
-#endif
+#endif /* !defined(HOST_TOOLCHAIN_MSC) */
 
   if ( !nostate )
   {
-#ifdef __WINDOWS__
+#if defined(HOST_TOOLCHAIN_MSC)
     if ( !embed_shared )
     { char buf[MAXPATHLEN];
       appendArgList(&tmpfiles, replaceExtension(cout, "exp", buf));
@@ -1240,7 +1312,7 @@ linkSharedObject()
   { soout = replaceExtension(out, soext, soname);
   }
 
-#ifdef __WINDOWS__
+#if defined(HOST_TOOLCHAIN_MSC)
   prependArgList(&ldoptions, "/dll");
 { char tmp[MAXPATHLEN];
   sprintf(tmp, "/out:%s", soout);
@@ -1248,16 +1320,22 @@ linkSharedObject()
 }
   concatArgList(&ldoptions, "", &ofiles);	/* object files */
   exportlibdirs();
-  appendArgList(&ldoptions, pllib);		/* libpl.lib */
+  if ( !nolibswipl )
+  { appendArgList(&ldoptions, pllib);		/* swipl.lib */
+    addOptionString(pllibs);
+  }
   concatArgList(&ldoptions, "", &libs);		/* libraries */
   concatArgList(&ldoptions, "", &lastlibs);	/* libraries */
-#else /*__WINDOWS__*/
+#else /* !defined(HOST_TOOLCHAIN_MSC) */
 #ifdef __CYGWIN__
   prependArgList(&ldoptions, SO_LDFLAGS);
   prependArgList(&ldoptions, soout);
   prependArgList(&ldoptions, "-o");		/* -o ctmp */
   concatArgList(&ldoptions, "", &ofiles);	/* object files */
-  appendArgList(&ldoptions, pllib);		/* -lpl */
+  if ( !nolibswipl )
+  { appendArgList(&ldoptions, pllib);		/* -lswipl */
+    addOptionString(pllibs);
+  }
   concatArgList(&ldoptions, "-L", &libdirs);    /* library directories */
   concatArgList(&ldoptions, "", &libs);		/* libraries */
   concatArgList(&ldoptions, "", &lastlibs);	/* libraries */
@@ -1275,18 +1353,18 @@ linkSharedObject()
 #endif /*SO_FORMAT_LDFLAGS*/
   concatArgList(&ldoptions, "", &ofiles);	/* object files */
   concatArgList(&ldoptions, "-L", &libdirs);    /* library directories */
-#ifndef O_SHARED_KERNEL
-  if ( embed_shared )
-#endif
-  { appendArgList(&ldoptions, pllib);		/* -lpl */
-  }
   concatArgList(&ldoptions, "", &libs);		/* libraries */
+#ifdef O_SHARED_KERNEL
+  if ( !nolibswipl )
+#endif
+  { appendArgList(&ldoptions, pllib);		/* -lswipl */
+  }
   concatArgList(&ldoptions, "", &lastlibs);	/* libraries */
 #ifdef __BEOS__
   appendArgList(&ldoptions, plexe);		/* last is executable */
 #endif
 #endif /*__CYGWIN__*/
-#endif /*__WINDOWS__*/
+#endif /* !defined(HOST_TOOLCHAIN_MSC) */
 
   callprog(ld, &ldoptions);
 }
@@ -1383,11 +1461,11 @@ createSavedState()
 static void
 copy_fd(int i, int o)
 { char buf[CPBUFSIZE];
-  int n;
+  ssize_t n;
 
   while( (n=read(i, buf, sizeof(buf))) > 0 )
   { while( n > 0 )
-    { int n2;
+    { ssize_t n2;
 
       if ( (n2 = write(o, buf, n)) > 0 )
       { n -= n2;
@@ -1405,7 +1483,7 @@ copy_fd(int i, int o)
 }
 
 
-#ifdef __WINDOWS__
+#if defined(HOST_TOOLCHAIN_MSC)
 void
 saveExportLib()
 { char ibuf[MAXPATHLEN];
@@ -1427,7 +1505,7 @@ saveExportLib()
     }
   }
 }
-#endif /*__WINDOWS__*/
+#endif /* defined(HOST_TOOLCHAIN_MSC) */
 
 
 void
@@ -1436,7 +1514,7 @@ createOutput()
 
   if ( verbose )
   {
-#ifdef __WINDOWS__
+#if defined(HOST_TOOLCHAIN_MSC)
     printf("\tcopy /b %s+%s %s\n", out, pltmp, out);
 #else
     printf("\tcat %s >> %s\n", pltmp, out);
@@ -1540,19 +1618,6 @@ main(int argc, char **argv)
     special = 2;
   else
     special = 0;
-					/* swipl-ld [-pl x] -E ...: behave as cpp */
-  if ( argc-special > 0 && streq(argv[special], "-E") )
-  { arglist cppoptions;
-    int i;
-
-    memset(&cppoptions, 0, sizeof(cppoptions));
-    for(i=special+1 ; i < argc; i++)
-      appendArgList(&cppoptions, argv[i]);
-
-    callprog(PROG_CPP, &cppoptions);
-
-    return 0;
-  }
 					  /* swipl-ld [-pl x] -v: verbose */
   if ( argc-special == 1 && streq(argv[special], "-v") )
   { arglist coptions;
@@ -1569,7 +1634,36 @@ main(int argc, char **argv)
 
   parseOptions(argc, argv);
   defaultProgram(&pl, PROG_PL);
-  getPrologOptions();
+
+  if ( build_defaults )
+  { nostate = TRUE;			/* not needed and Prolog won't run */
+    defaultProgram(&cc, C_CC);
+#ifdef PLBASE
+    defaultPath(&plbase, PLBASE);
+#else
+    defaultPath(&plbase, PLHOME);
+#endif
+    defaultPath(&plarch, PLARCH);
+    defaultProgram(&pllib, C_PLLIB);
+    addOptionString(C_LIBS);
+    appendArgList(&ldoptions, C_LDFLAGS);
+    appendArgList(&coptions, C_CFLAGS);
+    appendArgList(&cppoptions, C_CFLAGS);
+#ifdef SO_EXT
+    soext = strdup(SO_EXT);
+#endif
+#ifdef O_PLMT
+    ensureOption(&coptions, "-D_REENTRANT");
+    ensureOption(&cppoptions, "-D_REENTRANT");
+#ifdef _THREAD_SAFE			/* FreeBSD */
+    ensureOption(&coptions, "-D_THREAD_SAFE");
+    ensureOption(&cppoptions, "-D_THREAD_SAFE");
+#endif
+#endif
+  } else
+  { getPrologOptions();
+  }
+
   fillDefaultOptions();
 
   if ( show_version )
